#define LOOKUP_MAX_PROBES 10

extern const unsigned dcm_vrtable_from_str_len;
extern const uint8_t dcm_vrtable_from_str_empty;
extern const uint8_t dcm_vrtable_from_str_dict[];

extern const unsigned dcm_attribute_from_tag_len;
extern const uint16_t dcm_attribute_from_tag_empty;
extern const uint16_t dcm_attribute_from_tag_dict[];

extern const unsigned dcm_attribute_from_keyword_len;
extern const uint16_t dcm_attribute_from_keyword_empty;
extern const uint16_t dcm_attribute_from_keyword_dict[];

