#include <stdint.h>

const unsigned dcm_vrtable_from_str_len = 136;
const uint8_t dcm_vrtable_from_str_empty = 0xff;
const uint8_t dcm_vrtable_from_str_dict[136] = {
0xf, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0x15, 
0xff, 0xff, 0x14, 0xff, 0xff, 0xff, 0xff, 0xff, 
0xa, 0x1c, 0xff, 0xff, 0x8, 0x1, 0xff, 0xff, 
0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 
0xff, 0xff, 0xff, 0xff, 0xff, 0x16, 0xff, 0x1e, 
0xff, 0x2, 0x9, 0xff, 0xff, 0x1d, 0xff, 0x13, 
0x11, 0x19, 0xff, 0x7, 0xff, 0xff, 0xff, 0xff, 
0x0, 0xd, 0x12, 0xff, 0xff, 0xff, 0x1a, 0xff, 
0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 
0x5, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 
0xff, 0xff, 0xc, 0x4, 0xff, 0xff, 0xff, 0xff, 
0xff, 0xff, 0x21, 0xff, 0xff, 0xff, 0x20, 0xff, 
0xb, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 
0xff, 0x17, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 
0xff, 0xff, 0xff, 0xff, 0x10, 0x18, 0x1f, 0x6, 
0xff, 0xff, 0xff, 0xff, 0xff, 0x1b, 0xff, 0xff, 
0xff, 0xff, 0xff, 0x3, 0xff, 0xe, 0xff, 0xff, 
};

const unsigned dcm_attribute_from_tag_len = 19560;
const uint16_t dcm_attribute_from_tag_empty = 0xffff;
const uint16_t dcm_attribute_from_tag_dict[19560] = {
0xac1, 0xefe, 0xffff, 0xffff, 0xffff, 0xffff, 0x1e5, 0xffff, 
0xffff, 0xffff, 0x12ea, 0xffff, 0xffff, 0xffff, 0x185, 0xb9f, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xbf9, 
0xffff, 0xffff, 0xffff, 0xd45, 0xffff, 0xffff, 0x7f1, 0xffff, 
0xffff, 0xffff, 0xffff, 0x12e9, 0xaa6, 0xffff, 0xffff, 0xffff, 
0x4, 0x6d5, 0x1ff, 0xffff, 0xffff, 0xffff, 0x268, 0xffff, 
0xb9b, 0xffff, 0x192, 0xffff, 0x54b, 0xffff, 0xffff, 0x1209, 
0xffff, 0xffff, 0x7d7, 0x7de, 0xffff, 0xffff, 0xabe, 0xffff, 
0xaed, 0xffff, 0xffff, 0x10a1, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xef5, 0xffff, 0xd1e, 0xffff, 0x850, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x1050, 
0x1ec, 0x553, 0xffff, 0x33d, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xd48, 0xffff, 0x72a, 0x403, 0xffff, 0x985, 0xa66, 
0xda3, 0x630, 0xb11, 0xffff, 0xffff, 0xffff, 0xffff, 0x462, 
0xffff, 0xffff, 0x337, 0xffff, 0x12cd, 0xffff, 0xffff, 0xc9, 
0xd2d, 0xffff, 0xffff, 0x1008, 0xffff, 0xffff, 0xafb, 0xdaf, 
0x1200, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x822, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x939, 0x1ad, 0x941, 
0xf2c, 0xffff, 0xc1f, 0xffff, 0xffff, 0x85, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0x12ab, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xb44, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xa56, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x23b, 0x984, 0xffff, 
0xeb, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0x8, 0xe, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x52b, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0x57d, 0xffff, 0xffff, 0xc0, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0x3a3, 0xa99, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x124b, 0xffff, 
0xffff, 0x55d, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xde5, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xedb, 
0xffff, 0xffff, 0xc85, 0xaec, 0xffff, 0xffff, 0x53c, 0x28c, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xaa4, 0x91c, 
0x52c, 0xffff, 0xffff, 0xffff, 0x920, 0xffff, 0xffff, 0xffff, 
0xffff, 0x996, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0x671, 0xffff, 0xd4d, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0x6b4, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xd25, 
0xffff, 0x5e7, 0x6b9, 0xe78, 0xffff, 0x74a, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0x5d6, 0xeaa, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xe82, 0xffff, 0xffff, 0xffff, 
0xffff, 0xfeb, 0xffff, 0xffff, 0xea4, 0xffff, 0xffff, 0xffff, 
0xffff, 0xb80, 0x257, 0xffff, 0xffff, 0xffff, 0x1258, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x489, 0x963, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x91e, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xe89, 0xffff, 
0xffff, 0xffff, 0x120f, 0x1228, 0x9e, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0x9b2, 0xffff, 0x66d, 0xffff, 
0xffff, 0xffff, 0xffff, 0x2a, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0x4a2, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0x9c8, 0xb95, 0xcf4, 0xffff, 
0x1051, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0x703, 0xffff, 0xffff, 0xb40, 0x95e, 0xffff, 0x100e, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x82b, 
0xc59, 0xbc5, 0x1153, 0x7d8, 0xffff, 0xffff, 0xffff, 0xffff, 
0x5cb, 0x231, 0x25a, 0x701, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xc5, 0xffff, 0xffff, 0xffff, 0x105f, 
0xffff, 0xffff, 0xffff, 0x12e, 0x1054, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xc4c, 0xffff, 0x3c8, 0xffff, 0x11b6, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xb7b, 0xc7e, 0xffff, 0xffff, 0xffff, 0x10e1, 
0xffff, 0x1e4, 0xffff, 0x411, 0x59c, 0xffff, 0xffff, 0xab2, 
0x912, 0xca5, 0xe48, 0xffff, 0xffff, 0xffff, 0x19, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xe55, 0xffff, 0xc83, 0xffff, 
0x8fa, 0xffff, 0xffff, 0xffff, 0xc5c, 0xffff, 0xffff, 0x9ae, 
0xffff, 0x611, 0xd5e, 0xd6e, 0xffff, 0xffff, 0xffff, 0xffff, 
0x669, 0xffff, 0xffff, 0xffff, 0x6fc, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xe9a, 0xffff, 0xffff, 0xffff, 
0xffff, 0x56, 0xffff, 0x78e, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x731, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xfd5, 0xffff, 0xffff, 
0x5c7, 0xffff, 0x11c2, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xff3, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xbaf, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0x74, 0xffff, 0xc91, 0xffff, 0xffff, 0xffff, 0xffff, 
0xae1, 0xffff, 0x37a, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xead, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x61d, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x646, 
0xffff, 0xd75, 0xffff, 0xffff, 0xc4f, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x14d, 0xffff, 0xffff, 
0xffff, 0xffff, 0x3e, 0xcab, 0xffff, 0x127b, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xe1c, 0xffff, 0xffff, 0xffff, 
0xffff, 0x29a, 0xffff, 0xffff, 0xffff, 0x5b6, 0xffff, 0xffff, 
0xffff, 0x667, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0x8c8, 0xffff, 0x532, 0x961, 0xb31, 0x47f, 0x97a, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xaa3, 0xffff, 0xffff, 0xeea, 0xc3d, 0xffff, 
0xffff, 0xffff, 0xffff, 0x81a, 0xffff, 0xffff, 0xffff, 0x7ff, 
0xffff, 0xffff, 0xffff, 0xffff, 0x44d, 0xa68, 0xffff, 0xfa2, 
0xdd7, 0x2b4, 0x3d2, 0xffff, 0xffff, 0x978, 0xffff, 0x444, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xfe, 
0xe34, 0x114f, 0xffff, 0xffff, 0x4b9, 0x68a, 0xffff, 0xffff, 
0xda1, 0xf27, 0xf45, 0xffff, 0xffff, 0xffff, 0xcb3, 0xffff, 
0xffff, 0xa6b, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0x3f6, 0x9d7, 0xffff, 0xcc1, 0xffff, 
0xffff, 0xffff, 0x2bc, 0x24b, 0xffff, 0xb3b, 0xffff, 0xffff, 
0xffff, 0xffff, 0x77d, 0xffff, 0xffff, 0x7a2, 0x8b8, 0xffff, 
0xffff, 0xffff, 0x13b, 0x6b2, 0x505, 0xffff, 0xb6a, 0x1076, 
0xffff, 0x810, 0xffff, 0x47e, 0xcae, 0xffff, 0xffff, 0xc4e, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0x75a, 0xd12, 0x126f, 0x3a5, 0xffff, 
0xffff, 0x893, 0xeda, 0x114a, 0xffff, 0xffff, 0xffff, 0xf97, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xdb9, 0xffff, 
0xffff, 0xffff, 0xffff, 0x78c, 0xe8f, 0xffff, 0x1137, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x2fd, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xe9e, 0xffff, 0xffff, 0x638, 
0xe52, 0xffff, 0xffff, 0xffff, 0xffff, 0x8d6, 0xffff, 0xffff, 
0xc7f, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0x1016, 0xd76, 0xffff, 0x8d8, 0xffff, 0x1005, 0x875, 0xffff, 
0xffff, 0xffff, 0x9ed, 0x15a, 0xffff, 0xffff, 0x367, 0xffff, 
0xffff, 0xffff, 0xffff, 0x2f8, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xccd, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0x282, 0xffff, 0xffff, 0x998, 0xffff, 0xffff, 
0x8d4, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xca2, 0xffff, 
0xffff, 0x438, 0x9d2, 0x1254, 0xffff, 0x9fc, 0xbb1, 0xffff, 
0xffff, 0xffff, 0xffff, 0x105b, 0xffff, 0xffff, 0x11b1, 0x1000, 
0xffff, 0xffff, 0xffff, 0x738, 0xffff, 0xb70, 0xffff, 0xffff, 
0xffff, 0xffff, 0x12fa, 0xc44, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0x31, 0xffff, 0xffff, 0x6be, 0xd7c, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x14a, 0x931, 
0x59, 0x64d, 0xffff, 0xffff, 0x351, 0x81e, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0x733, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0x104b, 0xffff, 0xffff, 0x986, 0xffff, 0x12ce, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x576, 0xb20, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x23a, 
0xffff, 0xffff, 0xffff, 0xe5e, 0xffff, 0xffff, 0xbd6, 0x114b, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xb88, 0xe3a, 0xffff, 0xffff, 0xffff, 0xd06, 0xffff, 
0xffff, 0xffff, 0x338, 0xffff, 0xffff, 0xffff, 0xffff, 0x327, 
0x33f, 0x1181, 0xffff, 0xffff, 0xc66, 0x12fb, 0xffff, 0xffff, 
0xffff, 0x5bd, 0x1068, 0xffff, 0xffff, 0xffff, 0xffff, 0x6ac, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0x136, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0x429, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0x3c, 0x243, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x3c7, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xa2d, 0xbb2, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x94b, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0x77, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xf1f, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x68d, 0x81f, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x796, 
0x596, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x7f3, 
0x916, 0xffff, 0xffff, 0xce4, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0x12a8, 0x300, 0x947, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0x2a1, 0xd98, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x1ed, 0x16c, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0x4ef, 0xffff, 0xffff, 0xffff, 
0xffff, 0xde0, 0xe6c, 0xffff, 0x12db, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xd58, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x9cc, 0x12bf, 
0x11d5, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0x974, 0xffff, 0xffff, 0x14c, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0x291, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x11c5, 0xef2, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0x110, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0x80a, 0xffff, 0xffff, 0xffff, 0xffff, 0xff5, 
0xffff, 0xffff, 0x165, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0x8c1, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0x99e, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0x1205, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0x35c, 0xffff, 0x173, 0xd7e, 0x41c, 0x12f1, 0xffff, 0xffff, 
0xffff, 0xf8f, 0xa51, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0x1bb, 0xffff, 0xffff, 0x4fc, 0xffff, 0xc22, 0xed0, 0xffff, 
0xffff, 0x9a1, 0xffff, 0x559, 0x434, 0x5c, 0xffff, 0x41f, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xe0e, 0x66, 0x1198, 0xffff, 0x124, 0xde2, 0xffff, 0xffff, 
0x4e3, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x977, 
0xb69, 0x928, 0x537, 0xffff, 0xe20, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0x12a2, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0x4e1, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xf59, 0x4f9, 0x1135, 0xffff, 0xffff, 0xffff, 
0xf9e, 0xffff, 0xffff, 0xcfa, 0xffff, 0x111d, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0x348, 0x436, 0xaee, 0xffff, 
0xd57, 0xffff, 0xffff, 0x331, 0xa82, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xb22, 0xfce, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xe15, 0x101a, 0xffff, 0xa97, 0xb83, 0xc46, 0xffff, 0xffff, 
0xffff, 0xffff, 0x121c, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xa6, 0xffff, 0xffff, 
0xffff, 0xffff, 0x1088, 0xffff, 0xffff, 0xffff, 0x856, 0x696, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x44f, 0xffff, 
0xffff, 0xa7d, 0xffff, 0xce3, 0xffff, 0xffff, 0xffff, 0xffff, 
0x10b0, 0xffff, 0xffff, 0xa48, 0xc62, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0x1166, 0xff2, 0xffff, 0xffff, 0xfe1, 0xffff, 
0xffff, 0xffff, 0xffff, 0x7e4, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x130f, 
0x25f, 0xffff, 0xffff, 0xffff, 0xe67, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0x248, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xc8, 0xf69, 0x1175, 0xffff, 0x10c7, 0xffff, 0xffff, 
0xffff, 0xffff, 0x6c8, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xcff, 0xeb8, 0xffff, 0xffff, 
0x99f, 0xffff, 0xffff, 0xffff, 0x24c, 0x286, 0xf9b, 0xffff, 
0xffff, 0xffff, 0x59a, 0x244, 0x74f, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x9a, 0xffff, 0xffff, 
0xffff, 0xe14, 0xa18, 0xffff, 0xffff, 0xffff, 0xffff, 0xec5, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xa1d, 0xffff, 0xffff, 0x12c7, 0xf43, 
0xffff, 0xffff, 0x2da, 0xc1, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x6c3, 0xffff, 0xffff, 
0xffff, 0xffff, 0x10be, 0xffff, 0xffff, 0xddf, 0xa0b, 0xffff, 
0x1074, 0xf42, 0xffff, 0xffff, 0x12aa, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0x634, 0xffff, 0xffff, 0x2bd, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xc0a, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0x302, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xa27, 0xffff, 0x12a4, 0xffff, 0xffff, 0xffff, 0xffff, 0x3e8, 
0xffff, 0xffff, 0xffff, 0xffff, 0xe12, 0xffff, 0x492, 0xffff, 
0xffff, 0xffff, 0x525, 0x197, 0xffff, 0xffff, 0xffff, 0x798, 
0xffff, 0xca, 0xc3f, 0xff9, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0x2e2, 0xffff, 0x7ef, 0xffff, 0xffff, 0xb16, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x1be, 0x989, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xe97, 0x6b0, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x235, 0xffff, 
0xf53, 0xd69, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0x26a, 0xffff, 0xffff, 0xffff, 
0xcbb, 0xdf2, 0xffff, 0xffff, 0xffff, 0xb47, 0xffff, 0xffff, 
0xffff, 0x9bf, 0xfec, 0xffff, 0xffff, 0xffff, 0xff6, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xf68, 
0xffff, 0x12a6, 0xffff, 0x1128, 0xffff, 0xd10, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0x5c1, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x14f, 
0xffff, 0xffff, 0x2ba, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0x1a3, 0xffff, 0x1c1, 0x5e9, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0x126d, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x722, 
0xffff, 0xffff, 0x45b, 0xffff, 0xffff, 0x425, 0x6e9, 0xa05, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0x6af, 0xffff, 0x5de, 0xffff, 0xffff, 0x11f8, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xee2, 0xffff, 0xd0b, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0x9a8, 0xe53, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0x1189, 0xffff, 0xb79, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0x479, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0x27d, 0xffff, 0xffff, 0x109f, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0x12ac, 0x7c4, 0xffff, 0xffff, 0xd1c, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0x7c5, 0xbb8, 0xfda, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0x7e6, 0x10ef, 0xffff, 0xffff, 0xffff, 
0xffff, 0x114e, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x39f, 
0x12ca, 0x1301, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xd2b, 0xffff, 0xffff, 0xffff, 0xffff, 0x464, 0xffff, 0x115c, 
0xffff, 0xffff, 0xf22, 0xa5d, 0xffff, 0xffff, 0xffff, 0xffff, 
0x12a5, 0x715, 0x16f, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xc61, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x12d3, 0xffff, 0x126e, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xbcf, 0xffff, 
0xffff, 0x30f, 0xdfd, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0x92f, 0x976, 0xffff, 0xffff, 0xffff, 0xe2d, 
0xad8, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0x8b2, 0x662, 0xffff, 0x7f2, 0xb1a, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x16b, 
0xffff, 0xffff, 0xffff, 0xcc2, 0xa00, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x1239, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xa79, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x739, 0xb0c, 
0xffff, 0xffff, 0xffff, 0x957, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xed4, 0xffff, 0xffff, 0xdea, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0x735, 0xffff, 0xd6c, 0x10bd, 0xffff, 0x7af, 
0x69d, 0xffff, 0xffff, 0xffff, 0x3ef, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0x5c9, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xe4, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0x4bc, 0xffff, 0xffff, 0xc18, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x608, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x89, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x760, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xef9, 
0x67a, 0x105d, 0x1185, 0xffff, 0x209, 0xffff, 0x245, 0xffff, 
0xffff, 0xffff, 0x794, 0x2ef, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0x7ce, 0xffff, 0x788, 0xffff, 0xffff, 0xffff, 0xffff, 
0x2db, 0x6e, 0xab, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xb15, 0xffff, 0xe07, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x12d0, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0x1163, 0x64a, 0xffff, 0xffff, 
0xffff, 0xd11, 0xffff, 0xffff, 0xffff, 0x8a8, 0xffff, 0x97, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x512, 0xffff, 0xffff, 
0xffff, 0xffff, 0x8df, 0xc55, 0x109, 0x88c, 0xe56, 0xeb0, 
0xff0, 0x1271, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0x122a, 0xb18, 0xffff, 0x233, 0xffff, 0xd0, 
0xffff, 0xffff, 0xffff, 0x374, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0x120c, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0x373, 0xffff, 0xffff, 0xffff, 0xb2b, 0x68b, 0x10e8, 0x12d5, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x5ac, 0x568, 0xffff, 
0xffff, 0xb06, 0xffa, 0xffff, 0xffff, 0xa7f, 0xffff, 0xffff, 
0x7f6, 0xffff, 0x3c5, 0xf26, 0x27f, 0xffff, 0xc58, 0xffff, 
0xa7e, 0xb09, 0xffff, 0xffff, 0xffff, 0x1131, 0xeef, 0xffff, 
0x3d3, 0xffff, 0x1034, 0x1155, 0x3fa, 0x1260, 0xffff, 0xffff, 
0x1e9, 0x5be, 0x5dc, 0x1206, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x1187, 0xffff, 0xffff, 
0x6a, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x125f, 0xffff, 
0xffff, 0x7d0, 0x997, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xd70, 0xffff, 0xffff, 
0x1274, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x125, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xa44, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0x1bc, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0x4da, 0xffff, 0xffff, 0xffff, 0x180, 0xffff, 0xffff, 
0xffff, 0xffff, 0x766, 0xffff, 0x982, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xc8f, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0x15e, 0x452, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x20, 0x1106, 0x108, 
0x70e, 0xffff, 0xffff, 0x86f, 0xffff, 0x461, 0xb45, 0xffff, 
0xffff, 0x58d, 0xffff, 0xffff, 0xf1, 0xffff, 0xffff, 0xffff, 
0x47d, 0xf87, 0xffff, 0x11a1, 0xffff, 0xffff, 0x305, 0xffff, 
0xffff, 0xffff, 0xffff, 0x12d2, 0xffff, 0x2fb, 0x82f, 0xaeb, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x8ef, 0xffff, 
0xffff, 0xffff, 0xffff, 0xe7d, 0xffff, 0x10d6, 0xffff, 0x76f, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x1031, 0xee6, 
0xffff, 0xffff, 0xaca, 0xffff, 0xffff, 0xffff, 0x1158, 0xffff, 
0x66b, 0xffff, 0x786, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0x450, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x929, 
0x777, 0xd21, 0xdb0, 0x1178, 0xa91, 0xffff, 0xffff, 0xffff, 
0xbba, 0x48a, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x664, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x1013, 0x26d, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0x528, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0x172, 0x8cd, 0xffff, 0xffff, 0xffff, 0x7b6, 
0xffff, 0xffff, 0xffff, 0x511, 0x613, 0xffff, 0xffff, 0xffff, 
0xffff, 0x10e7, 0xffff, 0x65f, 0xffff, 0x7b5, 0xffff, 0xd23, 
0xffff, 0xffff, 0x4a5, 0xffff, 0x6e8, 0xffff, 0x6fd, 0xffff, 
0xffff, 0x11df, 0xb57, 0xffff, 0xffff, 0xffff, 0xc3, 0xab3, 
0xffff, 0x1f8, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0x682, 0x995, 0x12b0, 0xffff, 0x1ea, 0xffff, 
0xffff, 0xd7a, 0x11db, 0xcd7, 0xffff, 0xffff, 0x3cb, 0xffff, 
0xffff, 0x1122, 0x130c, 0xffff, 0xffff, 0xffff, 0x98b, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0x1193, 0xffff, 0xffff, 0x6df, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0x19f, 0xffff, 0xc5f, 0xffff, 0x240, 
0x84, 0x12d6, 0xdc2, 0xffff, 0x125e, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xa55, 0x125b, 0xffff, 
0xffff, 0xffff, 0xffff, 0xd30, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0x7a9, 0x339, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x39a, 0xfa, 0x755, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x2c1, 0x695, 
0xf25, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xcb6, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0x1211, 0x9c9, 0xffff, 0xffff, 
0xffff, 0x7db, 0xffff, 0x311, 0xffff, 0xffff, 0xffff, 0x6c0, 
0xffff, 0xffff, 0x1f2, 0xffff, 0xffff, 0xf49, 0x546, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0x1319, 0x1113, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0x2fe, 0xffff, 0xa89, 0xffff, 0x101f, 
0x1230, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0x547, 0xffff, 0xffff, 0x44b, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xc29, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffb, 0x1277, 0xffff, 0x9b, 
0xffff, 0xffff, 0xffff, 0xda, 0xffff, 0x1123, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0x72f, 0xffff, 0x103d, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xe1f, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x5e5, 
0xffff, 0xffff, 0xffff, 0xffff, 0xaae, 0xd93, 0xffff, 0xffff, 
0xffff, 0x474, 0xffff, 0xffff, 0xffff, 0xffff, 0x1176, 0xffff, 
0xffff, 0xffff, 0xffff, 0x122e, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xdc7, 0xffff, 0xfb8, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0x26b, 0xffff, 0xffff, 0xffff, 0x624, 0xbf1, 
0xccc, 0xffff, 0xffff, 0xffff, 0x128, 0xffff, 0xffff, 0xffff, 
0xffff, 0x12c8, 0xffff, 0xc32, 0xffff, 0x11b, 0xffff, 0xaac, 
0xffff, 0x126, 0x4f5, 0xffff, 0xc6d, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x1145, 0xffff, 0xffff, 
0xffff, 0xa71, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xce5, 0xffff, 0xa93, 0xb86, 0xffff, 
0xffff, 0xffff, 0xc1a, 0xffff, 0x12cc, 0x504, 0xffff, 0xffff, 
0x53e, 0x12ad, 0xffff, 0x105, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0x11b0, 0xffff, 0xffff, 0xafd, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0x170, 0x3fe, 0xffff, 0xffff, 
0xa3e, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xb93, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xd71, 0x628, 0xf3f, 0xffff, 0x1081, 0xdb4, 0xffff, 0xffff, 
0xffff, 0x7c6, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xcfe, 0xffff, 0xffff, 
0xffff, 0xffff, 0xbeb, 0xffff, 0xffff, 0xffff, 0x6de, 0xecd, 
0xf93, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x70f, 
0x8f7, 0xdbf, 0xffff, 0xffff, 0xffff, 0x67b, 0x150, 0xffff, 
0xffff, 0x1218, 0x23d, 0xffff, 0x8b9, 0x774, 0xf17, 0xffff, 
0xffff, 0xeae, 0xffff, 0xffff, 0xffff, 0xffff, 0x12a, 0x6c2, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xd4, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x659, 0x5ad, 0x973, 
0xb48, 0xffff, 0x2e9, 0x747, 0x8ec, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0x363, 0xffff, 0xffff, 0x1c7, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x11c6, 
0xffff, 0xffff, 0x110c, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0x169, 0xa67, 0xb54, 0xa84, 0x105c, 
0xbdd, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x539, 0xffff, 
0xffff, 0xffff, 0xce1, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x1eb, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x7b, 0x3db, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0x1132, 0xe84, 0xffff, 0xffff, 0x8fe, 0xcc7, 0x683, 
0xcb0, 0x8c, 0xe6, 0x37d, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0x10fc, 0xffff, 0x145, 0xffff, 0xae8, 
0xffff, 0xffff, 0xffff, 0x44e, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0x454, 0xffff, 0x36f, 0xcd8, 0xffff, 0xffff, 
0xf62, 0x252, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x11f3, 
0xffff, 0x2b6, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xf8, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x45e, 
0xffff, 0x1043, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xd56, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xe29, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0x307, 0x46c, 0x838, 0xfb6, 
0xffff, 0xffff, 0xffff, 0x60, 0x11b3, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xda8, 
0x1234, 0xffff, 0xffff, 0x1d8, 0xffff, 0xffff, 0x1172, 0xffff, 
0xffff, 0xb0f, 0x101d, 0xffff, 0xffff, 0xffff, 0xffff, 0xd97, 
0xffff, 0xffff, 0xffff, 0x404, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xf5b, 0xffff, 0x5da, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x697, 0xffff, 0xff, 
0xcef, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x6d0, 0xffff, 0x11ed, 
0xffff, 0xffff, 0x851, 0xbe3, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xe3b, 0xffff, 0xa7, 
0xffff, 0x27c, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0x8b0, 0x107d, 0xffff, 0x4ba, 0x10fe, 0xd02, 0xc92, 
0x3e6, 0xffff, 0x2ab, 0x12f8, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x1082, 0xffff, 0xb2e, 
0xffff, 0xffff, 0xffff, 0xffff, 0x7c0, 0xffff, 0xffff, 0x7d4, 
0x573, 0xffff, 0xffff, 0xffff, 0xffff, 0xa9f, 0xffff, 0xffff, 
0xffff, 0x393, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0x37e, 0x437, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x1040, 0x6e5, 0xffff, 
0xffff, 0x762, 0xffff, 0x11e, 0xffff, 0x1a9, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xf9f, 0xffff, 0xffff, 0xffff, 0xffff, 
0x53f, 0xffff, 0xffff, 0xffff, 0xe3f, 0xffff, 0x887, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0x10c9, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xadc, 0xffff, 0x6fa, 0xffff, 0xd3e, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0x808, 0xffff, 0xffff, 0xffff, 0x211, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0x6ab, 0x99c, 0xdf8, 0xffff, 
0xffff, 0xffff, 0x1ca, 0xffff, 0xffff, 0xffff, 0xffff, 0x289, 
0xffff, 0xffff, 0xffff, 0xbf, 0x23c, 0xb6b, 0xc0d, 0xdca, 
0x1265, 0xffff, 0xffff, 0xffff, 0xffff, 0x130, 0xffff, 0xffff, 
0xffff, 0x1304, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xcd4, 0xffff, 0xffff, 0x20d, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0x900, 0x1104, 0xffff, 0xffff, 0x57, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0x38f, 0x10e3, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0x883, 0xffff, 0xffff, 0x80f, 0xffff, 0xffff, 0xffff, 0xffff, 
0xdf, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0x3e2, 0xffff, 0xc78, 0xffff, 0xffff, 
0xa15, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0x804, 0xa20, 0x1097, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xbcc, 0x468, 0xffff, 0x3c2, 0xed2, 
0xffff, 0xffff, 0xffff, 0xc03, 0xffff, 0xc81, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x126c, 0xffff, 
0xffff, 0xffff, 0x773, 0x4c1, 0xe81, 0x118f, 0x108a, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0x73f, 0xffff, 0xffff, 0xffff, 0x897, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xc09, 0xea5, 0xc15, 0x12ff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xbdb, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0x9b7, 0xffff, 0x27, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xcb9, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xa36, 0xffff, 0xffff, 0xffff, 
0xffff, 0x91, 0xffff, 0xffff, 0x677, 0x2cd, 0x749, 0xaf7, 
0xb4a, 0xffff, 0xffff, 0xffff, 0x63f, 0xffff, 0x2f6, 0x6d8, 
0xffff, 0xffff, 0xffff, 0xffff, 0x536, 0x1118, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0x57e, 0x9da, 0xffff, 0xffff, 0xffff, 
0xffff, 0x952, 0xffff, 0xc07, 0x151, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xe0c, 0xffff, 0xffff, 0x76e, 0xffff, 
0xffff, 0xffff, 0xbd4, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0x1183, 0xffff, 0xffff, 0xffff, 0xffff, 
0xd19, 0xffff, 0xffff, 0xffff, 0xffff, 0x516, 0xffff, 0xffff, 
0xffff, 0xa0c, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x360, 
0xffff, 0x12e4, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xf4c, 0x130a, 0xffff, 0xffff, 0xffff, 
0xee9, 0xffff, 0x782, 0x814, 0xd31, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xc24, 0x1245, 
0xffff, 0xffff, 0xffff, 0xffff, 0x287, 0xffff, 0xffff, 0x889, 
0x1152, 0xffff, 0xffff, 0xffff, 0xb90, 0xffff, 0xffff, 0xffff, 
0xffff, 0xfdb, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xe42, 0xffff, 0xffff, 0xcd, 
0xffff, 0xd5b, 0xa5b, 0xffff, 0x95, 0x744, 0xffff, 0x55e, 
0xffff, 0xffff, 0xffff, 0xffff, 0x8e7, 0xffff, 0xffff, 0xffff, 
0xb5, 0x3f8, 0x9ce, 0xb21, 0x1112, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0x886, 0xffff, 0xffff, 0xffff, 0x9c7, 0x4b5, 
0x7a1, 0xffff, 0x2a2, 0xffff, 0xbb3, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0x5ea, 0x98a, 0xffff, 0x1d, 0x10f2, 0x115a, 
0x1248, 0xffff, 0x125d, 0xffff, 0x4d6, 0xffff, 0x1c2, 0xbf8, 
0xffff, 0x1ac, 0xffff, 0xffff, 0xffff, 0x8e5, 0xffff, 0xffff, 
0x1299, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0x926, 0xffff, 0xffff, 0xffff, 
0x284, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xd01, 0xffff, 0xffff, 
0xffff, 0xffff, 0x1a6, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xea, 0xffff, 0x41b, 0xffff, 0x6e0, 
0x6fe, 0x85d, 0xf85, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xea6, 0xffff, 0xffff, 0xffff, 0xffff, 0xa77, 0xffff, 0x86c, 
0x1195, 0xffff, 0xffff, 0xffff, 0xffff, 0xa4d, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x661, 
0x18b, 0xe23, 0xffff, 0xffff, 0xffff, 0xffff, 0x831, 0x820, 
0x8ce, 0xffff, 0x11e0, 0xffff, 0x1146, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xd7f, 0xbff, 0xffff, 0xe74, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x10b, 0x626, 
0xc2c, 0xffff, 0xffff, 0x60c, 0xda0, 0xffff, 0xffff, 0xffff, 
0xc82, 0xffff, 0xffff, 0xffff, 0xffff, 0x493, 0xffff, 0xffff, 
0x1252, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xf7e, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0x5ab, 0xffff, 0xe8d, 0xffff, 0xffff, 0xd84, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0x5eb, 0xffff, 0x9e4, 0xffff, 0xffff, 0x88d, 0xffff, 
0x106e, 0x118a, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0x737, 0xffff, 0x9, 0x1243, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xee0, 0x21b, 0x904, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0x1a1, 0xffff, 0xe4e, 0x79c, 0xffff, 0xffff, 
0xffff, 0xffff, 0x12e2, 0xac9, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0x43b, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0x6b3, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xf6f, 0xffff, 0xffff, 0xffff, 
0xffff, 0x128e, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xd3d, 
0xffff, 0xffff, 0xb17, 0xffff, 0xffff, 0xffff, 0xdd, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xc53, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0x139, 0xffff, 0xffff, 0x50b, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0x124a, 0xffff, 0xffff, 0xffff, 
0x522, 0x70a, 0xffff, 0xffff, 0xffff, 0x878, 0xffff, 0xffff, 
0x645, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x48, 
0xe73, 0xffff, 0x8ac, 0xffff, 0xffff, 0x0, 0xd2c, 0xffff, 
0xaa9, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0x408, 0xffff, 0x666, 0xffff, 0xe1d, 0xffff, 
0xffff, 0x629, 0xc2e, 0xdd4, 0x48f, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x1314, 
0xffff, 0xffff, 0x107, 0xffff, 0xffff, 0xffff, 0xffff, 0x557, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0x62d, 0xffff, 0xffff, 0x711, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x114, 0x1100, 
0x1186, 0xffff, 0xffff, 0x8d5, 0x1130, 0xffff, 0xffff, 0xfb4, 
0xffff, 0x49c, 0x545, 0xffff, 0xffff, 0x1237, 0xffff, 0xd, 
0xffff, 0x936, 0x99a, 0xffff, 0xffff, 0xffff, 0x1de, 0x12c9, 
0xffff, 0xffff, 0xffff, 0xffff, 0x1224, 0xffff, 0xffff, 0xffff, 
0xffff, 0xec, 0xa42, 0x111b, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0x9ca, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0x12ec, 0xffff, 0xffff, 0x21, 0xffff, 0x6cf, 
0xffff, 0xffff, 0x10bf, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0x109c, 0xffff, 0xf2b, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xb05, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xbd5, 0xffff, 0x121f, 0x4fd, 0xffff, 
0xffff, 0xffff, 0xf00, 0xffff, 0xffff, 0xffff, 0xffff, 0x5e, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0x55f, 0x10a2, 0xffff, 0xffff, 0xffff, 0x1026, 
0xffff, 0xffff, 0xe11, 0xffff, 0xffff, 0xffff, 0xffff, 0xac3, 
0xc33, 0xffff, 0xffff, 0xd4f, 0xffff, 0xffff, 0x7c, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x79d, 
0x7b0, 0xffff, 0xffff, 0xffff, 0x34d, 0x7ca, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xcec, 0xffff, 0xffff, 0xffff, 
0x442, 0xffff, 0xffff, 0xffff, 0xc80, 0xfbe, 0x132, 0xffff, 
0x251, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xae, 
0x62a, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xc21, 0xffff, 0xffff, 0xffff, 0xe75, 0xffff, 
0xffff, 0x10ba, 0xffff, 0xffff, 0xffff, 0x1227, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x94, 
0xffff, 0xffff, 0xd52, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0x3fd, 0xffff, 0x95c, 0xffff, 
0xffff, 0xaf0, 0xf96, 0x1296, 0xffff, 0xffff, 0x141, 0xffff, 
0xffff, 0xffff, 0xffff, 0xf30, 0xffff, 0x1db, 0x8a6, 0xffff, 
0xffff, 0xffff, 0xffff, 0x178, 0xffff, 0xffff, 0xffff, 0x8bd, 
0xe4d, 0xafc, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x51a, 
0x1173, 0x698, 0xffff, 0xffff, 0xffff, 0xffff, 0x329, 0xffff, 
0x1a7, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0x1111, 0xffff, 0xd3c, 0x1d4, 0xffff, 0xffff, 0x758, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xd39, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0x93c, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0x4f0, 0xffff, 0xffff, 0xffff, 0xa80, 
0xffff, 0xffff, 0xffff, 0xffff, 0x58f, 0x83c, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xbae, 0x717, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xc3a, 0xffff, 0xffff, 
0x550, 0xffff, 0x42d, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0x93a, 0xbc9, 0xffff, 0xffff, 0x12f5, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x933, 0xffff, 
0xffff, 0xffff, 0x9ba, 0x639, 0xffff, 0xffff, 0xffff, 0xffff, 
0x7ac, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x5fd, 0x6ba, 0xf5f, 
0xffff, 0x5a1, 0xffff, 0xffff, 0xffff, 0xffff, 0x3c6, 0xffff, 
0xba, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0x1101, 0xffff, 0xffff, 0x88a, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0x161, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0x3b4, 0xffff, 0xfc3, 0x126b, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xa9b, 0xffff, 0xffff, 0xffff, 
0xffff, 0xae9, 0xffff, 0x35e, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0x53a, 0x6f, 0xffff, 0x14e, 0xffff, 0xffff, 
0xffff, 0xffff, 0x791, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xcba, 0x1116, 0x681, 0x11cd, 0xffff, 0xffff, 0xffff, 
0xded, 0xff8, 0xffff, 0x6a0, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xb9d, 0x911, 0xffff, 0xffff, 0xc87, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x226, 0xffff, 0xffff, 
0x42e, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0x417, 0xffff, 0xffff, 0x863, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x4c0, 0xad9, 
0xffff, 0xa78, 0xffff, 0xffff, 0x4b1, 0xffff, 0xffff, 0x24e, 
0xffff, 0xffff, 0xffff, 0x90e, 0xde8, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x32d, 0x4d1, 0xc01, 
0xffff, 0xffff, 0xffff, 0xfc7, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0x111c, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xd0a, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xad, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0x1ef, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xf2d, 0x388, 0xffff, 0xffff, 0xffff, 0xaef, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xfc6, 0xca7, 0xffff, 0xffff, 
0xffff, 0xffff, 0xd40, 0xac8, 0x37f, 0xffff, 0xffff, 0xa6d, 
0xffff, 0xffff, 0xffff, 0x458, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xd5d, 0x228, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0x895, 0xffff, 0xffff, 0xffff, 0xffff, 0x5ba, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xa01, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xfaa, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0x372, 0xcbf, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0x35f, 0xffff, 0xffff, 0x8cf, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xf75, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0x11e4, 0x1cb, 0x30d, 0x4ff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0x1d0, 0xffff, 0xffff, 0xffff, 0x2e6, 0xffff, 0xffff, 
0xffff, 0xffff, 0x37b, 0xffff, 0xb29, 0xfd8, 0xffff, 0xffff, 
0x636, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xa43, 0xffff, 
0xffff, 0xffff, 0xe5a, 0x25d, 0x9d0, 0xffff, 0xffff, 0x16e, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x1096, 0xffff, 0xc1b, 
0x123c, 0xffff, 0xffff, 0x1c0, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0x154, 0x800, 0x13d, 0xffff, 0xffff, 
0xffff, 0xf7, 0xffff, 0xbcd, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x75c, 
0xffff, 0xffff, 0xffff, 0x1063, 0x8f4, 0x121b, 0xffff, 0x1303, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x11f4, 0xa12, 
0xffff, 0x5e1, 0xffff, 0x5f7, 0x10b9, 0xffff, 0xffff, 0xffff, 
0xffff, 0x406, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x861, 
0x81, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0x64c, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xda2, 0xffff, 0x116d, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xf0d, 0xffff, 0xffff, 0xffff, 
0xffff, 0x8c4, 0xcb1, 0xacf, 0xffff, 0xffff, 0xffff, 0xffff, 
0x113, 0xffff, 0x12f4, 0xffff, 0xffff, 0xf3a, 0xffff, 0x724, 
0x355, 0xffff, 0x591, 0xffff, 0xe7, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0x7bd, 0xffff, 0x7bb, 0xffff, 0xffff, 0xffff, 
0x4f8, 0x3e3, 0xb9, 0x10c2, 0x123f, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xc08, 0xffff, 0xffff, 
0xffff, 0x102a, 0xffff, 0xffff, 0xffff, 0xd53, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0x8a5, 0xffff, 0x10ee, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xb36, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xfdc, 0xffff, 0xffff, 0x9e0, 0xffff, 0xffff, 0x392, 0x746, 
0xffff, 0x1f4, 0xffff, 0x1150, 0xac0, 0xffff, 0xf99, 0x2b8, 
0xffff, 0xffff, 0xe95, 0xffff, 0xbbd, 0x792, 0xffff, 0xffff, 
0xffff, 0x960, 0xffff, 0xffff, 0xffff, 0x1119, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x784, 0xffff, 
0x41a, 0xae6, 0x1007, 0x575, 0x1120, 0xffff, 0xffff, 0x370, 
0x11f7, 0x120b, 0x4ad, 0x42f, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0x4df, 0x830, 0xffff, 0x106c, 0x261, 
0xffff, 0xa4c, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xadf, 0xffff, 0xffff, 0xc6b, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0x775, 0x56f, 0x340, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0x1b, 0x1090, 0xffff, 0x764, 0xffff, 
0xffff, 0xcd6, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0x55a, 0xffff, 0x270, 0xffff, 0xffff, 0x906, 
0x3f, 0xffff, 0xffff, 0xffff, 0x578, 0x779, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xee7, 0xffff, 0xffff, 0xa57, 0xffff, 
0xffff, 0xffff, 0x64e, 0xffff, 0xffff, 0x515, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x1071, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x6e2, 0x797, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x5b9, 0xa8c, 0xffff, 
0xb2f, 0x1067, 0x62c, 0x1018, 0xffff, 0xffff, 0x60a, 0xffff, 
0x179, 0x5f9, 0xf4b, 0x11d2, 0x1264, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xb03, 0x1ee, 0xffff, 0xffff, 
0x51d, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xa19, 0xffff, 0x202, 0xa28, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0x6d3, 0x685, 0x343, 0x552, 
0xd67, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xf7d, 
0xffff, 0xffff, 0xffff, 0xc2a, 0xffff, 0xffff, 0x8a0, 0x1c6, 
0xbbf, 0xffff, 0x51e, 0xffff, 0xffff, 0xffff, 0x1129, 0x2b, 
0xffff, 0xffff, 0xb94, 0xa3d, 0xc1d, 0xffff, 0x858, 0x122c, 
0xe9f, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xb00, 0xffff, 
0xffff, 0xffff, 0x8af, 0xb3a, 0xffff, 0xffff, 0x1070, 0xffff, 
0xffff, 0xffff, 0x5c5, 0xffff, 0xffff, 0xffff, 0x1292, 0xffff, 
0xffff, 0x4fa, 0xffff, 0xffff, 0xffff, 0xffff, 0x70b, 0xffff, 
0x789, 0x8a9, 0xffff, 0xffff, 0x4aa, 0x1290, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x369, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0x1027, 0xffff, 0xffff, 0x48b, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xbde, 0xffff, 0x128a, 0xffff, 0xffff, 0xa1c, 0xffff, 
0xffff, 0xffff, 0xffff, 0x4c8, 0x123e, 0xffff, 0x3dd, 0xffff, 
0xe8c, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0x2a3, 0x117e, 0xffff, 0xffff, 0xffff, 0xffff, 0x4e8, 
0xffff, 0xffff, 0xffff, 0x4f2, 0xffff, 0xffff, 0xa5c, 0x1cc, 
0xffff, 0xffff, 0x6d4, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0x649, 0xffff, 0xffff, 0x78a, 0x9d1, 0xdde, 0xffff, 0x665, 
0xffff, 0xaf, 0xffff, 0x11a5, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0x4c5, 0x322, 0x10f1, 0xffff, 0xffff, 0xffff, 
0x1e0, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0x34, 0xffff, 0xffff, 0xffff, 0xa74, 0xffff, 0xffff, 0xffff, 
0xffff, 0x709, 0xffff, 0x440, 0x987, 0xffff, 0xffff, 0xffff, 
0xffff, 0x106d, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xd6, 
0xc27, 0xfa3, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0x2c3, 0xffff, 0x1293, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xddb, 0xffff, 0xffff, 0x9f5, 0xde7, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0x87f, 0xffff, 0x118e, 0xffff, 0xffff, 
0xffff, 0xffff, 0x33b, 0xeac, 0x1194, 0xffff, 0xffff, 0x859, 
0xffff, 0x208, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xc2d, 0x4f7, 0x76c, 0x87a, 0x12fe, 
0xffff, 0xffff, 0xa9, 0xffff, 0xffff, 0xffff, 0xd85, 0xd33, 
0xffff, 0xffff, 0xd4b, 0xffff, 0xffff, 0xffff, 0x509, 0xffff, 
0x1162, 0xffff, 0x1e, 0xffff, 0xffff, 0xffff, 0xffff, 0xc72, 
0xffff, 0x4a1, 0xffff, 0xffff, 0xffff, 0x11bf, 0xffff, 0xaf8, 
0xffff, 0xffff, 0x258, 0xffff, 0x8c2, 0xffff, 0x3a1, 0xc2b, 
0xfcf, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xbfe, 
0x6f8, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xdd2, 0xffff, 
0xa5a, 0x1136, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xa65, 0xa4, 0xffff, 0xffff, 0xffff, 0x606, 0x110e, 0xffff, 
0x186, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x44, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0x24f, 0x1251, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x9d4, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xbdc, 0xffff, 0xffff, 
0x4cd, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x1139, 0xffff, 
0x994, 0xffff, 0x52a, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0x3f2, 0x49f, 0xe39, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xe44, 0xffff, 0xffff, 0xffff, 
0xffff, 0xcf1, 0x12af, 0xffff, 0xffff, 0xffff, 0x6c6, 0xffff, 
0xffff, 0x819, 0xffff, 0x239, 0x1196, 0x688, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0x66f, 0xffff, 0xffff, 0xffff, 0xffff, 
0xfb1, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xe8, 0xffff, 0xffff, 0xffff, 0x3a7, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xa3a, 0xffff, 0x113e, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xf80, 0x1006, 0x117a, 
0x93b, 0xf4f, 0xffff, 0xffff, 0xa52, 0xffff, 0xffff, 0xffff, 
0xffff, 0x880, 0xffff, 0xffff, 0xffff, 0xffff, 0x5d8, 0xffff, 
0xffff, 0xf9d, 0x12dc, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xf20, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xe1e, 0x105a, 0xffff, 0xffff, 0xffff, 0xffff, 0xef, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xcfc, 0x117f, 0xffff, 0x2ee, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xbb7, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0x422, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xb3e, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x862, 
0xffff, 0xffff, 0x114c, 0xffff, 0x110d, 0xb71, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0x36c, 0xe99, 0x113f, 0xffff, 0xd5a, 
0xffff, 0xffff, 0xdda, 0xffff, 0x2c0, 0x790, 0xcc4, 0x1022, 
0x12e0, 0xffff, 0xffff, 0x79f, 0xffff, 0xffff, 0x839, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xb26, 0xffff, 0xffff, 
0xffff, 0xffff, 0x60e, 0xa3, 0xffff, 0xffff, 0xffff, 0x210, 
0xea7, 0x116f, 0xffff, 0xffff, 0x5c6, 0xabf, 0xffc, 0x1267, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0x466, 0xffff, 0xffff, 0xcc6, 0xee8, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x73, 0xffff, 0xffff, 
0x147, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x1e8, 
0xffff, 0x1023, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xdb6, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0x11b4, 0xffff, 0x3bd, 0xffff, 0xffff, 0xffff, 0x6fb, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x783, 
0xffff, 0xffff, 0x11dd, 0xffff, 0x79b, 0xffff, 0xffff, 0xffff, 
0x2d5, 0xad5, 0xcbe, 0x10cf, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0x427, 0x10b3, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0x21d, 0xffff, 0x86a, 0xffff, 0xffff, 
0xe9, 0xbd0, 0x7c1, 0xffff, 0xffff, 0xffff, 0xffff, 0xab0, 
0xffff, 0x21c, 0x3f4, 0xffff, 0xffff, 0x4cb, 0x199, 0xffff, 
0x7eb, 0xf9, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0x122, 0xffff, 0x70d, 0xffff, 
0x7b1, 0xa64, 0xe79, 0xffff, 0x346, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0x10a3, 0xaa8, 0xffff, 0x10, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0x34e, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xb07, 0xffff, 0xffff, 
0xffff, 0xffff, 0x115, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x3b8, 0x521, 
0xffff, 0xe47, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0x2c5, 0xffff, 0xffff, 0xe7a, 0xffff, 0xffff, 0x391, 0xffff, 
0xffff, 0xffff, 0xffff, 0x706, 0xd16, 0x312, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xada, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x7da, 
0x944, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x58, 0xffff, 0xb25, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x9d, 0xffff, 0xffff, 
0xf70, 0x102f, 0x1c3, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x118, 0x3aa, 0x3a8, 
0xffff, 0xb53, 0xffff, 0x2c9, 0xffff, 0xb6d, 0xffff, 0x66e, 
0xffff, 0xa83, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0x412, 0xffff, 0x769, 0xffff, 0x397, 0x11e1, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x69, 0x90, 0xffff, 
0xffff, 0xffff, 0x103c, 0x117d, 0x8e6, 0xffff, 0xffff, 0x1306, 
0x5b1, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xdcd, 0xffff, 0x3cf, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0x971, 0xae2, 0xffff, 0x12f6, 0xa2f, 0x89e, 0x670, 0xffff, 
0xffff, 0xffff, 0x885, 0xffff, 0xffff, 0xffff, 0x1c, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x1225, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0x264, 0xffff, 0xb1f, 0xffff, 0xffff, 0x8b4, 
0xffff, 0xffff, 0xb35, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0x10f5, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x635, 0xc77, 0xffff, 
0xffff, 0x80c, 0xffff, 0xffff, 0xffff, 0xffff, 0x63a, 0xffff, 
0xa4f, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0x10cb, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xe66, 0xffff, 0xffff, 0xffff, 0xffff, 0x15c, 0xaba, 0xffff, 
0x123d, 0xffff, 0xffff, 0x759, 0x5d0, 0xffff, 0xffff, 0xffff, 
0xef1, 0x93f, 0x75, 0xf21, 0x1259, 0xffff, 0xaa7, 0xffff, 
0xffff, 0xabb, 0x288, 0x1029, 0x725, 0x1103, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0x2aa, 0x503, 0xffff, 0xffff, 0x7fd, 
0xffff, 0xffff, 0xa58, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xddc, 0x1ba, 0x1086, 
0x51c, 0x10d5, 0x10fb, 0xffff, 0xffff, 0xffff, 0xffff, 0xebf, 
0x25e, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0x1020, 0xffff, 0x65, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x304, 0xffff, 0xf1d, 
0x12c4, 0xffff, 0xffff, 0xffff, 0x672, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0x30e, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x98c, 0xffff, 
0xffff, 0xffff, 0x12b4, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xd6d, 0xf61, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xa06, 0xffff, 0xffff, 0x4fb, 0x908, 0x793, 0xf0f, 0xffff, 
0x1295, 0xffff, 0x3a0, 0xffff, 0x171, 0xffff, 0xffff, 0xffff, 
0x56a, 0xe85, 0x833, 0xffff, 0xccf, 0xec3, 0xffff, 0xffff, 
0xffff, 0xffff, 0xf77, 0xffff, 0x640, 0xffff, 0x306, 0x3ff, 
0xc99, 0xefc, 0x1059, 0xffff, 0x259, 0x1e3, 0x837, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x872, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xedf, 0x123, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xc06, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0x4c6, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xaf6, 
0xffff, 0xffff, 0xffff, 0xb51, 0xffff, 0xffff, 0xffff, 0x903, 
0xffff, 0xffff, 0x623, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0x88e, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xc57, 
0x96a, 0xffff, 0xffff, 0xffff, 0xffff, 0xb34, 0xf7b, 0xffff, 
0xaa1, 0xffff, 0xffff, 0xffff, 0x1147, 0xffff, 0xffff, 0xffff, 
0x678, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xf64, 0xffff, 0xffff, 0x128b, 0xffff, 0xffff, 0xffff, 0xe76, 
0x361, 0xffff, 0xffff, 0xffff, 0x7e5, 0xffff, 0xfe2, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x1263, 0xffff, 
0xffff, 0xffff, 0xffff, 0xe10, 0xffff, 0x11a3, 0xffff, 0xffff, 
0x51, 0xffff, 0x1114, 0xffff, 0xffff, 0x12c, 0xd09, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x1b2, 0xffff, 0xffff, 
0xffff, 0xffff, 0xd26, 0x497, 0xffff, 0xffff, 0xffff, 0xffff, 
0x1033, 0xffff, 0xffff, 0xffff, 0xffff, 0xc34, 0xffff, 0x321, 
0x1025, 0x526, 0xffff, 0x87e, 0xffff, 0xffff, 0x69b, 0x1297, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xfee, 0xffff, 0xffff, 0xffff, 0x11eb, 0xffff, 
0xffff, 0xfbf, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0x10e0, 0xffff, 0xffff, 0x20a, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0x1065, 0xffff, 0xffff, 0xffff, 0x118d, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x1223, 0xffff, 0xffff, 
0x1085, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x11dc, 0x827, 
0xa86, 0xd61, 0xffff, 0xffff, 0xb39, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x1298, 
0xffff, 0xffff, 0xffff, 0xdfb, 0xffff, 0xffff, 0xffff, 0xffff, 
0x9b9, 0xffff, 0xffff, 0xffff, 0x599, 0xffff, 0xffff, 0xffff, 
0xffff, 0x9ff, 0xffff, 0xffff, 0x91a, 0xffff, 0xffff, 0x874, 
0xffff, 0xffff, 0x879, 0xffff, 0x1287, 0xffff, 0x115f, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xecf, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x726, 0xffff, 
0xffff, 0x1075, 0xffff, 0xffff, 0xffff, 0xf5e, 0xc13, 0xffff, 
0xfc9, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x359, 
0xffff, 0xffff, 0xffff, 0xffff, 0x77b, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0x82e, 0x96, 0xad3, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0x1232, 0xf92, 0xffff, 0x1a4, 0x32, 0xffff, 
0xffff, 0xffff, 0xffff, 0x1261, 0xffff, 0x1056, 0xf56, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0x273, 0xffff, 0x8dc, 0xb33, 0xffff, 0xffff, 0xc49, 
0xffff, 0xffff, 0xffff, 0xa2, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0x12cb, 0xffff, 0xffff, 0xffff, 0xffff, 0x507, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x138, 0x3be, 0xffff, 
0xffff, 0xffff, 0xffff, 0x4b3, 0xc6a, 0xffff, 0xffff, 0x668, 
0xffff, 0xffff, 0xffff, 0xe69, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xfa8, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0x41e, 0xc47, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0x66c, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0x71, 0x3a2, 0xffff, 0x198, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0x12c5, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0x3bc, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xaa2, 0xffff, 0x1294, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x98e, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xbfd, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x386, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x663, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x9ab, 0xffff, 
0x7d6, 0xffff, 0x7ae, 0x12fc, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xb, 0x280, 0xb8b, 0xe4b, 
0x1036, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0x7aa, 0xc39, 0xffff, 0xffff, 0xffff, 0x402, 0x533, 0x622, 
0xffff, 0x3da, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xf31, 0xffff, 0x487, 0x10b2, 0xffff, 0xffff, 
0xc16, 0xffff, 0x5c8, 0xffff, 0xffff, 0xffff, 0xffff, 0x107f, 
0xd80, 0xffff, 0xffff, 0xffff, 0x8b3, 0x10e2, 0x2d2, 0xffff, 
0xffff, 0xffff, 0x6ea, 0xffff, 0xffff, 0xffff, 0x4e0, 0xffff, 
0xbb, 0xffff, 0xffff, 0x242, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0x7ad, 0xffff, 0xffff, 0xffff, 0xc02, 0xffff, 
0xffff, 0xffff, 0xffff, 0xd42, 0xffff, 0x959, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0x19c, 0xffff, 0x315, 0xffff, 
0x491, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0x447, 0x2ea, 0xffff, 0x2f0, 0xd9f, 0x91b, 0x1060, 
0xf6, 0x3af, 0xffff, 0x368, 0x11e8, 0x1257, 0xffff, 0x9d9, 
0xbd2, 0xbe2, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0x362, 0xffff, 0xffff, 0xffff, 0xffff, 0x7a8, 0xffff, 
0x98, 0xae3, 0x129e, 0xffff, 0xffff, 0xffff, 0xffff, 0x6f3, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x92d, 0x6cd, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xa22, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xf3c, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xcb2, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xfaf, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0x11a8, 0xffff, 0xffff, 0x54a, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0x1316, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xf04, 0xffff, 0xffff, 0xffff, 0xffff, 
0x116b, 0xd6a, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xa50, 
0x87b, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0x1207, 0x159, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0x40, 0xc42, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xcdc, 
0xffff, 0xffff, 0xffff, 0xffff, 0xe3, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0x349, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xfde, 0xffff, 0xa02, 0xab9, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x347, 0x46e, 0xab6, 
0xdef, 0xffff, 0xffff, 0xffff, 0xffff, 0x126a, 0xffff, 0xffff, 
0xa85, 0x765, 0xffff, 0xffff, 0xe09, 0xf0, 0xffff, 0xffff, 
0xffff, 0xffff, 0x5a9, 0xffff, 0x95d, 0xffff, 0xffff, 0x453, 
0x5e2, 0x10c4, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x395, 
0x8f3, 0xffff, 0xffff, 0xffff, 0xffff, 0x12bd, 0xffff, 0xffff, 
0x1011, 0xebe, 0xffff, 0xffff, 0x501, 0xffff, 0xffff, 0xc90, 
0x11ae, 0xffff, 0xffff, 0x602, 0xffff, 0xffff, 0xffff, 0x5b3, 
0x1087, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xeb4, 0x26c, 0xffff, 0xffff, 0xffff, 
0xffff, 0x7e0, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0x3e1, 0xffff, 0xffff, 0xffff, 
0xe7c, 0x1210, 0xffff, 0xffff, 0xd35, 0xffff, 0xffff, 0xcb4, 
0xffff, 0x130d, 0xffff, 0xffff, 0xe26, 0xffff, 0xffff, 0xffff, 
0x836, 0xffff, 0xffff, 0x4d0, 0xffff, 0xffff, 0xffff, 0xffff, 
0x848, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x4c, 
0xffff, 0x9c6, 0xffff, 0x168, 0xd18, 0x768, 0x1f3, 0xffff, 
0xffff, 0xf8c, 0xffff, 0xd50, 0x121d, 0xf3d, 0x1d1, 0x2c7, 
0xffff, 0x97f, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xee, 0xffff, 0xffff, 0xffff, 0x14b, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0x6d6, 0xffff, 0xffff, 0x60f, 0xc6, 0xffff, 
0xffff, 0x5d, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xe22, 0x384, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0x3d5, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0x17b, 0xffff, 0xffff, 0xffff, 0xffff, 
0x43, 0xffff, 0xffff, 0xffff, 0xffff, 0x76a, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0x7cd, 0xffff, 0xffff, 0xffff, 
0x1037, 0xffff, 0xffff, 0xffff, 0xa45, 0x344, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0x206, 0xffff, 0xdc5, 0xffff, 0xffff, 0xc, 0xfd2, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x5e8, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0x730, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x11c3, 0xffff, 
0xffff, 0xffff, 0xffff, 0x8e, 0xffff, 0xf5, 0x108b, 0x334, 
0xffff, 0xffff, 0xffff, 0x5a3, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0x227, 0xffff, 0xffff, 0xffff, 0xdd6, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xc7d, 0xfe8, 0x10dc, 0xdc4, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0x10ed, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xe01, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0x5a, 0xdce, 0xef4, 0xffff, 
0xffff, 0x1202, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xd43, 0xff4, 0xffff, 0xffff, 
0xffff, 0x7d1, 0xffff, 0xffff, 0xffff, 0x8fd, 0x1e6, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xf46, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x9d6, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x91f, 0xa4e, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x5bb, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0x8c5, 0x177, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xfa9, 
0x112f, 0x38e, 0x33, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0x451, 0xffff, 0xffff, 0x52f, 0x78b, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xd64, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xc95, 0x102d, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xd8d, 0xffff, 0x753, 0xffff, 0xffff, 0xffff, 
0x1133, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x1280, 0xffff, 
0xe2a, 0x1148, 0xffff, 0xffff, 0x9c1, 0xfc4, 0xffff, 0x38, 
0x1311, 0xffff, 0xffff, 0xffff, 0xffff, 0x723, 0x834, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0x1256, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0x10a9, 0xffff, 0xffff, 0xfba, 
0xffff, 0xb0b, 0x1a5, 0xffff, 0xffff, 0x10d, 0xffff, 0xffff, 
0xffff, 0x283, 0x65d, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xf11, 0xffff, 0xffff, 0xffff, 
0xcf9, 0x5c3, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xea8, 0xffff, 0xffff, 0x983, 
0x6a3, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0x99, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0x3f3, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0x324, 0xffff, 0xffff, 0xa09, 0xffff, 0x26f, 
0xffff, 0xffff, 0xffff, 0xffff, 0x1d9, 0x457, 0x485, 0x12c3, 
0xffff, 0xffff, 0xa9a, 0xffff, 0xffff, 0x1041, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0x10eb, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xab4, 0xffff, 0xffff, 0x12cf, 0xffff, 
0xffff, 0xcda, 0xffff, 0xffff, 0xffff, 0xd03, 0xffff, 0xffff, 
0x975, 0xc4d, 0xb59, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xb6f, 0x482, 0xffff, 0x40a, 0xffff, 0xffff, 
0x7a, 0xffff, 0x472, 0x4b0, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xd4e, 
0xffff, 0xffff, 0xffff, 0x92e, 0xffff, 0x881, 0xffff, 0xd1b, 
0xffff, 0xffff, 0xa2a, 0xffff, 0xffff, 0x78, 0x74d, 0x34b, 
0xae4, 0xffff, 0xffff, 0x740, 0xffff, 0xffff, 0x563, 0xffff, 
0xffff, 0xffff, 0xffff, 0x36b, 0xffff, 0xffff, 0xffff, 0xffff, 
0x1039, 0xffff, 0x9b1, 0xf58, 0xffff, 0xffff, 0x1073, 0xffff, 
0xffff, 0xffff, 0x18c, 0xc69, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0x1281, 0x1246, 0xffff, 0xffff, 0xffff, 0xffff, 
0x20c, 0x3ee, 0xffff, 0xffff, 0x686, 0xffff, 0xffff, 0xffff, 
0x83d, 0xa17, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0x77f, 0xffff, 0xffff, 0x2dd, 0xffff, 0xffff, 0xffff, 
0xffff, 0xb08, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x94a, 
0xffff, 0xffff, 0xffff, 0xffff, 0x345, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xcb8, 0xe50, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xa4a, 0x754, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0x2e5, 0xbe5, 0xffff, 0x13e, 0xffff, 0xffff, 0xffff, 
0xffff, 0x181, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0x5d7, 0x3ce, 0xffff, 0xffff, 0xffff, 0xffff, 0xf7c, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0x83b, 0xe87, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xd2f, 
0x4f3, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x1c8, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xe7e, 
0x10ac, 0xf7a, 0xffff, 0xffff, 0xffff, 0x107c, 0x127, 0xf13, 
0xffff, 0xffff, 0xffff, 0x6e1, 0xffff, 0x1233, 0xffff, 0xffff, 
0xffff, 0xffff, 0x9b6, 0xffff, 0xffff, 0x6e4, 0xffff, 0xe94, 
0xffff, 0xe4c, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xd0c, 
0xffff, 0xffff, 0xffff, 0xffff, 0x620, 0x8d3, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0x9bc, 0xffff, 0xffff, 0x10a7, 
0xffff, 0xffff, 0x520, 0xffff, 0xca9, 0xffff, 0xffff, 0x899, 
0xffff, 0xa96, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x1dd, 
0xffff, 0x600, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xe43, 0xffff, 0xffff, 0xd66, 0xfa1, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xab8, 0xffff, 0xffff, 0xffff, 0xffff, 
0x28f, 0xffff, 0xffff, 0xffff, 0x188, 0x187, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x4c4, 
0xffff, 0xffff, 0xffff, 0xd51, 0xffff, 0x255, 0xffff, 0xffff, 
0xffff, 0xffff, 0xed5, 0xffff, 0xffff, 0xffff, 0xffff, 0x1268, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0x736, 0xe38, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xec4, 0x4e5, 
0x55, 0x4a4, 0x54e, 0x483, 0x6c4, 0xce7, 0xffff, 0xffff, 
0xffff, 0xffff, 0xa1a, 0xffff, 0xe2f, 0xffff, 0x17f, 0xffff, 
0x424, 0x12da, 0xffff, 0x6e3, 0xffff, 0x10ce, 0xffff, 0x1253, 
0xffff, 0xf32, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0x3d8, 0xebc, 0xffff, 0xffff, 0x653, 0xb5e, 0xffff, 0xffff, 
0xffff, 0x2af, 0xffff, 0x870, 0xffff, 0x4a0, 0x104d, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x11d7, 
0xffff, 0x58a, 0xffff, 0x658, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0x90a, 0xffff, 0xffff, 0xffff, 0x44c, 0x281, 0xd14, 
0xffff, 0xffff, 0xffff, 0xffff, 0xd5c, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0x3b3, 0xffff, 0xffff, 0x68, 0xc05, 
0xffff, 0x10a8, 0xe6f, 0x59f, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0x12f2, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xb99, 0xffff, 0x1275, 0xffff, 0x821, 0xffff, 
0x471, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xd0e, 0xffff, 
0xbfc, 0xffff, 0xffff, 0xffff, 0xffff, 0x543, 0x1309, 0xa87, 
0xffff, 0xffff, 0x907, 0xffff, 0xfd3, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0x101, 0x120, 0xc0c, 0xffff, 0x51b, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xcd2, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0x103b, 0xffff, 0xffff, 0x11d, 0xffff, 
0xffff, 0x7f5, 0xffff, 0x771, 0xffff, 0xb56, 0xffff, 0x2a7, 
0x17d, 0xa13, 0xf24, 0xffff, 0x572, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0x2f5, 0x5f0, 0x751, 0x102c, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0x1127, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xd7d, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0x86b, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0x19d, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xb82, 0x98f, 0xffff, 0x8bc, 
0xf65, 0xffff, 0xffff, 0xba3, 0xffff, 0x49e, 0x57f, 0xa9c, 
0xffff, 0xffff, 0xffff, 0xffff, 0x927, 0x4b4, 0xffff, 0xffff, 
0xaad, 0xffff, 0xffff, 0xffff, 0xffff, 0x42b, 0x54f, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0x2ca, 0xffff, 0xffff, 0xffff, 0x119b, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xdb, 0x632, 0xffff, 0x9df, 0xffff, 0x77e, 0x11cf, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x1115, 0x8ed, 0xd8e, 
0x382, 0xffff, 0xffff, 0xffff, 0xffff, 0x540, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xd3, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xa8f, 0xffff, 0xffff, 0xffff, 0xffff, 0x9be, 
0x389, 0x116c, 0xffff, 0xffff, 0x80, 0x267, 0x8bb, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0x1a2, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x12e8, 0xffff, 
0x4c7, 0xffff, 0xffff, 0xffff, 0xffff, 0x463, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0x64, 0xac2, 0x7cb, 0xfdf, 0xffff, 0xffff, 
0xffff, 0xffff, 0xe62, 0xffff, 0xffff, 0xffff, 0xffff, 0x3c0, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0x5a6, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x2ac, 0x364, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x10a, 
0xffff, 0xd94, 0xffff, 0x223, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xc4b, 0xdee, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xd8b, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x7a4, 
0xffff, 0x787, 0xffff, 0x498, 0xffff, 0xffff, 0xb46, 0xffff, 
0xffff, 0xffff, 0xda9, 0x865, 0xfcb, 0x867, 0x1169, 0xffff, 
0xffff, 0xffff, 0x175, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x3ae, 0xffff, 0xffff, 
0xffff, 0xffff, 0x101b, 0xffff, 0x2ad, 0x130e, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0x59d, 0xc41, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0x3fc, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0x67d, 0xffff, 0xffff, 0xffff, 0x1214, 0xffff, 0xffff, 0xffff, 
0xd1, 0x1aa, 0xffff, 0xffff, 0x9eb, 0xffff, 0xffff, 0x10e6, 
0xffff, 0xffff, 0xffff, 0xffff, 0x127f, 0xffff, 0x8ca, 0x10da, 
0xce2, 0xffff, 0xffff, 0xe58, 0xffff, 0xffff, 0xffff, 0xffff, 
0x439, 0x72c, 0xffff, 0x1d7, 0xffff, 0xe93, 0xe88, 0xffff, 
0x9f2, 0xffff, 0xffff, 0xffff, 0xffff, 0x116a, 0xffff, 0xffff, 
0xffff, 0xffff, 0x480, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x7e3, 
0x12b5, 0xffff, 0xffff, 0xffff, 0xffff, 0x523, 0x46f, 0xffff, 
0xc68, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xd92, 
0xffff, 0xffff, 0xffff, 0xffff, 0x8c0, 0xffff, 0xffff, 0x5fb, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xb98, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x83, 0xbc, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x10a0, 0x93e, 0x5e6, 
0xffff, 0xffff, 0x1057, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x1255, 0xffff, 0x224, 
0x3f1, 0x4b7, 0x570, 0x35a, 0x5e0, 0xf63, 0x10d2, 0xffff, 
0xffff, 0x5ed, 0xffff, 0xffff, 0xffff, 0xf0e, 0xffff, 0xffff, 
0xffff, 0xbe, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xfac, 
0x325, 0xe19, 0xe21, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xa25, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xa37, 0xa1e, 0xffff, 0xfb, 0xffff, 
0xffff, 0xffff, 0xffff, 0x888, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xc97, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xbb9, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xf76, 0xffff, 0xffff, 0x809, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0x10ea, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xfc, 0x67e, 0x8f0, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0x9e3, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xdbe, 0xffff, 0xffff, 0xffff, 0xf23, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0x11ce, 0xffff, 0xffff, 0xffff, 0x1249, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0x29d, 0xffff, 0xffff, 0x3a4, 0xffff, 
0xffff, 0xffff, 0xbee, 0xffff, 0xffff, 0x852, 0xa6c, 0x10d7, 
0x1cd, 0xffff, 0xa8a, 0xffff, 0x11a7, 0xffff, 0x84d, 0xc45, 
0xffff, 0xffff, 0xffff, 0xffff, 0x23e, 0xffff, 0xffff, 0xffff, 
0xfab, 0xffff, 0x100d, 0xffff, 0x12a3, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xcd5, 0x1e1, 0xd60, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0x988, 0xffff, 0xa49, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x2cb, 0xffff, 0x294, 
0x564, 0x1052, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xd9b, 
0xffff, 0xffff, 0xffff, 0xffff, 0x22a, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xd82, 0xffff, 0xe1, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xe2b, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xd99, 0xffff, 0xffff, 0xffff, 
0xffff, 0x6, 0xffff, 0xffff, 0x924, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x123b, 0xffff, 0xffff, 
0x2ec, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xc3c, 
0xffff, 0x6a2, 0x129, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0x940, 0x617, 0x106, 0xffff, 0xffff, 0x8aa, 0xd5f, 
0x30b, 0xffff, 0xffff, 0xffff, 0x43f, 0xffff, 0xa33, 0x112b, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0x12de, 0xb58, 0xffff, 0xffff, 0x9d5, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x571, 0xdd3, 
0xffff, 0xffff, 0xffff, 0x96f, 0xffff, 0xffff, 0xb42, 0xffff, 
0xffff, 0xffff, 0x237, 0x407, 0xffff, 0xd8a, 0xffff, 0xffff, 
0xffff, 0xb77, 0xffff, 0xffff, 0x18d, 0xffff, 0x71d, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0x11f, 0x2f9, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xa0a, 0xffff, 
0xffff, 0xffff, 0xffff, 0x12be, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x4b8, 0xffff, 
0xffff, 0xffff, 0x82d, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xb0a, 0xffff, 0xffff, 0xffff, 0x8a2, 0x11ad, 0xffff, 0xffff, 
0xffff, 0x100, 0x710, 0xffff, 0xffff, 0xffff, 0xd00, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x12d7, 0xffff, 
0xffff, 0x203, 0xf7f, 0xffff, 0xffff, 0x555, 0xffff, 0xdfc, 
0x11d6, 0x5c0, 0x9a4, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0x87c, 0x849, 0xffff, 0xffff, 0xffff, 
0x7ea, 0xffff, 0xffff, 0xffff, 0xe2e, 0xffff, 0x732, 0xffff, 
0x8cb, 0xbda, 0xffff, 0x31c, 0xffff, 0xffff, 0x152, 0xffff, 
0x10df, 0xffff, 0x3b2, 0xffff, 0x5ef, 0x8d1, 0xffff, 0xffff, 
0xffff, 0xece, 0xdf5, 0x2e8, 0xffff, 0xffff, 0xe6e, 0x48c, 
0x9f6, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xd81, 0xd47, 0xcdb, 0x2d6, 0x38a, 0x2f, 0xbec, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xfb2, 0xffff, 0xffff, 0xffff, 0xffff, 0x2b9, 0xf88, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xbbc, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0x531, 0xffff, 0xffff, 0xffff, 
0xffff, 0x85e, 0x3eb, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0x11ab, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xdb3, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xfd7, 0xffff, 0x1208, 0xffff, 0xffff, 0xffff, 0xffff, 0xd2, 
0x319, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0x96c, 0x529, 0x9c0, 0x428, 0x10cc, 0x195, 0xffff, 
0xffff, 0xffff, 0x1066, 0xffff, 0xec9, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0x625, 0xffff, 0xffff, 0xffff, 0xffff, 
0xe61, 0x10b8, 0x707, 0xffff, 0xffff, 0x955, 0xffff, 0xffff, 
0x1fa, 0x6ff, 0xe04, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0x5c4, 0xffff, 0xc48, 0xaaa, 0x1308, 
0xffff, 0xffff, 0xb5f, 0xffff, 0xffff, 0xffff, 0x5af, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x8d7, 0xa29, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0x1094, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0x680, 0x97e, 0xf15, 0xffff, 0xbf2, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0x3e5, 0xffff, 0xeff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xe51, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0x8a4, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xf6c, 0xffff, 0x12f, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0x222, 0x376, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0x146, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0x4d2, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xbe0, 0x1093, 0x32b, 
0x3ac, 0x826, 0xffff, 0xffff, 0x3d4, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xfd9, 0xa24, 0x4c2, 0xffff, 0xffff, 
0xffff, 0xffff, 0x10ae, 0x10e4, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0x694, 0xb72, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xe9c, 0xffff, 0xcb7, 0x1b3, 0x684, 0x11a2, 0x785, 
0xa3b, 0xffff, 0xffff, 0xffff, 0xb68, 0x12bb, 0x8bf, 0xb55, 
0x45a, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x1d5, 
0xffff, 0xcb5, 0xf83, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0x86e, 0xffff, 0x5ee, 0xffff, 0xffff, 0xffff, 0xffff, 
0x1bd, 0x524, 0x9f, 0xffff, 0xffff, 0x8b7, 0xffff, 0xa10, 
0xbc8, 0x5d1, 0xffff, 0x86, 0xffff, 0xffff, 0xffff, 0x11d0, 
0xffff, 0xffff, 0xffff, 0xffff, 0x5db, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0x83a, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xd27, 0xe32, 0x4b6, 0xffff, 
0xffff, 0x6b7, 0xf3b, 0xffff, 0xffff, 0xffff, 0x32f, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x6c5, 0xffff, 
0xffff, 0xffff, 0x5bc, 0x401, 0x6f6, 0xed8, 0xffff, 0xffff, 
0xffff, 0x5a4, 0xffff, 0xffff, 0x5f1, 0xffff, 0xffff, 0x841, 
0x8b1, 0x11fa, 0xffff, 0xffff, 0xffff, 0xffff, 0x62e, 0x513, 
0xffff, 0x1168, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x87d, 
0xffff, 0x413, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0x1238, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xdeb, 0xffff, 0xffff, 0x8d, 0xffff, 0xffff, 0xffff, 
0xffff, 0xb67, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x699, 
0xffff, 0xffff, 0xca4, 0xd77, 0x8b6, 0xee1, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x86d, 0xffff, 
0xffff, 0x160, 0xffff, 0x18f, 0xffff, 0xc9f, 0x142, 0x816, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x47, 0x1046, 
0xffff, 0x5a0, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0x5, 0x2d3, 0xc63, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0x4d3, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xb6e, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0x19e, 0xffff, 0xffff, 0xffff, 0xffff, 0x143, 0x22e, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0x4ca, 0xa1b, 0xb41, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0x6c, 0xffff, 0xed, 0x42, 
0xffff, 0x73d, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xbc3, 0xe17, 0xb3d, 0xffff, 0xffff, 0x4a9, 
0x8c9, 0xd54, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x79e, 
0xffff, 0xffff, 0x8d0, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0x27b, 0x473, 0x95a, 0x970, 0x335, 
0xffff, 0xffff, 0xcf2, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xb89, 0x426, 0xffff, 0xffff, 0xffff, 0xf29, 0xfd6, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xad7, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0x11f1, 0x22, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0x341, 0xffff, 0xffff, 0xffff, 
0xffff, 0x6f0, 0xffff, 0x679, 0xffff, 0x7d2, 0xfcd, 0x10c1, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xdb5, 0x1e7, 
0xffff, 0x465, 0xffff, 0x925, 0x6cc, 0xffff, 0xffff, 0xeed, 
0x488, 0xffff, 0xa8d, 0x90f, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0x1bf, 0xffff, 0x674, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0x11a6, 0xffff, 0xffff, 0x12b1, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0x43a, 0x5df, 0xffff, 0xffff, 0xffff, 
0x8f5, 0xffff, 0x951, 0xffff, 0xffff, 0xffff, 0xceb, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0x92a, 0xffff, 0xffff, 0xcbd, 0x7e7, 0xe25, 0xffff, 
0xffff, 0x301, 0x650, 0xffff, 0xffff, 0xffff, 0xa1, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x3d, 0xffff, 0xffff, 
0xffff, 0x8ab, 0x106a, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xc9b, 0xf39, 0xffff, 0x1156, 0xffff, 0xc7a, 0xffff, 
0xe5d, 0xe8a, 0xffff, 0xffff, 0xffff, 0xffff, 0x292, 0xffff, 
0x1288, 0xffff, 0xffff, 0xe3d, 0xf5c, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0x445, 0xffff, 0x12b, 0xb8e, 0x592, 
0x2d4, 0xffff, 0xffff, 0xbc6, 0x11fe, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0x85c, 0xc6f, 0xffff, 0xffff, 
0x103, 0x106f, 0x11b8, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0x577, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x673, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xc7, 0xe59, 
0x6e7, 0xffff, 0xffff, 0xd4a, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0x3ec, 0xffff, 0x7c7, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xa8e, 0xffff, 0xd72, 0xabd, 0xffff, 0x67c, 0x366, 
0x443, 0x687, 0xbf0, 0xffff, 0xffff, 0xffff, 0xffff, 0x980, 
0xffff, 0xffff, 0xffff, 0x958, 0xffff, 0xf1c, 0xffff, 0xffff, 
0xffff, 0xffff, 0x124f, 0xa21, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0x218, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x12ae, 0xffff, 0xffff, 
0xffff, 0xffff, 0xfe7, 0xffff, 0xffff, 0xffff, 0x10b6, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0x495, 0x876, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0x714, 0xffff, 0xffff, 0xe5, 0xdf3, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0x27e, 0xffff, 0xbe1, 0xffff, 
0xffff, 0x65b, 0xffff, 0xffff, 0xffff, 0x303, 0xffff, 0xffff, 
0x12b3, 0xb64, 0xffff, 0xffff, 0xffff, 0x396, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0x3f9, 0xffff, 0x253, 0xffff, 0xffff, 
0xffff, 0x5cf, 0xffff, 0x109d, 0x8f9, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xfc2, 
0xffff, 0xffff, 0x7b2, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0x964, 0xffff, 0xffff, 0xf2, 0xffff, 0xffff, 
0xc79, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0x317, 0x82c, 0x950, 0xffff, 0xffff, 0xffff, 0x15f, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0x7a5, 0xffff, 0x59e, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x3d9, 0xffff, 0xffff, 
0xffff, 0xdfe, 0xffff, 0xffff, 0xffff, 0xe4a, 0x22c, 0xffff, 
0xffff, 0xffff, 0xffff, 0x46, 0x1df, 0xc36, 0xffff, 0xffff, 
0xffff, 0x5e4, 0x12b9, 0xffff, 0xffff, 0xde6, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0x5f5, 0x4b2, 0xffff, 0xffff, 
0xe77, 0xffff, 0xffff, 0xffff, 0x9b5, 0xffff, 0x45, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x213, 0x3de, 0x6ed, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0x9c2, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0x1184, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0x1266, 0xffff, 0x112a, 0xffff, 0xffff, 0x3c4, 0xffff, 
0xffff, 0xffff, 0x3f0, 0xffff, 0xffff, 0xffff, 0x11e2, 0xffff, 
0xffff, 0xffff, 0x133, 0x9b4, 0x115b, 0xffff, 0x7f0, 0xef6, 
0xffff, 0xffff, 0x818, 0xffff, 0xffff, 0xffff, 0x7, 0x594, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xf6b, 0xffff, 
0x7cf, 0xc0b, 0xffff, 0xffff, 0xffff, 0xffff, 0x456, 0xffff, 
0xffff, 0xce, 0xaf4, 0x3b6, 0xffff, 0xffff, 0xffff, 0xffff, 
0x716, 0xffff, 0xffff, 0xffff, 0xd1d, 0xfc1, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0x7d3, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0x750, 0xffff, 0xffff, 0xffff, 0xffff, 0x7df, 0xbac, 
0xe9b, 0xffff, 0xffff, 0x229, 0xa23, 0xeb6, 0x118c, 0xffff, 
0xffff, 0xffff, 0x58c, 0xffff, 0xffff, 0xffff, 0xffff, 0x61a, 
0xffff, 0xffff, 0xffff, 0xffff, 0xfe9, 0xd07, 0xffff, 0x36, 
0xffff, 0xffff, 0xea0, 0x85a, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xa2c, 0xffff, 0x4e7, 0x9e1, 0xffff, 0x420, 
0xffff, 0x972, 0xffff, 0xffff, 0xffff, 0xffff, 0x1142, 0x11bc, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xd3f, 0xffff, 
0xffff, 0x937, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0x421, 0xffff, 0xffff, 0x10fa, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xf71, 0xffff, 0xffff, 0xbd8, 0xbea, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0x8eb, 0xffff, 0xffff, 0xd24, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x323, 0xf9a, 
0x2fa, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x61e, 0xffff, 
0xffff, 0xffff, 0xc52, 0x107a, 0x10db, 0xffff, 0xffff, 0xffff, 
0xffff, 0xbce, 0x6b, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x11a, 0xffff, 
0xffff, 0xffff, 0x41, 0xffff, 0xffff, 0xffff, 0x1089, 0x1f0, 
0x49a, 0xffff, 0xd55, 0xffff, 0xffff, 0xffff, 0xffff, 0x8a7, 
0xfd1, 0x2e1, 0xffff, 0xffff, 0xffff, 0x117b, 0xffff, 0xffff, 
0x68f, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x917, 
0xffff, 0x29b, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xef3, 0x1014, 0xa4b, 0xed6, 0x1219, 0x272, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0x7bf, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0x10ad, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x7f8, 
0xffff, 0xd63, 0xffff, 0xffff, 0x1282, 0xffff, 0xb66, 0xffff, 
0x6aa, 0x813, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x10ab, 
0x77c, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x123a, 0xffff, 
0xffff, 0x8b, 0xdc8, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0x567, 0x12f0, 0xffff, 0x9c, 0xc11, 
0xffff, 0x857, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0x71b, 0x2e4, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x6db, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x1179, 0xffff, 
0x21e, 0xffff, 0xffff, 0xaea, 0xffff, 0xffff, 0xffff, 0xffff, 
0x18a, 0xffff, 0x6bf, 0xffff, 0x11de, 0xffff, 0xffff, 0xffff, 
0xffff, 0x7c3, 0x10b4, 0xade, 0x942, 0x566, 0xffff, 0xffff, 
0xffff, 0xffff, 0x153, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xf81, 0xffff, 0xffff, 0xffff, 0xffff, 
0x57b, 0x7c8, 0x5c2, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x675, 
0xffff, 0x506, 0xffff, 0xffff, 0xffff, 0x99b, 0xffff, 0xffff, 
0xffff, 0x12bc, 0xffff, 0x8fc, 0xffff, 0xffff, 0xffff, 0xffff, 
0x5d5, 0xc3e, 0x31f, 0x5b4, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xb0d, 0xffff, 0xffff, 0x824, 
0xaab, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x913, 
0xffff, 0xffff, 0xa6e, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xeab, 0x5a2, 0x9a9, 0x1289, 0xffff, 0x183, 0xffff, 0xffff, 
0xffff, 0xffff, 0xec6, 0x73c, 0x868, 0xab1, 0xffff, 0xffff, 
0xa16, 0xffff, 0xffff, 0xffff, 0xffff, 0x365, 0xa40, 0xab5, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x1121, 0xffff, 
0x90d, 0xffff, 0xffff, 0x2cf, 0xffff, 0xce0, 0xffff, 0xffff, 
0xffff, 0xffff, 0x5f4, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xdff, 0xffff, 0xffff, 0xffff, 0xbc7, 0xffff, 0xffff, 
0xffff, 0xffff, 0x12dd, 0xffff, 0x104e, 0xffff, 0xc88, 0xffff, 
0xffff, 0xffff, 0x81c, 0xffff, 0xffff, 0xffff, 0xffff, 0x55c, 
0xffff, 0x2c, 0x1b8, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0x60d, 0xffff, 0xffff, 0xffff, 0xcb, 0xffff, 0xffff, 
0xffff, 0x431, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x108d, 0xb60, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xeeb, 0xffff, 0x113d, 
0xffff, 0xffff, 0xac, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xf34, 0xffff, 0xffff, 0xffff, 0xf66, 0xcfd, 0xffff, 0x10fd, 
0xffff, 0x9e9, 0xffff, 0x85b, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0x50d, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0x37, 0xffff, 0xffff, 0xffff, 0xffff, 0x476, 0xd8c, 0x12f9, 
0xffff, 0xc86, 0x1300, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xa08, 0xffff, 0xffff, 0xffff, 0x894, 0xc0e, 0xffff, 0xffff, 
0xffff, 0xffff, 0x308, 0xffff, 0xffff, 0x12b2, 0x1f, 0xffff, 
0xffff, 0x75e, 0xd7b, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0x1078, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0x85f, 0xffff, 0xffff, 0x6a5, 0xffff, 0xffff, 0x3ea, 0x278, 
0x9f7, 0xffff, 0x416, 0xffff, 0xffff, 0xffff, 0xffff, 0x8e3, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0x10ca, 0xe90, 0xffff, 0xee3, 0xffff, 0xffff, 
0xffff, 0xffff, 0x30, 0xd78, 0xca1, 0xffff, 0x12c6, 0xffff, 
0xffff, 0xffff, 0xffff, 0xefd, 0xffff, 0x162, 0x96b, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x3cc, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xc12, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0x96e, 0x1b9, 0xffff, 0xdcb, 0x68e, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xa1f, 0xffff, 0xffff, 0xffff, 0xaff, 
0x713, 0xd36, 0x7bc, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xcf6, 0xe30, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0x69f, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0x2de, 0x94d, 0xffff, 0x357, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xbc2, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x28b, 0xe83, 0xffff, 
0xffff, 0xffff, 0xffff, 0xb5d, 0xf1a, 0x12, 0x220, 0x4bd, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x35, 0x829, 
0xbbb, 0xffff, 0xffff, 0xffff, 0xffff, 0x3e0, 0xffff, 0xba8, 
0xffff, 0xffff, 0xffff, 0xffff, 0x6d1, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0x117c, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0x10b7, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xfea, 0xffff, 0xffff, 0xffff, 0xf74, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x593, 0x843, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x1141, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0x12f3, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x3d6, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0x1080, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0x48e, 0xffff, 0xffff, 0x6c9, 
0xffff, 0x66a, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xac6, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x6d7, 0xc51, 0xffff, 
0xffff, 0xffff, 0x1278, 0x7e1, 0xb4b, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0x46d, 0xef7, 0xa39, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xcf7, 0xffff, 0xffff, 0x62f, 0x8e2, 0x40e, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x1083, 
0xffff, 0x110b, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0x8dd, 0xffff, 0xffff, 0x871, 0xffff, 
0xffff, 0xffff, 0x89f, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0x57a, 0xc28, 0xffff, 0xffff, 0xffff, 0xffff, 0xa69, 0xffff, 
0x6d, 0x992, 0xffff, 0xffff, 0x10b5, 0xffff, 0xffff, 0xffff, 
0xffff, 0x10bb, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0x554, 0xffff, 0xffff, 0x10f4, 0xffff, 0xffff, 0x1f9, 0x2d1, 
0xe65, 0xffff, 0xffff, 0xffff, 0xf2e, 0xef8, 0xffff, 0xffff, 
0x770, 0x1fc, 0x7fc, 0xffff, 0xffff, 0xffff, 0xb3, 0xffff, 
0x9a0, 0x3b7, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x484, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x114d, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0x1042, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0x9ec, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xb73, 0xffff, 
0xffff, 0xab7, 0xfe0, 0x214, 0xffff, 0xffff, 0xffff, 0xffff, 
0xb92, 0xffff, 0x7b8, 0xffff, 0x510, 0xffff, 0xffff, 0xffff, 
0xffff, 0xd38, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0x11f9, 0xffff, 0xba4, 0x103a, 0xffff, 0xffff, 0x8ad, 
0x127a, 0xffff, 0x1221, 0x968, 0x909, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0x5a7, 0xffff, 0xffff, 0xffff, 0x14, 
0xaf3, 0x1235, 0xffff, 0xffff, 0xffff, 0xffff, 0x734, 0xffff, 
0xffff, 0xffff, 0xafa, 0xffff, 0xeb5, 0x6ad, 0xffff, 0xffff, 
0xc84, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x9de, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xa95, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xb0e, 0xffff, 0xffff, 
0x59b, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0x519, 0xffff, 0xffff, 0xd05, 0xffff, 0xffff, 
0x1192, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xf9c, 
0x119, 0xffff, 0xbe6, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xcd0, 0xffff, 
0xffff, 0x39b, 0xffff, 0xffff, 0xffff, 0xffff, 0xc5a, 0xe03, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xcc0, 0xffff, 0xffff, 
0x72, 0xffff, 0xffff, 0xffff, 0xffff, 0xe0a, 0xffff, 0xffff, 
0xffff, 0x67, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x94c, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xd44, 0xffff, 0x63d, 
0xffff, 0xffff, 0xffff, 0x129a, 0x11fd, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0x193, 0xffff, 0xffff, 0xffff, 0x1279, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xb4d, 0xffff, 
0xffff, 0xffff, 0xffff, 0xbe4, 0xffff, 0x869, 0xeb7, 0xb96, 
0xffff, 0xffff, 0x1216, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x7e8, 0x914, 0xffff, 
0xffff, 0xffff, 0xffff, 0x854, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0x48d, 0x565, 0xffff, 0x46b, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0x15b, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xb84, 0xffff, 0x9bb, 
0x467, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0x9fb, 0xb5b, 0xffff, 0xffff, 0xffff, 0xb1b, 
0x4ea, 0xffff, 0xffff, 0x542, 0x10e5, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xf90, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xdc9, 0xa2e, 
0xffff, 0xffff, 0x88f, 0x991, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xea3, 0x113a, 0xffff, 0x9a5, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0x1134, 0xffff, 0xffff, 0xffff, 0xffff, 
0xcc, 0xffff, 0xffff, 0xffff, 0xec1, 0xffff, 0x12d, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xf3e, 0xffff, 
0xffff, 0xffff, 0xffff, 0xc3b, 0xd2e, 0xe1b, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x29c, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xb5c, 0x11f5, 0xffff, 0xffff, 0xffff, 
0x9e6, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x285, 
0xffff, 0xffff, 0xffff, 0xfb5, 0x905, 0x80e, 0xffff, 0xffff, 
0xffff, 0x336, 0xffff, 0x1312, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0x1170, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0x81d, 0xe00, 0xffff, 0x409, 0xcde, 0xffff, 
0xffff, 0xffff, 0xffff, 0xd2a, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0x82a, 0x10f9, 0xffff, 0xc23, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0x590, 0x1f5, 0x238, 0x2c2, 0xffff, 0x1038, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x30c, 0xffff, 0xffff, 
0xffff, 0xffff, 0x605, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0x898, 0xffff, 0x11d3, 0xffff, 0x1204, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x40c, 0xf8d, 0xfa6, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x11be, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xad6, 0xffff, 0xffff, 0xffff, 0xffff, 0xa63, 0xc20, 0xfdd, 
0xffff, 0xffff, 0x5f, 0xbbe, 0xffff, 0xffff, 0xd1f, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xc76, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xd73, 0xffff, 0xffff, 0xffff, 0xffff, 
0x92, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0x19b, 0xffff, 0xffff, 0xffff, 
0xe8b, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0x29, 0xffff, 0x8db, 0xffff, 
0xffff, 0xffff, 0x34f, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0x120e, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0x956, 0xffff, 0xffff, 0xffff, 
0xc4a, 0x11ba, 0x127e, 0x265, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0x84e, 0xa70, 0xffff, 0xd6b, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0x2b0, 0x387, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x122d, 0xffff, 0xffff, 
0xa35, 0xffff, 0xffff, 0xffff, 0x266, 0x741, 0x115e, 0xffff, 
0xe33, 0xffff, 0x1212, 0xffff, 0xfc0, 0x33a, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x11e6, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x1004, 0xffff, 
0xffff, 0xffff, 0xffff, 0xbed, 0xffff, 0xb3f, 0xffff, 0x50f, 
0xdd0, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0x496, 0xffff, 0x11da, 0xffff, 
0x500, 0xb04, 0x1276, 0x616, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0x459, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0x2a0, 0x64f, 0xffff, 0xffff, 0xffff, 0x33e, 
0xffff, 0xf60, 0xffff, 0xffff, 0xffff, 0xe7f, 0xffff, 0xffff, 
0xc56, 0xde4, 0xf86, 0x11ac, 0xffff, 0x967, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x3dc, 0x98d, 
0x4eb, 0x158, 0x71c, 0x10d0, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0x124c, 0xffff, 0xffff, 0xffff, 0xffff, 0x805, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0x63c, 0xc64, 0xabc, 0xe0d, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x12f7, 0xc73, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0x727, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xacb, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xcf, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xdd8, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0x3d1, 0xffff, 0xffff, 0xffff, 0xffff, 0xd3b, 
0xf50, 0xffff, 0xffff, 0xb91, 0xffff, 0xffff, 0xffff, 0x50c, 
0xf10, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0x3a6, 0xffff, 0xffff, 0xffff, 0xffff, 0xb37, 
0xffff, 0xffff, 0xffff, 0x4f4, 0x328, 0xffff, 0x514, 0xaf9, 
0xffff, 0xffff, 0x1125, 0xffff, 0xffff, 0xffff, 0xffff, 0xe63, 
0xffff, 0xffff, 0xffff, 0xffff, 0x1af, 0xffff, 0xd1a, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x700, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xa90, 0xffff, 0xffff, 0xffff, 
0x603, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0x1241, 0xffff, 0xffff, 0x400, 0x5d2, 0x13c, 
0xffff, 0xffff, 0x326, 0x39c, 0xffff, 0x4a, 0xc50, 0x1171, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0x194, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0x17, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x7d5, 
0xffff, 0xffff, 0x46a, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0x7cc, 0xffff, 0xffff, 0x144, 
0x4ec, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xf1b, 
0x1310, 0xffff, 0xffff, 0x7ec, 0xf02, 0xe57, 0x460, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x155, 0x129f, 0xf4a, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x3b9, 0x56d, 0xbab, 
0xffff, 0x8ff, 0xffff, 0xffff, 0xffff, 0x262, 0xffff, 0xb4f, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xec7, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xb19, 
0x38b, 0x932, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xce8, 0xffff, 0xfa5, 0x11b9, 0xffff, 0x90b, 0xd08, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xa5f, 0xba5, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xadd, 0xffff, 0xd88, 0x1273, 0x891, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0x807, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x63b, 0xffff, 
0xffff, 0xffff, 0xffff, 0xfb7, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0x9a7, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xc75, 
0xffff, 0xffff, 0x7ee, 0xffff, 0x92c, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xfca, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0x12e3, 0xffff, 0xffff, 0xffff, 0xffff, 0x101c, 0x1124, 
0x954, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0x269, 0xffff, 0x11f2, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xc38, 0x1091, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x5b, 0xffff, 
0xd13, 0xbfb, 0xffff, 0xffff, 0xffff, 0xffff, 0x705, 0xffff, 
0xffff, 0xffff, 0x842, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0x815, 0xffff, 0x42c, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0x69e, 0xffff, 0xffff, 0xb76, 
0x1182, 0xffff, 0xffff, 0xffff, 0x79, 0xffff, 0x3df, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0x104a, 0xfd4, 0xffff, 0xffff, 0xffff, 0xffff, 0xb32, 
0xffff, 0xffff, 0xffff, 0xffff, 0xc9a, 0xffff, 0xffff, 0x399, 
0x51f, 0xffff, 0xaa, 0xffff, 0xffff, 0xffff, 0xd87, 0xca0, 
0xffff, 0xffff, 0xffff, 0x39d, 0x1149, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xa92, 0xffff, 0x254, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xb50, 0x1044, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0x69a, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0x20f, 0xffff, 0xffff, 0xffff, 0x256, 0xffff, 0xdbb, 
0xc17, 0xffff, 0xffff, 0xffff, 0x371, 0x455, 0x108c, 0xffff, 
0x1b1, 0xffff, 0xffff, 0xea2, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0x518, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0x81b, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x4db, 0xd68, 
0x8d9, 0x618, 0xffff, 0xffff, 0x720, 0x1143, 0xffff, 0xffff, 
0xffff, 0x5e3, 0x97b, 0xffff, 0x9b0, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xf03, 0xffff, 0xffff, 0xffe, 0x1313, 
0x96d, 0x11fb, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x3cd, 0xffff, 
0xffff, 0xffff, 0x5a5, 0xffff, 0xffff, 0xffff, 0xffff, 0x691, 
0x9ad, 0xffff, 0xffff, 0xffff, 0x37c, 0xffff, 0xffff, 0xb6, 
0xffff, 0xffff, 0x299, 0xffff, 0x41d, 0x68c, 0xffff, 0x1305, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0x1001, 0xd9a, 0xffff, 0xccb, 0x11bd, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xb7, 0x263, 0x17a, 
0xd9d, 0xffff, 0xffff, 0xffff, 0xc4, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0x84b, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xfd, 0xb28, 
0x432, 0xb4, 0x1229, 0xffff, 0xcf8, 0xffff, 0xfed, 0x10c, 
0x89b, 0xffff, 0xffff, 0xffff, 0xffff, 0x2a6, 0xffff, 0xffff, 
0x607, 0xffff, 0xffff, 0x1283, 0xffff, 0xffff, 0x277, 0xffff, 
0xffff, 0xffff, 0xa03, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x32e, 0xffff, 0x74e, 
0xffff, 0xffff, 0x6dd, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0x2be, 0xbf7, 0xffff, 0x380, 0x13f, 0xffff, 0xffff, 
0xffff, 0xffff, 0xea9, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0x35d, 0x2f4, 0xffff, 0xffff, 0x2ce, 0x767, 0x3b, 0x469, 
0xe35, 0xffff, 0xffff, 0xffff, 0xffff, 0x5d9, 0xffff, 0xffff, 
0x5aa, 0xffff, 0xffff, 0xffff, 0xffff, 0x7f9, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xa9d, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0x11bb, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x801, 
0xffff, 0xffff, 0xffff, 0x946, 0xffff, 0x23, 0xffff, 0x873, 
0x12e7, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0x2e7, 0x588, 0x102, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0x30a, 0xaf1, 0xffff, 0xffff, 
0xffff, 0xffff, 0xb7f, 0xffff, 0xffff, 0x64b, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x7e9, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x9bd, 0xffff, 0xffff, 
0x43d, 0xace, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0x10ec, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0x494, 0xffff, 0xe86, 0xffff, 0xe9d, 0xffff, 0x1015, 0x4c3, 
0xffff, 0xc96, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x601, 
0x3b0, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xa14, 0x643, 0xffff, 0x7a7, 0xffff, 0x45f, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0x1e2, 0xffff, 0xffff, 0xffff, 0xffff, 0xc8c, 0x7f7, 0x1da, 
0xe64, 0xffff, 0xffff, 0xffff, 0xf98, 0xffff, 0xffff, 0x1117, 
0xffff, 0x828, 0xc26, 0xffff, 0xffff, 0xffff, 0xffff, 0x103f, 
0xd41, 0xffff, 0xffff, 0xffff, 0xf09, 0xffff, 0xffff, 0xffff, 
0x915, 0xffff, 0xffff, 0xffff, 0x12b8, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x4e6, 0xa9e, 0xffff, 
0xffff, 0x279, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0x65e, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0x2dc, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xfe6, 0xffff, 0x12a1, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x2a5, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0x310, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0x190, 0x57c, 0x80b, 0xffff, 0xdf9, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0x7fa, 0xb85, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0x94e, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x562, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xbb5, 0xffff, 0xecc, 0x1099, 0xffff, 0x580, 0xffff, 0x49, 
0x12b6, 0xffff, 0xffff, 0xffff, 0xffff, 0x352, 0xd15, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0x10ff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x9db, 
0x1a0, 0x652, 0xffff, 0xffff, 0xffff, 0x127c, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0x174, 0xffff, 0xffff, 0x53d, 0xffff, 
0xa31, 0xffff, 0xffff, 0xcdd, 0xffff, 0x9d8, 0x62, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0x1180, 0xffff, 0xffff, 0xffff, 0xb10, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xc71, 0xffff, 0x47b, 0x166, 0xffff, 0x693, 
0x275, 0xf5d, 0xffff, 0xffff, 0x390, 0xb0, 0x4ed, 0x56c, 
0xffff, 0x969, 0xffff, 0xc14, 0xffff, 0xffff, 0x2d9, 0xffff, 
0xffff, 0x212, 0x582, 0xf47, 0xffff, 0x110a, 0xffff, 0xffff, 
0xffff, 0x18e, 0xcd1, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xe41, 0xffff, 0xffff, 0xffff, 0x910, 0xffff, 
0xffff, 0xffff, 0xcce, 0xffff, 0xffff, 0xffff, 0x11e9, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0x1019, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x216, 
0xffff, 0x116, 0xffff, 0xffff, 0xffff, 0xffff, 0x73e, 0xffff, 
0xffff, 0x1286, 0xffff, 0xffff, 0xffff, 0x477, 0xffff, 0xae0, 
0xffff, 0xa0e, 0xca8, 0xffff, 0xffff, 0xffff, 0x67f, 0xffff, 
0xffff, 0xffff, 0xffff, 0xf0b, 0xffff, 0xe60, 0x49b, 0xffff, 
0xffff, 0xffff, 0x2e0, 0xc94, 0xffff, 0xffff, 0xffff, 0x12c1, 
0x7be, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0x729, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0x10d4, 0xffff, 0xffff, 0xffff, 0x478, 
0xffff, 0xffff, 0x26e, 0xffff, 0xffff, 0x728, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xf16, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x9a6, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xcd3, 0xffff, 
0x100b, 0xffff, 0xffff, 0xffff, 0x134, 0xffff, 0xd46, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0x5b5, 0x8ba, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xa04, 0x7a6, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0x2fc, 0xa94, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xdf4, 
0x44a, 0xffff, 0x3c3, 0xffff, 0xffff, 0xffff, 0x6f4, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0x290, 0x313, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xe24, 0xffff, 0x12ba, 0xffff, 0xffff, 0xffff, 0xd83, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xc8e, 0x6b5, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x11c1, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x4dc, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0x1174, 0x60b, 0xffff, 0x9ef, 0xffff, 0xffff, 0xffff, 0x1272, 
0xffff, 0xffff, 0x4f1, 0xffff, 0xffff, 0xffff, 0x1201, 0xa5, 
0x1058, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x949, 0xbad, 
0xe8e, 0xffff, 0x6eb, 0xffff, 0xffff, 0x358, 0xa81, 0x50a, 
0xa61, 0x704, 0x107e, 0xffff, 0xffff, 0xb12, 0xffff, 0xffff, 
0x812, 0xffff, 0xffff, 0x918, 0xe2c, 0x609, 0x350, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0x5ec, 0xffff, 0xffff, 0x353, 0xffff, 0xffff, 0xffff, 
0xffff, 0x91d, 0x9fd, 0xffff, 0xffff, 0xffff, 0xd4c, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0x1009, 0xffff, 0xffff, 0xffff, 
0x129c, 0xffff, 0x587, 0xffff, 0x6f5, 0xffff, 0xf94, 0x122f, 
0xffff, 0x3b1, 0xffff, 0xffff, 0xa46, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x11ef, 0xffff, 
0xffff, 0xd91, 0xffff, 0xffff, 0x3ba, 0xc67, 0x10d3, 0xffff, 
0x111, 0xffff, 0xffff, 0xffff, 0xffff, 0x12fd, 0xffff, 0xffff, 
0x93d, 0xffff, 0xffff, 0xffff, 0xffff, 0x772, 0xffff, 0xd86, 
0xede, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x121a, 0xffff, 
0xffff, 0xffff, 0xffff, 0xa7b, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0x7b7, 0xffff, 0xffff, 0xffff, 0xffff, 0x53, 0x7ba, 
0xffff, 0xffff, 0xb75, 0xbe8, 0xb1d, 0xffff, 0x5ca, 0xffff, 
0xffff, 0xffff, 0x34c, 0xffff, 0x112, 0x586, 0xffff, 0x97c, 
0xffff, 0xffff, 0xba9, 0xffff, 0xdc1, 0xffff, 0xffff, 0xf, 
0xffff, 0xffff, 0xffff, 0xffff, 0x137, 0x12a0, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xbaa, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xbf4, 
0xc9d, 0xffff, 0xffff, 0x1215, 0xffff, 0xffff, 0xdd5, 0xffff, 
0xffff, 0xdad, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0x104, 0xc5d, 0xffff, 0xffff, 0xffff, 0x6c7, 0xffff, 0xffff, 
0xffff, 0x10d8, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0x1154, 0xffff, 0xffff, 0xffff, 0xa26, 0xe13, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x47c, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0x8ae, 0xb30, 0xffff, 0xb1e, 
0xffff, 0x65a, 0x80d, 0x517, 0xa73, 0xebb, 0x1048, 0x561, 
0xffff, 0xffff, 0x375, 0xffff, 0xffff, 0xfa4, 0xffff, 0x17c, 
0xf05, 0xf2a, 0xffff, 0xffff, 0xdb7, 0xffff, 0xb2, 0xffff, 
0xffff, 0x1160, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0x18, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xd90, 0x10af, 
0xffff, 0xffff, 0x28d, 0x9cd, 0xffff, 0xffff, 0xffff, 0xffff, 
0x70c, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xeb9, 
0x119d, 0xffff, 0x1077, 0xffff, 0xffff, 0xa2b, 0xdcc, 0xffff, 
0xffff, 0xf33, 0xffff, 0xffff, 0x11af, 0x1fd, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x2f2, 
0xffff, 0xffff, 0x11d9, 0xffff, 0x441, 0xffff, 0xffff, 0xffff, 
0xee5, 0x778, 0xffff, 0xffff, 0xffff, 0x1302, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0x74b, 0xf67, 0xffff, 0xffff, 0xc2, 
0x8da, 0xffff, 0xffff, 0xdd1, 0x2, 0x832, 0xffff, 0x113b, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xda7, 0xffff, 0x27a, 
0xffff, 0xffff, 0x1242, 0xffff, 0x1247, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0x316, 0xffff, 0xffff, 0xffff, 0xffff, 
0x32a, 0x3ca, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xef0, 0xffff, 0xffff, 0xffff, 0xffff, 0x1062, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xb78, 0xee4, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x795, 0xffff, 0xffff, 
0xffff, 0x9ac, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0x8a3, 0xffff, 0xffff, 0xffff, 0xe36, 0xffff, 0xffff, 0xffff, 
0xffff, 0xaf5, 0xe68, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xf1e, 0xffff, 0xffff, 0xffff, 0xffff, 0xf37, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x128f, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x612, 0xb5a, 
0x8f, 0xffff, 0xffff, 0x225, 0x2eb, 0x776, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0x2df, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0x12d4, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x10f6, 0x31a, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xc2f, 0x6ee, 0xffff, 
0xffff, 0xffff, 0xbef, 0xed3, 0xffff, 0xa8b, 0xffff, 0xffff, 
0x56b, 0x1098, 0xffff, 0xffff, 0xffff, 0x9cb, 0x88b, 0xffff, 
0xffff, 0xffff, 0x135, 0xc5b, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x131, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x2d8, 0x8e1, 0xffff, 
0xffff, 0xffff, 0xffff, 0xac5, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0x4cf, 0xffff, 0xffff, 0xffff, 0xffff, 0xa75, 0xffff, 
0xffff, 0xffff, 0xffff, 0xb14, 0xa53, 0x102b, 0x721, 0xeb2, 
0xffff, 0x8e0, 0xba7, 0xffff, 0x8f2, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0x7b3, 0xffff, 0x10e9, 0xf6e, 0xffff, 0xffff, 
0xffff, 0xffff, 0xf84, 0x8fb, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x6f1, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0x846, 0xffff, 0xf36, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x756, 
0xffff, 0x2ff, 0xffff, 0xffff, 0xffff, 0x4af, 0xffff, 0xffff, 
0xcac, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xe46, 0xffff, 0xffff, 0xce6, 0xffff, 0xffff, 0xd17, 
0xffff, 0xffff, 0xffff, 0xd95, 0xffff, 0xffff, 0x415, 0x581, 
0xf91, 0xacc, 0xffff, 0x52, 0xffff, 0xffff, 0x551, 0x935, 
0xd9c, 0xffff, 0x1f7, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xf2f, 0xffff, 
0x29e, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0x1157, 0xffff, 0xffff, 0xffff, 0x10de, 0x11f0, 0xffff, 
0xffff, 0xffff, 0xffff, 0x2ae, 0xe37, 0xffff, 0xb02, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x12d9, 
0xdc, 0xffff, 0xffff, 0xffff, 0xffff, 0x4de, 0xbd1, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0x4ae, 0x1110, 0x93, 0xffff, 
0x1055, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x43e, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x1165, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xc7c, 0xfbb, 0x1092, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xe4f, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xe5b, 0xffff, 0xffff, 0x9e7, 
0xffff, 0xffff, 0x2a9, 0xf73, 0xffff, 0xffff, 0x33c, 0x930, 
0xffff, 0xffff, 0xd0f, 0x948, 0xf8a, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0x6ce, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0x73a, 0xffff, 0x6d9, 0x1317, 0xffff, 0xffff, 0x36d, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0x3b5, 0xffff, 0xffff, 0xffff, 0x121e, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0x627, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0x26, 0xffff, 0x12c2, 0xb4e, 0xffff, 0xffff, 0xffff, 
0xf79, 0xffff, 0xffff, 0x9f0, 0xffff, 0x418, 0xffff, 0xffff, 
0xffff, 0x6a9, 0xffff, 0xffff, 0xffff, 0xffff, 0x92b, 0xffff, 
0xffff, 0xffff, 0x11fc, 0x4c9, 0xffff, 0x633, 0x4a3, 0xffff, 
0xffff, 0xffff, 0xb3c, 0xffff, 0xffff, 0xffff, 0xffff, 0x651, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0x966, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0x10c6, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0x2f3, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xed7, 0xffff, 0xffff, 0x1010, 0x61, 0x1f6, 0x757, 
0xffff, 0xffff, 0xffff, 0xffff, 0x840, 0xaf2, 0x6bb, 0xffff, 
0xffff, 0xffff, 0xffff, 0xdf1, 0xffff, 0xffff, 0x5ff, 0xffff, 
0xffff, 0xffff, 0x54, 0xffff, 0xffff, 0x120d, 0xb81, 0xffff, 
0xffff, 0x108e, 0x11cc, 0xffff, 0xffff, 0x825, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0x549, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0x7f, 0x1315, 0xf55, 0xffff, 0xa8, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0x1069, 0xffff, 0xffff, 0xffff, 0xffff, 0xb2d, 0xffff, 0xffff, 
0x1cf, 0x6bd, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0x3ed, 0xd8f, 0xffff, 0x9a3, 
0xffff, 0x560, 0x1244, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xad4, 0xffff, 0xffff, 0xffff, 0x761, 0x855, 0xffff, 0xf8e, 
0x103e, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x9aa, 0xb01, 0xffff, 
0xffff, 0xffff, 0x111a, 0xffff, 0x866, 0xffff, 0xffff, 0x124e, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x1d6, 
0x660, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x637, 0xffff, 
0xffff, 0xffff, 0xffff, 0x52d, 0x45d, 0xffff, 0xffff, 0x204, 
0x6f9, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x1035, 0xffff, 
0xb9a, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x72b, 
0x1fb, 0x835, 0xffff, 0xffff, 0x860, 0xe31, 0xf95, 0xffff, 
0xffff, 0xffff, 0xffff, 0x410, 0xffff, 0x938, 0xffff, 0x5b8, 
0xa88, 0xffff, 0x548, 0xffff, 0xffff, 0xffff, 0xd37, 0xd3a, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0x189, 0xea1, 0xffff, 0xffff, 0x1d2, 
0xed1, 0xffff, 0xffff, 0xffff, 0xffff, 0x333, 0xffff, 0xffff, 
0x1222, 0x398, 0xffff, 0xffff, 0xb7e, 0xffff, 0x597, 0xffff, 
0xca6, 0x449, 0x655, 0x6a7, 0xffff, 0xf19, 0x11ff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x1140, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x1107, 0x1231, 0xffff, 
0xffff, 0xffff, 0xc8b, 0xffff, 0xffff, 0x448, 0x104c, 0x109e, 
0xffff, 0x36e, 0xad2, 0xffff, 0xffff, 0xffff, 0x9e2, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x5bf, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xafe, 0xffff, 0xffff, 
0xffff, 0xffff, 0x11a0, 0x22b, 0xffff, 0x7a3, 0xe3c, 0xffff, 
0x1284, 0xffff, 0xd7, 0xac7, 0xffff, 0x2c4, 0xffff, 0xffff, 
0xffff, 0xe3e, 0x125c, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xd5, 0x9fe, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0x36a, 0xffff, 0xffff, 0xffff, 
0x6f2, 0xa0f, 0xcc5, 0xffff, 0xffff, 0xffff, 0xdfa, 0xffff, 
0xffff, 0x1177, 0xffff, 0xffff, 0xe92, 0xffff, 0xffff, 0xf12, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xb43, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xe06, 0xffff, 0xffff, 
0xffff, 0xffff, 0x574, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0x2c8, 0xe96, 0xffff, 0xffff, 0xffff, 0xffff, 0x111f, 
0x10c3, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0x99d, 0x11d4, 0xffff, 0x3f5, 0xffff, 0xffff, 0xffff, 0x5f6, 
0x1167, 0x39, 0xaa5, 0x63e, 0xad1, 0x25c, 0x120a, 0xffff, 
0x9c4, 0xffff, 0x84a, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0x184, 0xffff, 0xffff, 0x1b0, 0xe02, 0xf6d, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xfc5, 0xffff, 0xffff, 0x5b7, 
0xffff, 0x9c3, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x40d, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x8ea, 0xcf3, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xde1, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0x71a, 0xffff, 0x1213, 0x845, 
0xffff, 0x12e1, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0x844, 0x1024, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xfc8, 0xffff, 0xffff, 0x61c, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0x113c, 0xffff, 0x78f, 0xffff, 
0xc40, 0xf14, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x71f, 
0xfbc, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0x10a4, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xa30, 0xffff, 0xffff, 0xffff, 0xffff, 
0xf3, 0x11ea, 0xffff, 0xffff, 0xffff, 0x619, 0x2b1, 0x129d, 
0xffff, 0xeec, 0x52e, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xf89, 0xffff, 0xffff, 0xffff, 0x1084, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x118b, 0x201, 
0xffff, 0xffff, 0x642, 0xffff, 0xffff, 0x15, 0xffff, 0x40f, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xcea, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0x3a9, 0x433, 0xffff, 0xffff, 0xc00, 0xffff, 0xffff, 0xffff, 
0x1191, 0xffff, 0x4a8, 0x50e, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xbb0, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xb7a, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0x1159, 0x853, 0xffff, 0xffff, 0xa62, 0xffff, 0x182, 0xffff, 
0xffff, 0x56e, 0xffff, 0x817, 0xffff, 0xffff, 0xffff, 0x11c9, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x10e, 0x752, 
0xffff, 0xffff, 0xa, 0xffff, 0xffff, 0x100f, 0x6dc, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x7b4, 0xffff, 0xffff, 
0x7b9, 0x615, 0xffff, 0xffff, 0x621, 0x6ec, 0xffff, 0x508, 
0x1109, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0x9a2, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x89d, 0xba0, 0xffff, 
0xffff, 0xffff, 0xffff, 0xac4, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xf5a, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0x7e2, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x9f4, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x12df, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x692, 0xffff, 
0xffff, 0x3bf, 0xd9e, 0xdcf, 0x781, 0xffff, 0x4b, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0x1105, 0xffff, 0xffff, 0xffff, 
0xffff, 0xf0c, 0xffff, 0xffff, 0x76b, 0xffff, 0xffff, 0xffff, 
0xe54, 0x1226, 0xffff, 0x7a0, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xe1a, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0x922, 0xf06, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0x921, 0xffff, 0xffff, 0xffff, 0xffff, 0xeb1, 0xffff, 0xf82, 
0xffff, 0xffff, 0xffff, 0xffff, 0xbb6, 0xffff, 0xffff, 0xffff, 
0xc31, 0xffff, 0xffff, 0x430, 0xb8, 0x1161, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0x65c, 0xc37, 0xffff, 0x207, 0x486, 0x4d9, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xdc3, 0x8ee, 0xd59, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x16, 0xffff, 0xffff, 
0x2b2, 0xffff, 0xda4, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x70, 0xffff, 
0xffff, 0xffff, 0x934, 0xf01, 0x2e, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0x314, 0x116e, 0x8e8, 0xb63, 0xaa0, 
0xffff, 0x119a, 0xde, 0xffff, 0xa60, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0x6da, 0xffff, 0xffff, 0xf08, 0x6ca, 
0xffff, 0xffff, 0xffff, 0x217, 0xffff, 0xffff, 0xffff, 0xc5e, 
0xd96, 0xffff, 0xffff, 0xc19, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0x31e, 0x5dd, 0xffff, 0xffff, 0xffff, 0xffff, 0x10c5, 0x9f9, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xfcc, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x94f, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xa07, 0xffff, 0x276, 
0xffff, 0xffff, 0x111e, 0x8a1, 0x1220, 0xffff, 0x10d9, 0xffff, 
0xffff, 0xffff, 0xffff, 0x657, 0xffff, 0xffff, 0x176, 0xda6, 
0xffff, 0xffff, 0xffff, 0xffff, 0xec8, 0x9ea, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0x379, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x234, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xe5f, 0x896, 0x502, 
0x61b, 0xe49, 0xffff, 0xffff, 0xe6a, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0x7d9, 0xffff, 0xc8d, 0xffff, 
0xffff, 0xffff, 0x196, 0xdb2, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x17e, 0xffff, 
0xdc6, 0xffff, 0xffff, 0x11e7, 0xffff, 0xffff, 0xb9c, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x1a8, 
0xf4d, 0x419, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0x5d3, 0xffff, 0xbf6, 0x296, 0xffff, 
0xecb, 0xffff, 0x2d0, 0xffff, 0xffff, 0x1b4, 0x6cb, 0xffff, 
0xffff, 0xffff, 0x43c, 0x3e7, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0x28, 0xffff, 0xffff, 0xffff, 0x31b, 
0xffff, 0x2bf, 0xffff, 0x1064, 0xffff, 0x5f8, 0x221, 0x127d, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0x4e4, 0x11ec, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0x47a, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x78d, 0xffff, 
0xffff, 0x1a, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0x11a4, 0xffff, 0x780, 0xffff, 0xffff, 0xffff, 0xd8, 
0x3d7, 0xffff, 0xffff, 0xffff, 0xffff, 0x2bb, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xc7b, 0xffff, 0xffff, 0xffff, 
0x4fe, 0x129b, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0x115d, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x1012, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0x354, 0xffff, 0xffff, 0xff1, 0x530, 
0x6ef, 0xffff, 0x11b5, 0xfe3, 0xffff, 0x4bf, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x614, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x4f, 0xd74, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xa76, 0xffff, 0x72e, 0x10bc, 0xffff, 
0xffff, 0xffff, 0x1003, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0x8d2, 0xffff, 0x119c, 0xffff, 0xffff, 0xffff, 0xbf3, 0xf28, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x119e, 0x656, 0xd9, 
0x5b0, 0xffff, 0xffff, 0xb9e, 0xf07, 0xffff, 0xffff, 0xffff, 
0x890, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x205, 0xffff, 
0xa34, 0xffff, 0xffff, 0x1190, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xde9, 0xffff, 0x10b1, 0xb52, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x55b, 0xffff, 0x309, 
0xcc8, 0x107b, 0x579, 0xffff, 0xcaa, 0xffff, 0xffff, 0xffff, 
0x4d8, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x13a, 0xffff, 
0x12a9, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0x246, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0x414, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xbd, 0xffff, 0xffff, 0xffff, 0x61f, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x718, 0xffff, 
0x10f, 0xffff, 0x84f, 0xffff, 0x8a, 0xffff, 0xffff, 0xffff, 
0x11b7, 0x1079, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0x2e3, 0x6e6, 0xffff, 0xffff, 0xb7c, 0xffff, 0x589, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x12d1, 
0xffff, 0xffff, 0xffff, 0xffff, 0x100c, 0xffff, 0x604, 0xffff, 
0x2d7, 0xffff, 0xffff, 0x8c7, 0xffff, 0xffff, 0xffff, 0xffff, 
0x74c, 0xe72, 0xffff, 0xffff, 0x10c8, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xd49, 0xffff, 0xffff, 0xffff, 0xffff, 0x742, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0x1188, 0xffff, 0xffff, 0xffff, 
0xffff, 0xba1, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0x1102, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0x7c2, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x10c0, 
0xffff, 0xffff, 0xffff, 0xffff, 0x9d3, 0xffff, 0xffff, 0x5f2, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xfa7, 0x990, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xb8d, 0x965, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xbca, 0xffff, 0xffff, 0xffff, 0xffff, 0x9e8, 0xffff, 0xffff, 
0xffff, 0xffff, 0xe08, 0xa41, 0xffff, 0x109b, 0xffff, 0xffff, 
0xffff, 0xffff, 0x9b8, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0x297, 0xcc3, 0xffff, 0x101e, 0xffff, 0xffff, 
0xffff, 0xffff, 0x595, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xedd, 0xffff, 
0xffff, 0x53b, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xe0b, 0xffff, 0xffff, 0xffff, 0xffff, 0x1dc, 0x104f, 0xffff, 
0x1f1, 0xffff, 0xffff, 0x7fe, 0xffff, 0xffff, 0xffff, 0xb65, 
0xffff, 0x9f8, 0x1032, 0x1291, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0x271, 0xdae, 0xd32, 0xffff, 0x22f, 0xfff, 0xffff, 
0xffff, 0xffff, 0x378, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xe05, 0xffff, 0xffff, 0xffff, 0x981, 0xb2a, 0xffff, 
0x25b, 0xffff, 0x10f3, 0xffff, 0x75d, 0x4ce, 0x83f, 0x6b1, 
0xd29, 0x2b3, 0xcdf, 0xfb0, 0xffff, 0xffff, 0xffff, 0x901, 
0xffff, 0xffff, 0x38d, 0xffff, 0xffff, 0xffff, 0x381, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x864, 
0x7dd, 0x10dd, 0xffff, 0xffff, 0x1285, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x148, 0xffff, 
0x110f, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0x140, 0xffff, 0xffd, 0xffff, 0xb8f, 0xffff, 0xffff, 
0xffff, 0xffff, 0xbe7, 0xffff, 0xcfb, 0xffff, 0xffff, 0xffff, 
0x745, 0x6a6, 0xcd9, 0x12ed, 0xffff, 0xffff, 0x4a7, 0xffff, 
0xffff, 0xffff, 0x105e, 0xffff, 0xffff, 0xffff, 0xffff, 0x902, 
0x3e9, 0x24, 0x31d, 0x446, 0xc89, 0xffff, 0xffff, 0x877, 
0xa59, 0xc93, 0xd65, 0x12d8, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0x1b5, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0x63, 0xf0a, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xc43, 0xffff, 0xffff, 0xffff, 0xffff, 
0x122b, 0xffff, 0xbd9, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xf52, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x12b7, 0xffff, 
0xffff, 0xffff, 0x42a, 0xcf5, 0xf8b, 0x11cb, 0xffff, 0xffff, 
0xffff, 0xffff, 0x28a, 0xffff, 0xffff, 0xffff, 0xffff, 0x7e, 
0x1fe, 0x4bb, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0x39e, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0x499, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xfd0, 
0x882, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0x1ae, 0x405, 0xad0, 0xeb3, 0xffff, 0xffff, 0x610, 0xffff, 
0x10d1, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xe98, 0xffff, 
0xffff, 0x250, 0x5ce, 0xffff, 0xffff, 0x1269, 0xffff, 0xffff, 
0xe0, 0xa32, 0x163, 0xce9, 0xffff, 0xa7a, 0xffff, 0xbe9, 
0xffff, 0xa11, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0x799, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x823, 
0xffff, 0xffff, 0x149, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0x75b, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x583, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x62b, 0xffff, 
0xb74, 0x802, 0xffff, 0xffff, 0xffff, 0xffff, 0xe40, 0xffff, 
0xffff, 0xf41, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x49d, 
0xffff, 0xffff, 0x8cc, 0x8de, 0xffff, 0xffff, 0xffff, 0xffff, 
0x76d, 0xffff, 0xffff, 0x689, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0x748, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0x10f8, 0xffff, 0xffff, 0xffff, 0x4be, 0x6b6, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xae7, 0x475, 0xfb3, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0x100a, 0xffff, 0xf4, 0xffff, 0xffff, 0xb49, 0xb2c, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xc6e, 0x83e, 0xffff, 0xffff, 0xffff, 0x945, 0x953, 
0xffff, 0xe28, 0xffff, 0xffff, 0xffff, 0xfad, 0xffff, 0xffff, 
0xffff, 0x1017, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xfb9, 0xffff, 0xffff, 0x6c1, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0x10f0, 0xffff, 0xffff, 0xffff, 0xffff, 0x6f7, 0xffff, 
0xffff, 0xffff, 0xffff, 0x892, 0xffff, 0x377, 0xbd7, 0x1151, 
0xcf0, 0xffff, 0x112d, 0xffff, 0x654, 0xffff, 0xdac, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x1028, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xaaf, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xa0d, 0xffff, 0xbc1, 0xffff, 0xffff, 
0xffff, 0xffff, 0x10aa, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0x102e, 0x342, 0x556, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xd34, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x32c, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xebd, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xc25, 0xffff, 0xffff, 0x167, 0x2ed, 0xbc4, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x11c4, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x811, 
0xffff, 0xb7d, 0x11c8, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0x1c5, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0x584, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0x3bb, 0x538, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x1144, 
0xffff, 0xffff, 0xffff, 0xffff, 0x5d4, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x40b, 
0x90c, 0xffff, 0xffff, 0x6a1, 0xffff, 0xffff, 0xffff, 0xadb, 
0xffff, 0xffff, 0xddd, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x320, 0x5cd, 0x8c3, 
0xffff, 0x7d, 0xffff, 0xf40, 0xffff, 0xffff, 0xa6a, 0xffff, 
0xffff, 0xffff, 0xffff, 0x3a, 0x702, 0xffff, 0xffff, 0x847, 
0x1236, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x4e, 
0xffff, 0x125a, 0xffff, 0xffff, 0x993, 0xb8c, 0x481, 0xffff, 
0xffff, 0xffff, 0x7dc, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x249, 0x743, 
0xffff, 0xffff, 0xffff, 0xd0d, 0xffff, 0xffff, 0xffff, 0xffff, 
0x28e, 0xdb1, 0xffff, 0xffff, 0xffff, 0xffff, 0x10cd, 0xffff, 
0x534, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x88, 0x385, 0x89c, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x7f4, 0x50, 
0xffff, 0xffff, 0xffff, 0x544, 0xffff, 0xffff, 0x157, 0xffff, 
0x121, 0xbc0, 0xffff, 0x22d, 0xffff, 0xffff, 0x719, 0xffff, 
0xffff, 0xffff, 0x35b, 0xffff, 0xffff, 0xc0f, 0x75f, 0x11b2, 
0xb8a, 0xffff, 0xffff, 0xffff, 0x4ac, 0xc35, 0xffff, 0xffff, 
0x2d, 0xbdf, 0x8f8, 0xb24, 0xa98, 0xffff, 0xffff, 0xffff, 
0x9b3, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xf38, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0x215, 0x58b, 0xced, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x1, 0xcad, 0xffff, 
0xffff, 0xc04, 0xffff, 0xffff, 0xffff, 0xffff, 0xa3f, 0xffff, 
0xa54, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x1c4, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x29f, 0xcca, 0x1250, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0x1053, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0x1138, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x2f1, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xf54, 0xffff, 0xffff, 0xffff, 
0xffff, 0x19a, 0xffff, 0xffff, 0xffff, 0x109a, 0xffff, 0xffff, 
0x919, 0xffff, 0xffff, 0x11c0, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xb27, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0x3f7, 0xffff, 0xffff, 0xffff, 
0xffff, 0x4d4, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0x21f, 0xffff, 0xffff, 0xffff, 0xffff, 0x8f6, 0x12ef, 0xffff, 
0xffff, 0xffff, 0xffff, 0xd04, 0xffff, 0xffff, 0xf48, 0xffff, 
0xcaf, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x7ab, 
0xc98, 0x20e, 0x7fb, 0xffff, 0xffff, 0xffff, 0x7ed, 0xffff, 
0x79a, 0xffff, 0xffff, 0xe6d, 0xffff, 0xffff, 0x1ce, 0x25, 
0xf18, 0xb87, 0x1072, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0x97d, 0xffff, 0xffff, 0xffff, 0x4dd, 
0xffff, 0x5f3, 0xffff, 0xffff, 0x1262, 0xffff, 0xffff, 0xffff, 
0x11d1, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xb61, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xbd3, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xe18, 0x1049, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0x541, 0xffff, 0x219, 0x2a4, 0xffff, 0x4d, 0x641, 
0xf4e, 0xffff, 0x69c, 0xffff, 0x4ee, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xb1c, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x200, 0xffff, 
0xffff, 0xffff, 0xffff, 0xff7, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0x117, 0x9ee, 0x11, 0x435, 0xffff, 0xffff, 0xffff, 
0xffff, 0x884, 0xffff, 0xffff, 0x84c, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xca3, 
0xffff, 0xb1, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x631, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0x9dc, 0xffff, 0x490, 0xeca, 0x1307, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xdf0, 0x274, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0x8c6, 0xffff, 0xffff, 0xffff, 0xa0, 0xffff, 
0x6a4, 0xffff, 0xffff, 0xffff, 0x3, 0xffff, 0xffff, 0x7c9, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0x1164, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x6b8, 0x8f1, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0x690, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xcee, 0xffff, 
0xffff, 0xffff, 0x3c1, 0x676, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xc74, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0x4d7, 0xf51, 0xffff, 0xb62, 
0x6bc, 0xffff, 0x3ab, 0xffff, 0xffff, 0x2b7, 0x9af, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0x1095, 0xffff, 0xb97, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0x803, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xdba, 0xffff, 0xffff, 0xffff, 0xfef, 0xffff, 0x82, 
0xf35, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xefb, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xe80, 0x356, 0x5b2, 0x230, 0xffff, 0xffff, 
0x54c, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0x241, 0xffff, 0x11f6, 0xffff, 0xffff, 0xffff, 0xffff, 0xf72, 
0xffff, 0x164, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xeba, 
0xffff, 0x16d, 0x3fb, 0xffff, 0x9cf, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xfbd, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x9f1, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x34a, 0xffff, 0xffff, 
0x535, 0xffff, 0xffff, 0xfa0, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0x708, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x9c5, 
0xffff, 0xffff, 0xffff, 0xdd9, 0xffff, 0xffff, 0x12e5, 0xffff, 
0xffff, 0x1b6, 0xf78, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0x1ab, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0x58e, 0xc1c, 0xdbc, 0xffff, 0xffff, 0xffff, 
0x11ee, 0xffff, 0xffff, 0xffff, 0x3e4, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0x11d8, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0x15d, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xa5e, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xa6f, 0xffff, 0xffff, 0xffff, 0x943, 0x5fe, 0xefa, 0xffff, 
0xffff, 0xffff, 0xffff, 0x236, 0xc54, 0xffff, 0xffff, 0xffff, 
0xffff, 0xde3, 0x23f, 0xe45, 0x9dd, 0xfe4, 0x1061, 0x3ad, 
0xba2, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0x4cc, 0xffff, 0xffff, 0xffff, 0xffff, 0x569, 0x806, 0x87, 
0xffff, 0xffff, 0x1270, 0xffff, 0xffff, 0x1126, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x77a, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0x295, 0xffff, 0xffff, 0x24d, 0x1197, 0xffff, 0xffff, 
0xffff, 0x1c9, 0x293, 0xffff, 0x71e, 0xffff, 0xffff, 0xffff, 
0xa7c, 0x6a8, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xd79, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x9e5, 0xeaf, 
0xffff, 0x2a8, 0xffff, 0xffff, 0xffff, 0xffff, 0x9fa, 0xffff, 
0xffff, 0xffff, 0xffff, 0x383, 0x10a5, 0x332, 0xffff, 0xffff, 
0x923, 0xbfa, 0xf44, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xa38, 0xffff, 0xffff, 0xc9c, 0xffff, 0xffff, 0x130b, 0x1b7, 
0x1021, 0xffff, 0x1047, 0xffff, 0x4f6, 0xffff, 0xffff, 0xffff, 
0xa47, 0xffff, 0xe2, 0x4ab, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0x4a6, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0x318, 0x54d, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0x962, 0xffff, 0x5a8, 0xffff, 0x10f7, 0xffff, 
0xfae, 0xffff, 0xffff, 0xffff, 0x13, 0xffff, 0xa72, 0x12a7, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x1203, 0xffff, 0xffff, 
0xffff, 0xffff, 0xc9e, 0xd89, 0xec0, 0x191, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x9f3, 0xffff, 0xffff, 
0xfe5, 0x5fc, 0x1045, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xc65, 0xffff, 0x95f, 0xffff, 0x76, 
0xd62, 0xffff, 0x5fa, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0x5ae, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x119f, 
0xdab, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x8e9, 
0xffff, 0xffff, 0x394, 0xc1e, 0x12e6, 0xffff, 0xe7b, 0xffff, 
0xffff, 0x999, 0xffff, 0x89a, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0x979, 0xffff, 0xffff, 0xcc9, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xeee, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xacd, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xdf6, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x24a, 
0xffff, 0x232, 0xffff, 0x128c, 0xffff, 0xffff, 0xffff, 0x598, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xc30, 0xffff, 0xffff, 0x527, 0x647, 0xffff, 0xffff, 0xffff, 
0x4d5, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xe16, 0xffff, 
0xffff, 0x108f, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x4e2, 
0xffff, 0xffff, 0x128d, 0xffff, 0xc70, 0xffff, 0x423, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xbf5, 0xffff, 
0xffff, 0x1030, 0x10a6, 0xffff, 0xffff, 0xffff, 0xffff, 0xf6a, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x73b, 0xffff, 0x644, 
0xffff, 0xffff, 0xba6, 0xffff, 0xd6f, 0xffff, 0xb6c, 0x45c, 
0xffff, 0x11aa, 0xedc, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xe0f, 0xffff, 0xffff, 0x38c, 0x470, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xb4c, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0x12eb, 0xf57, 0xffff, 0xffff, 0xffff, 0xec2, 0xffff, 0xffff, 
0x330, 0xffff, 0xffff, 0xb13, 0xe91, 0xffff, 0xffff, 0x1199, 
0x3c9, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x8e4, 0xffff, 
0xffff, 0xffff, 0x558, 0xffff, 0xffff, 0xffff, 0xffff, 0xed9, 
0xffff, 0xffff, 0x124d, 0xffff, 0xffff, 0xffff, 0xffff, 0x2f7, 
0xffff, 0xffff, 0x16a, 0xffff, 0x12ee, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0x1d3, 0xe6b, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0x4e9, 0xd22, 0xffff, 0xffff, 0x11e5, 
0xffff, 0xffff, 0xffff, 0xe71, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xbcb, 0xffff, 0xffff, 0xffff, 0xffff, 0xbb4, 0xffff, 
0x11a9, 0xffff, 0x585, 0xffff, 0xffff, 0xdc0, 0xffff, 0x3d0, 
0x712, 0xffff, 0x6ae, 0xffff, 0xffff, 0xffff, 0x11e3, 0x11c7, 
0x247, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xe27, 0xffff, 0xffff, 0xffff, 0xc8a, 0xffff, 0xffff, 
0xffff, 0x20b, 0xffff, 0x8b5, 0xffff, 0xffff, 0xffff, 0x763, 
0xb38, 0xffff, 0xffff, 0x298, 0xffff, 0x95b, 0xffff, 0xffff, 
0x5cc, 0xffff, 0x72d, 0xffff, 0xffff, 0xffff, 0x1108, 0xffff, 
0xc6c, 0xffff, 0xdb8, 0xffff, 0xd20, 0xffff, 0xffff, 0xffff, 
0x112e, 0xffff, 0xffff, 0x2c6, 0xffff, 0xffff, 0xffff, 0x1217, 
0xb23, 0xdf7, 0x11c, 0x1240, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0x260, 0xdaa, 0x1318, 0xda5, 0xffff, 
0x156, 0xe5c, 0xe70, 0xffff, 0xffff, 0xffff, 0xffff, 0x2cc, 
0xffff, 0xffff, 0xffff, 0xffff, 0x112c, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x1002, 0x106b, 
0xffff, 0xffff, 0x21a, 0x648, 0xffff, 0xffff, 0x2b5, 0xffff, 
0xffff, 0xa3c, 0xc10, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0x6d2, 0xdec, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x11ca, 0xcbc, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xd28, 0xffff, 0x12c0, 0xffff, 0x8be, 0xc60, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xdbd, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xae5, 0xffff, 0xffff, 
};

const unsigned dcm_attribute_from_keyword_len = 19560;
const uint16_t dcm_attribute_from_keyword_empty = 0xffff;
const uint16_t dcm_attribute_from_keyword_dict[19560] = {
0xf8e, 0xffff, 0xffff, 0x871, 0xd23, 0x5cd, 0xffff, 0xffff, 
0xffff, 0x114d, 0x1208, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0x1192, 0xffff, 0xe97, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xf33, 0xffff, 0x2f0, 
0xffff, 0xffff, 0xffff, 0xffff, 0xe8f, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xad5, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xe63, 0xffff, 0xb1c, 0xffff, 0xec8, 0xffff, 0xffff, 
0x611, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x7a4, 0x49d, 
0xffff, 0xffff, 0xffff, 0x1e, 0xb8a, 0xffff, 0xffff, 0xffff, 
0xf7e, 0xffff, 0xffff, 0xffff, 0xffff, 0x861, 0xffff, 0xffff, 
0xffff, 0xffff, 0x10fe, 0xffff, 0xffff, 0xffff, 0xc61, 0xffff, 
0xa45, 0xffff, 0xffff, 0xffff, 0x67, 0x5a5, 0xffff, 0xfb8, 
0xffff, 0x85f, 0x300, 0x74c, 0xffff, 0xffff, 0xab5, 0xffff, 
0xffff, 0xffff, 0xffff, 0x93, 0xffff, 0x3df, 0x12e8, 0xffff, 
0xffff, 0x266, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0x103d, 0xa12, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xeb9, 0xffff, 0xffff, 0x8d, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0x53c, 0xffff, 0xffff, 0xffff, 0xffff, 0x389, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x6c2, 
0xffff, 0xffff, 0x241, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xb9b, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0x617, 0xffff, 0xffff, 0x5e3, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0x786, 0x99, 0xffff, 0x11ae, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0x864, 0xfb7, 0x315, 0xffff, 0xffff, 
0xffff, 0xffff, 0xa47, 0x4f9, 0xffff, 0xffff, 0xdf, 0xffff, 
0xffff, 0xffff, 0xffff, 0xc66, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0x12ca, 0xffff, 0xffff, 0xffff, 0x547, 0xffff, 0xffff, 
0x53a, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x246, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0x95b, 0x8b8, 0xffff, 0xffff, 
0xffff, 0xffff, 0x4b4, 0x103b, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0x1c7, 0xffff, 0x682, 0x511, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0x7ec, 0xe25, 0xffff, 0xffff, 
0x127f, 0xffff, 0xffff, 0xffff, 0xffff, 0xf3d, 0xffff, 0xffff, 
0xffff, 0xae6, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x120, 
0x48c, 0x498, 0xffff, 0xffff, 0xffff, 0x32c, 0x9d9, 0xffff, 
0xffff, 0x655, 0xffff, 0x2fd, 0x1b5, 0xcc9, 0xffff, 0xffff, 
0xffff, 0x4cc, 0xb58, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0x67a, 0xffff, 0xffff, 0xe03, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0x91d, 0xffff, 0x186, 0x2ed, 
0xffff, 0xffff, 0x573, 0xffff, 0xadf, 0xffff, 0xffff, 0xffff, 
0x1172, 0x2c6, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x2f5, 0x1213, 
0xa07, 0x230, 0x1164, 0x12aa, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0x20f, 0xffff, 0xc98, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x626, 0xffff, 
0xffff, 0xffff, 0x1074, 0xffff, 0x1040, 0xffff, 0xffff, 0x8e4, 
0xffff, 0xffff, 0xffff, 0x9bc, 0x4ca, 0xffff, 0x3ca, 0xffff, 
0x863, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xb8d, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0x81, 0x9, 0xffff, 0xffff, 0xffff, 0xffff, 0x1262, 0xffff, 
0x11ad, 0xfc8, 0x1016, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0x9c4, 0xffff, 0xffff, 0xd3f, 0xffff, 0xd7b, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xe79, 0x1c4, 0x9a2, 
0xb44, 0xffff, 0x11fa, 0xffff, 0xc8d, 0xffff, 0x797, 0xaec, 
0xde7, 0x1024, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0x76d, 0xffff, 0xffff, 0xba, 0xe4e, 0xffff, 
0xffff, 0xb8e, 0xffff, 0x76b, 0xfba, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0x11ed, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x7be, 0xffff, 
0x118d, 0xffff, 0xffff, 0xffff, 0x8db, 0x4eb, 0xffff, 0x973, 
0xffff, 0xffff, 0xffff, 0xffff, 0x9ad, 0x1227, 0xffff, 0xffff, 
0xffff, 0xee, 0x4a5, 0x7a3, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0x492, 0x9ca, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0x1173, 0x1229, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0x215, 0xcd8, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xd80, 0x130b, 0xdbb, 0xffff, 0xffff, 0xd71, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x30a, 0x431, 
0x11fc, 0xffff, 0xffff, 0x11ee, 0xffff, 0x44b, 0xd4, 0x99d, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x48f, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0x558, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xabe, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x124c, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0x4ce, 0x6d5, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x1038, 0xffff, 
0xffff, 0xffff, 0x2b3, 0xffff, 0xffff, 0xffff, 0xffff, 0x142, 
0xffff, 0xd9a, 0xffff, 0x83a, 0xffff, 0xffff, 0xffff, 0x49f, 
0xcf3, 0x9d4, 0x4cd, 0xd45, 0xf7f, 0x1085, 0xffff, 0xffff, 
0xffff, 0xffff, 0xd6b, 0xdd8, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xdb, 0xb95, 0xc99, 0xffff, 0xffff, 0xffff, 
0x6f8, 0xffff, 0xffff, 0x219, 0xffff, 0xb7a, 0xffff, 0x12e, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x827, 0x8fe, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x125d, 0xffff, 0xffff, 
0xffff, 0xb53, 0xffff, 0x1152, 0xffff, 0xffff, 0xffff, 0x532, 
0xffff, 0xffff, 0xffff, 0x592, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0x875, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xfb5, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x11c7, 0x4a8, 0xffff, 
0x41c, 0xffff, 0xffff, 0xffff, 0x12fe, 0xffff, 0x47d, 0xac3, 
0xe0e, 0x11e1, 0xffff, 0xffff, 0xffff, 0xd87, 0xe86, 0xffff, 
0xc10, 0xffff, 0xffff, 0xffff, 0xc37, 0xffff, 0x716, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0x2df, 0xffff, 0xfa4, 0x580, 
0x10c5, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x66c, 0xffff, 
0x3c3, 0x8e2, 0xffff, 0x265, 0xffff, 0xeb3, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0x174, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0x1ff, 0x540, 0x8c3, 0xffff, 0x61b, 0x73, 0x91a, 
0x93e, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x48b, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0x1c1, 0x584, 0xaa9, 0xc4d, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x53b, 
0xffff, 0x488, 0xffff, 0x10e7, 0xffff, 0xffff, 0xffff, 0xffff, 
0x12a8, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0x11b8, 0x111c, 0x1142, 0x9ed, 0x3a6, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x7ee, 0xffff, 
0xffff, 0x12fa, 0x790, 0xa24, 0xae0, 0xffff, 0xffff, 0xbd, 
0xffff, 0xffff, 0x8ce, 0xe7d, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0x4ea, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x11f7, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0x10d, 0xffff, 0xffff, 0xffff, 
0x28e, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xbcf, 0xffff, 0x818, 0xffff, 0xffff, 0x609, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0x570, 0x12f8, 0xffff, 0xffff, 
0xffc, 0xffff, 0xb89, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0x1cd, 0xffff, 0xffff, 0x8ca, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0x302, 0xffff, 0xffff, 0x3b1, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xa8, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0x119f, 0xffff, 0xffff, 0x99f, 0x32b, 
0xadc, 0xbb0, 0xead, 0x7bf, 0xea9, 0x6b9, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xd9f, 0xffff, 0x67e, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x92, 
0x3cc, 0xb78, 0x12db, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0x1db, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xea3, 0xffff, 0xffff, 0xffff, 0x935, 0x18d, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xe30, 0xffff, 0xffff, 0xffff, 0x12a2, 0xffff, 0xffff, 
0xc1b, 0x1166, 0xffff, 0xb3a, 0xffff, 0xffff, 0x994, 0xebd, 
0xfef, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x518, 
0x21c, 0xffff, 0xffff, 0xffff, 0xffff, 0x1001, 0xffff, 0xffff, 
0xffff, 0x3af, 0x983, 0xeab, 0xffff, 0xffff, 0xa02, 0xffff, 
0xffff, 0xffff, 0x4c8, 0xc79, 0x86d, 0xffff, 0x79d, 0xffff, 
0xffff, 0xffff, 0x538, 0xc71, 0xffff, 0xffff, 0xd03, 0xffff, 
0x1f1, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0x239, 0xffff, 0xdf4, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0x1037, 0xef2, 0xffff, 0xbbf, 
0xffff, 0x200, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0x5f9, 0xffff, 0xffff, 0x320, 0x6ec, 0x8e9, 0xffff, 0xffff, 
0xffff, 0x345, 0xffff, 0xffff, 0xffff, 0x3f, 0xffff, 0x36b, 
0xffff, 0xc50, 0xffff, 0xb4, 0xffff, 0xffff, 0xffff, 0xbd1, 
0xe15, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xe6c, 0xffff, 
0xffff, 0xffff, 0x1189, 0xffff, 0x33f, 0x888, 0xbe0, 0xffff, 
0x11cd, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x62b, 
0x1158, 0xffff, 0xffff, 0xffff, 0xb5f, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x3ea, 0xffff, 0xffff, 
0xb01, 0xffff, 0xffff, 0xbee, 0x1230, 0xa90, 0xffff, 0x117, 
0xffff, 0xffff, 0x143, 0x65, 0xffff, 0xffff, 0xffff, 0xffff, 
0x4c, 0x792, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0x420, 0xffff, 0xffff, 0xffff, 
0x505, 0xffff, 0xffff, 0xffff, 0x680, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0x3a4, 0xffff, 0xffff, 0xffff, 0xe44, 0x11c6, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0x587, 0xffff, 0xffff, 0xfa3, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0x444, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x6a4, 
0xffff, 0xcef, 0xffff, 0xffff, 0xffff, 0xc4b, 0xffff, 0xffff, 
0xffff, 0xa03, 0xffff, 0x3b8, 0xffff, 0x10e, 0xffff, 0xffff, 
0xffff, 0x12b, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xacc, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xa1b, 0xffff, 
0xffff, 0xffff, 0xffff, 0xb21, 0xffff, 0xffff, 0xffff, 0xffff, 
0x1162, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x225, 
0xffff, 0xffff, 0xffff, 0xffff, 0x781, 0xffff, 0xffff, 0xffff, 
0x327, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xfdb, 0xb69, 0xa2, 0xffff, 0xffff, 
0xd6f, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x111a, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x1b6, 0x11a0, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0x28d, 0x834, 0xffff, 0xffff, 
0xffff, 0x3d4, 0xffff, 0xffff, 0x115c, 0xffff, 0xffff, 0x1c6, 
0xffff, 0xffff, 0x560, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xb82, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xd9, 0x495, 0xf7b, 
0xffff, 0xffff, 0xffff, 0xe96, 0x273, 0xffff, 0xffff, 0x99c, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xfb4, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xccc, 0x3ee, 0xffff, 
0xffff, 0xffff, 0xffff, 0x4f4, 0xeaa, 0xffff, 0xffff, 0x710, 
0xffff, 0xffff, 0xffff, 0x8b1, 0xffff, 0xffff, 0xffff, 0xd82, 
0xffff, 0xffff, 0xffff, 0x761, 0xffff, 0x644, 0xffff, 0xffff, 
0xffff, 0xffff, 0x3c0, 0xd6d, 0xed5, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xf4f, 0xffff, 0x936, 0xffff, 0xffff, 
0xffff, 0x1122, 0xffff, 0xffff, 0xffff, 0x1161, 0xa9f, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0x37b, 0x11f2, 0xffff, 0xffff, 
0xffff, 0xffff, 0x14c, 0xffff, 0xffff, 0xffe, 0xffff, 0xffff, 
0xffff, 0x10a2, 0xffff, 0x1cc, 0x8d4, 0xef6, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0x7bc, 0xc05, 0x110, 0xffff, 0xffff, 0xffff, 0x74e, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x130f, 0xffff, 0xffff, 
0xffff, 0x23d, 0xffff, 0x5ef, 0xffff, 0xffff, 0xda5, 0xffff, 
0xffff, 0xffff, 0x484, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0x10d0, 0xffff, 0x2c9, 0xffff, 
0xffff, 0x475, 0xffff, 0xffff, 0xbc3, 0xffff, 0xffff, 0xffff, 
0xffff, 0x6cb, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xe27, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xd9b, 0xffff, 0xffff, 0x9d5, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x2ff, 
0x1198, 0x7e, 0xffff, 0xffff, 0xffff, 0x11d4, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0x974, 0xe16, 0xffff, 0xf64, 
0xffff, 0xdf5, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0x90a, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xeb1, 0x34e, 0xffff, 0xffff, 0xffff, 0x891, 
0xffff, 0x9fd, 0xe1a, 0xe1f, 0xffff, 0xffff, 0xffff, 0x65b, 
0xffff, 0xffff, 0xffff, 0xffff, 0x32, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x9bf, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xd4d, 0x1107, 0x112e, 
0xffff, 0xffff, 0xe53, 0xe06, 0xffff, 0xffff, 0x548, 0xf0b, 
0xf9f, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x465, 0x84d, 0xffff, 
0xffff, 0xbc, 0x15b, 0xffff, 0x1fe, 0xc8e, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0x4b9, 0xffff, 0x4c9, 0xffff, 0xffff, 
0xad0, 0x1141, 0xffff, 0xffff, 0xdcf, 0xffff, 0x1006, 0xffff, 
0xffff, 0x123b, 0xa2d, 0xffff, 0x1086, 0xffff, 0xffff, 0xffff, 
0xffff, 0xa05, 0xaba, 0xffff, 0xb76, 0xffff, 0xffff, 0x893, 
0xffff, 0xffff, 0x1133, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xd43, 0xffff, 0xe07, 0xffff, 0x33d, 0x9ab, 0xffff, 
0xffff, 0x1140, 0xffff, 0xffff, 0x11e6, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0x11c2, 0x2ea, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0x127e, 0xffff, 0xffff, 0xffff, 0xffff, 0x4d2, 0xffff, 
0xffff, 0xffff, 0x10a, 0xffff, 0xffff, 0x1075, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0x11ac, 0xffff, 0x481, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0x4b8, 0x769, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0x2d5, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xb92, 0x23, 
0xffff, 0x684, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x21f, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x3e, 
0xaaa, 0x68c, 0xffff, 0x56d, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0x4f6, 0xfc4, 0x1048, 0xffff, 0xffff, 
0x565, 0xc22, 0xffff, 0xd5c, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0x3a, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xeb8, 0xffff, 0x744, 0x735, 
0xcd9, 0x698, 0xffff, 0x5c7, 0xffff, 0xffff, 0x52a, 0xffff, 
0xffff, 0x4e, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0x1071, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xf61, 0xffff, 0x93f, 0xffff, 0xffff, 0xffff, 0xffff, 
0x12d6, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xa85, 0xffff, 0xffff, 0x1174, 0xffff, 0xffff, 0xffff, 0xffff, 
0x3e7, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xa0a, 0xffff, 0xffff, 0xffff, 0xffff, 0xac5, 0xbce, 
0xffff, 0x41e, 0xffff, 0xffff, 0xffff, 0x49b, 0xffff, 0xffff, 
0x783, 0xffff, 0xd2c, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xf03, 0xffff, 0xffff, 0xffff, 0x791, 0x947, 
0x12bc, 0xffff, 0xffff, 0x27, 0xffff, 0xffff, 0xffff, 0x12c1, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xe70, 0xffff, 0xb3c, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xd48, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0x7a, 0xffff, 0x1136, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0x6bb, 0xffff, 0xffff, 0xffff, 0xd1, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0x10df, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xb6, 0xffff, 
0xffff, 0xffff, 0xffff, 0x7f4, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xc7d, 0xffff, 0xffff, 0xffff, 0xffff, 
0x652, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0x1223, 0xffff, 0xffff, 0xffff, 0x41, 0x709, 
0xc34, 0xffff, 0xffff, 0x954, 0xffff, 0x1a4, 0xffff, 0x80a, 
0xffff, 0xf3, 0xffff, 0xffff, 0xffff, 0xbd6, 0xffff, 0xffff, 
0xffff, 0xffff, 0x3fb, 0xffff, 0xffff, 0xffff, 0x2c0, 0xffff, 
0xffff, 0xffff, 0xffff, 0x259, 0xffff, 0xffff, 0xecd, 0xffff, 
0xffff, 0x97e, 0xe0a, 0xbd7, 0xffff, 0xffff, 0xffff, 0xffff, 
0x10de, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x78f, 0xffff, 
0xffff, 0xd63, 0xffff, 0xffff, 0xffff, 0x578, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0x17c, 0xffff, 0xffff, 0x103, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x1196, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x8ac, 0xffff, 0xffff, 
0xfbf, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0x87a, 0xffff, 0xffff, 0xffff, 0xffff, 
0x523, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0x96c, 0x11bf, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0x81e, 0xffff, 0xffff, 0x8d2, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0x33, 0xffff, 0xffff, 0xffff, 0xffff, 
0xf8b, 0xffff, 0x624, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0x71a, 0xffff, 0x121e, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0x8e6, 0xa19, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x890, 0xffff, 0xffff, 
0x8a3, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x773, 0xffff, 
0xffff, 0x20b, 0xffff, 0xffff, 0x132, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0x9dd, 0x3, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xc44, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x11b, 0x7f9, 0xffff, 
0xffff, 0x574, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xc7, 0x6d9, 0xa91, 0xffff, 0x87e, 0xffff, 
0xffff, 0xffff, 0xffff, 0x1f4, 0x20d, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0x1e0, 0xffff, 0xffff, 0x4d7, 0xffff, 
0xffff, 0x2f4, 0xe8c, 0x1135, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x387, 0xffff, 0xffff, 
0xffff, 0xe50, 0xffff, 0xffff, 0x6f0, 0xffff, 0xffff, 0x7c5, 
0xb4b, 0x751, 0xffff, 0xffff, 0xffff, 0x121, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xfe2, 0xffff, 0xffff, 0xffff, 0x71f, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x8fb, 
0x1159, 0xffff, 0xffff, 0xffff, 0xffff, 0x249, 0xffff, 0xffff, 
0xffff, 0x4a3, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0x1ca, 0xffff, 0x3e0, 0x110c, 0xffff, 0xffff, 
0x85d, 0xffff, 0xffff, 0x42a, 0xffff, 0xa5c, 0xffff, 0xb1e, 
0x12da, 0x23c, 0xffff, 0x110e, 0xffff, 0xffff, 0xffff, 0xffff, 
0x10c3, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xd2, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0x180, 0xa15, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0x603, 0xd00, 0x121b, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0x599, 0xffff, 0xffff, 0xf09, 0x2b, 0xffff, 0xffff, 
0xffff, 0xf5, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x1009, 0xffff, 
0xffff, 0x591, 0xffff, 0xffff, 0xffff, 0xcb7, 0xffff, 0xdf7, 
0xffff, 0xffff, 0xe35, 0xffff, 0xffff, 0xca7, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xda0, 0xffff, 
0x8a8, 0xffff, 0xffff, 0xffff, 0x1c3, 0x4d, 0x2cf, 0x4c5, 
0x122b, 0xffff, 0xffff, 0x66, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xc65, 0xffff, 0xffff, 0xd29, 
0xffff, 0x39b, 0xffff, 0xffff, 0x20, 0xffff, 0xffff, 0xffff, 
0x425, 0xffff, 0x1206, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0x701, 0x363, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0x5cf, 0xee8, 0xffff, 0xffff, 0xcba, 
0xffff, 0xffff, 0xffff, 0xffff, 0x30c, 0xffff, 0xffff, 0xffff, 
0x451, 0xffff, 0x51f, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x10c8, 
0xffff, 0x9b2, 0xffff, 0x1219, 0xffff, 0xffff, 0xffff, 0x43b, 
0xffff, 0xffff, 0xffff, 0xffff, 0xab1, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xc55, 
0xffff, 0xffff, 0xffff, 0x2f1, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xa28, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xda6, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xba0, 0x9f6, 0xa53, 0xe36, 0xffff, 0xffff, 0xffff, 0xffff, 
0x3d5, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x544, 
0x1003, 0x7b0, 0xffff, 0x837, 0x804, 0xffff, 0xffff, 0xffff, 
0x1272, 0xffff, 0xffff, 0xffff, 0x6de, 0x214, 0xffff, 0xe3a, 
0xffff, 0xffff, 0x7c4, 0xffff, 0xffff, 0x5d, 0xffff, 0xffff, 
0x1cb, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xa2e, 0xffff, 
0xffff, 0xffff, 0xffff, 0xe68, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xd77, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0x894, 0x112, 0xf77, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x18a, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x4f, 0xffff, 0xabf, 
0xaea, 0xffff, 0x3a7, 0xffff, 0xf0, 0x1178, 0xffff, 0x8aa, 
0xffff, 0xffff, 0x476, 0xffff, 0xffff, 0xffff, 0xffff, 0xeb0, 
0x1270, 0x115a, 0xffff, 0x4a9, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xf42, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0x1108, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0x956, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xbbb, 0xffff, 0x467, 0x114f, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xe5c, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0x4c7, 0xffff, 0x148, 0xffff, 0xffff, 
0x4b5, 0xd9e, 0xffff, 0xffff, 0x6fd, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0x436, 0xffff, 0xffff, 0xffff, 0xa2b, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0x106d, 0xe17, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xea1, 0xffff, 
0x377, 0x6cf, 0x8f1, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xd90, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xaf7, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xee5, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0x437, 0xffff, 0xffff, 0x1257, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xa8b, 0x11b5, 0xffff, 0xffff, 
0x12e3, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xafe, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x27e, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0x6fc, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0x1044, 0xffff, 0xffff, 0xd1a, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x177, 0x659, 0xffff, 
0xffff, 0x12f3, 0xffff, 0xffff, 0xcc2, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x10cb, 
0x3dd, 0xffff, 0xffff, 0xffff, 0xffff, 0x124a, 0xffff, 0xffff, 
0xffff, 0xffff, 0x391, 0xffff, 0xffff, 0x766, 0xffff, 0xffff, 
0xffff, 0xffff, 0x618, 0xffff, 0xffff, 0xffff, 0x109, 0xe0d, 
0x668, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xd74, 0xffff, 0xffff, 
0xffff, 0xffff, 0x36e, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xa54, 0xffff, 0x15, 0x61, 0x32a, 0x1042, 
0x667, 0xffff, 0xffff, 0xffff, 0xffff, 0x849, 0xffff, 0xffff, 
0xffff, 0x844, 0x10f3, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0x941, 0xb7f, 0xffff, 0x586, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0x12cd, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0x47e, 0xf53, 0xffff, 0xffff, 0xfc0, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xc96, 0xf89, 0xffff, 
0xffff, 0xffff, 0x5fd, 0xffff, 0x128f, 0xb45, 0xffff, 0xffff, 
0xeef, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xa9, 0xd59, 0xffff, 0xffff, 0xffff, 
0xfdf, 0xffff, 0x1026, 0xf10, 0xffff, 0xbcc, 0xffff, 0x2a, 
0xffff, 0xffff, 0x6cd, 0xffff, 0xffff, 0x86e, 0xffff, 0xffff, 
0xc16, 0xb81, 0xffff, 0xffff, 0x16, 0xffff, 0xffff, 0xffff, 
0xffff, 0x60a, 0xffff, 0xffff, 0xffff, 0x694, 0xfcf, 0xffff, 
0xffff, 0x1255, 0xffff, 0x93d, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0x477, 0xffff, 0xc69, 0x3db, 0xe84, 
0xd18, 0xffff, 0xffff, 0xffff, 0xe7f, 0xffff, 0xffff, 0xffff, 
0x7c, 0x778, 0x825, 0xffff, 0xffff, 0x1073, 0xffff, 0xffff, 
0x429, 0xffff, 0xffff, 0xffff, 0xffff, 0xd5d, 0xffff, 0xffff, 
0x97b, 0xffff, 0x398, 0x6ac, 0x29c, 0x114b, 0x2f, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0x79c, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x411, 0x10d4, 0x10eb, 
0xffff, 0xffff, 0x126, 0x326, 0xe62, 0x12b3, 0xffff, 0x8b, 
0xffff, 0xcf, 0xf4e, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0x120d, 0xffff, 0x362, 0xffff, 0xffff, 0x561, 0xffff, 0x11e5, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xc1a, 0xffff, 0xffff, 
0xffff, 0x117b, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xbc5, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x6f4, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0x102f, 0xffff, 0xffff, 0xa8e, 0xffff, 
0xffff, 0x189, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0x4e3, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x161, 0x102c, 0xffff, 
0x92e, 0x12f7, 0xffff, 0xffff, 0xffff, 0xffff, 0x1b9, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xc07, 0xffff, 
0xffff, 0x990, 0xffff, 0xa93, 0xc2a, 0xe2, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xdfb, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xb16, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0x4a6, 0x58b, 0x780, 0x213, 
0x2af, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0x2b8, 0xed1, 0x11c1, 0x296, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0x813, 0x56a, 0xffff, 0xffff, 
0x21e, 0xffff, 0xb7b, 0x1276, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xac9, 0xd02, 0xffff, 0xffff, 0xe6, 0x190, 0xb90, 
0xffff, 0xffff, 0xffff, 0x51a, 0x22c, 0xffff, 0xced, 0xffff, 
0xa0f, 0xffff, 0xf2, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xb43, 0xffff, 0x789, 0xffff, 
0x392, 0xffff, 0x17d, 0xffff, 0x6bd, 0x992, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x6f3, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xebf, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0x9a9, 0xb1f, 0x54d, 0xffff, 0xffff, 0xffff, 
0x848, 0xffff, 0xffff, 0xffff, 0x118a, 0xffff, 0x722, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0x1302, 0x12b5, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0x487, 0x852, 0xf57, 0xffff, 0xffff, 
0xffff, 0xffff, 0xf31, 0xffff, 0xffff, 0xffff, 0xffff, 0x1de, 
0xffff, 0xffff, 0x122c, 0x898, 0xffff, 0xffff, 0x648, 0x7e9, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0x12e2, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xfe1, 0xffff, 0xffff, 0xffff, 0x4fd, 0xffff, 0xffff, 0x216, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x901, 0x1a9, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xa01, 0xffff, 0xffff, 0x6d1, 0xffff, 0xffff, 0xffff, 0x1191, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xa67, 
0xffff, 0xffff, 0xffff, 0x294, 0xffff, 0x379, 0x66f, 0xffff, 
0xffff, 0xffff, 0xffff, 0xec, 0x1096, 0xffff, 0xffff, 0xffff, 
0xffff, 0xaff, 0xbb6, 0xffff, 0xffff, 0x8ed, 0xffff, 0xffff, 
0xffff, 0x106b, 0x11a7, 0xffff, 0xffff, 0xb85, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0x5c6, 0xffff, 0xffff, 0x12bb, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x666, 0xffff, 
0x12f, 0xa7e, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x1080, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xbd5, 0xffff, 0xffff, 0xffff, 0xf0a, 0x8f9, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0x9da, 0xffff, 0x1273, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0x38a, 0x419, 0xffff, 0xffff, 0xffff, 
0xffff, 0x242, 0xffff, 0x85c, 0xffff, 0xffff, 0xffff, 0xffff, 
0x2c1, 0x228, 0x494, 0xc93, 0xe56, 0xffff, 0xffff, 0xffff, 
0xb07, 0xb51, 0xffff, 0x12e9, 0xffff, 0xffff, 0xffff, 0xeac, 
0xffff, 0xffff, 0x9fc, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0x8c1, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0x4b1, 0xffff, 0xffff, 0xffff, 0x114e, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xf32, 0x66a, 0xfad, 0xfe9, 0xffff, 0x5a7, 
0xe91, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0x2c8, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0x217, 0xffff, 0x1023, 0xffff, 0x116f, 0x1081, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xca1, 0x11bb, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xe29, 
0xffff, 0xffff, 0xe74, 0xf6d, 0xffff, 0xffff, 0xffff, 0xffff, 
0xdbf, 0xed2, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xe11, 
0xffff, 0xffff, 0xffff, 0xb8, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xe3d, 0xffff, 0xffff, 0x45b, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xd68, 0xffff, 
0xffff, 0xffff, 0xffff, 0x178, 0x80e, 0xc6f, 0xf0d, 0xffff, 
0xffff, 0xffff, 0x433, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0x48d, 0xffff, 0xffff, 0xffff, 0x18e, 
0xffff, 0xffff, 0xffff, 0xffa, 0x1056, 0xffff, 0xffff, 0x8a6, 
0xffff, 0x7b2, 0xa32, 0xffff, 0xffff, 0x8d0, 0x129c, 0xffff, 
0xffff, 0x3d6, 0xffff, 0xffff, 0xffff, 0xffff, 0x372, 0xffff, 
0xffff, 0xffff, 0xffff, 0x9aa, 0xffff, 0xffff, 0x2b0, 0x8c5, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0x557, 0xffff, 0xffff, 0xefb, 0xffff, 0xffff, 0xffff, 
0xffff, 0x9b9, 0xffff, 0xffff, 0xffff, 0x1f5, 0xffff, 0xffff, 
0x23f, 0x7e3, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xcb2, 0xffff, 0xffff, 0x747, 0x8f2, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0x6b0, 0xffff, 0xffff, 0xffff, 
0xa1f, 0xffff, 0x963, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0x58e, 0xffff, 0xffff, 0x355, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0x27a, 0xc2f, 0xffff, 0xffff, 0xdcb, 0x579, 
0xffff, 0xffff, 0x107d, 0xffff, 0xffff, 0xffff, 0xffff, 0x128c, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x11c5, 0x2ae, 
0x5d5, 0xffff, 0xffff, 0x1311, 0xffff, 0xffff, 0x330, 0xffff, 
0x908, 0x124, 0xd0a, 0xffff, 0xffff, 0xffff, 0x455, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x2cc, 0x1a5, 0xffff, 
0xffff, 0x2d2, 0xffff, 0xffff, 0xffff, 0x5e4, 0xf1e, 0xffff, 
0xffff, 0xffff, 0xffff, 0x5d7, 0xb97, 0xffff, 0xffff, 0xffff, 
0xffff, 0xe98, 0x3bc, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xc3e, 0xffff, 0x317, 0xe85, 0x707, 0xffff, 0x424, 
0xffff, 0xffff, 0x718, 0x1199, 0xffff, 0xffff, 0x4fb, 0xffff, 
0xffff, 0xffff, 0xffff, 0x4d0, 0xfed, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0x10e1, 0xffff, 0xffff, 0xffff, 0x58, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0x11b4, 0x7f2, 0xaf3, 0xffff, 
0xffff, 0xf6a, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0x422, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xb32, 
0xf9e, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x7a9, 
0x374, 0xffff, 0x503, 0xffff, 0xe9a, 0xffff, 0x8e3, 0x3e3, 
0x763, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0x11c, 0x198, 0xaa3, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xc7e, 
0xffff, 0xffff, 0xffff, 0xffff, 0x68b, 0xeb4, 0x1190, 0xffff, 
0x61e, 0xffff, 0x78d, 0x1011, 0x1299, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0x367, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0x206, 0xffff, 0x3cf, 0x30e, 0xffff, 0x285, 0xffff, 
0xffff, 0xffff, 0xf82, 0x1027, 0xffff, 0xffff, 0xffff, 0xffff, 
0x5a3, 0x11ff, 0x3a9, 0x92c, 0xe05, 0xffff, 0x124d, 0xffff, 
0xf73, 0x3d1, 0xffff, 0x12c2, 0x1097, 0xffff, 0xffff, 0xffff, 
0x513, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0x1029, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x432, 0xffff, 0xffff, 
0xffff, 0xffff, 0x125f, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xfbd, 0xffff, 0xffff, 0xf59, 0xa41, 0xffff, 
0xffff, 0x917, 0xffff, 0xb4d, 0xa88, 0xffff, 0xffff, 0x4e4, 
0xffff, 0x94f, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xf1a, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0x342, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x9d0, 0xffff, 
0xffff, 0xc29, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x89c, 
0xffff, 0x26a, 0x12e6, 0xffff, 0x116a, 0xffff, 0x12a0, 0xa8d, 
0xffff, 0xffff, 0xffff, 0xe78, 0x10c1, 0x66e, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xbf5, 0x9e, 0xd32, 0xffff, 0xffff, 
0xffff, 0xffff, 0x40e, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xb9e, 0x12b0, 0xffff, 0xb34, 
0x33b, 0xffff, 0x854, 0x93c, 0xc74, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0x41a, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xe3, 0xffff, 0xffff, 0xffff, 
0xffff, 0x899, 0xffff, 0xffff, 0xffff, 0xffff, 0xd12, 0xab8, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x749, 0xffff, 
0x7b9, 0xbfa, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xe2e, 0xffff, 0x6c7, 0xffff, 0xffff, 0x10b, 
0x160, 0xffff, 0xffff, 0xe8e, 0xffff, 0xfae, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x3ed, 0x5f8, 
0xffff, 0xffff, 0x115b, 0x3ad, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0x1b1, 0x832, 0xffff, 0xffff, 0x7a5, 0x6ee, 
0x982, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x1089, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x11e2, 0x820, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x128a, 0x126c, 0xfe3, 
0xffff, 0x8d1, 0x150, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0x7c0, 0xc6a, 0xffff, 0xffff, 0xffff, 
0x1236, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xd2b, 0xffff, 0xffff, 0x1318, 0xddb, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xafb, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0x1287, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xb0b, 0xffff, 0xffff, 0xffff, 0xd24, 0xffff, 0x774, 
0x7b4, 0xffff, 0xffff, 0x583, 0x111e, 0x104b, 0xffff, 0x73c, 
0xffff, 0xffff, 0xffff, 0x415, 0xffff, 0xffff, 0xcf6, 0xffff, 
0x114a, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x2fa, 
0xffff, 0xffff, 0xffff, 0xffff, 0xea5, 0xffff, 0xffff, 0xffff, 
0xffff, 0x1d6, 0x3b5, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0x376, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xa89, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x117d, 0x1d5, 
0x59c, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0x784, 0xffff, 0xffff, 0x337, 0xffff, 0xffff, 0xffff, 0xb13, 
0xffff, 0xffff, 0xe46, 0xffff, 0x291, 0x690, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xd7f, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0x6d7, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0x737, 0xffff, 0xffff, 0xffff, 0xffff, 0x20c, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x89a, 
0xffff, 0xffff, 0xffff, 0xffff, 0xd41, 0xffff, 0xec0, 0x100d, 
0xffff, 0xffff, 0x616, 0xe48, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xcda, 0xffff, 0x54a, 0xffff, 
0xffff, 0xffff, 0x114, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0x454, 0xffff, 0xffff, 0xffff, 0x120e, 0xace, 0xffff, 0xd75, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0x5d4, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0x64e, 0xffff, 0xffff, 0xbf8, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0x9d2, 0xffff, 0xffff, 0xffff, 0xffff, 0x110f, 0xffff, 0xffff, 
0x26e, 0xffff, 0xffff, 0xffff, 0xc68, 0xdc3, 0xffff, 0xff6, 
0xffff, 0x2f2, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0x469, 0xffff, 0xffff, 0x6dc, 0x125a, 0xffff, 0xffff, 0xffff, 
0xffff, 0x459, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xb1b, 0xffff, 0xffff, 0xec7, 0xaad, 0xffff, 0xffff, 0xffff, 
0xffff, 0x167, 0xffff, 0xffff, 0x126d, 0xffff, 0xffff, 0x6c8, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xe4, 0xffff, 0xffff, 
0x70c, 0xf2a, 0x1165, 0xe59, 0x89d, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0x662, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0x6e6, 0xc8a, 0xffff, 0x566, 0xffff, 0xffff, 
0xffff, 0xffff, 0x865, 0xede, 0xffff, 0xffff, 0xb9f, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xad1, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0x57f, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xf3c, 
0xffff, 0xce3, 0xffff, 0xffff, 0x67c, 0xffff, 0x1207, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x6eb, 
0x394, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0x2e4, 0x1193, 0xffff, 0xd7a, 0x3bb, 0xffff, 0xffff, 0xffff, 
0xffff, 0xd51, 0xde3, 0xffff, 0xffff, 0xffff, 0xd26, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xee3, 0xffff, 0x106c, 
0xffff, 0x11c0, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0x101, 0x569, 0xffff, 0xffff, 0xffff, 0x354, 
0x955, 0xffff, 0xffff, 0xffff, 0xffff, 0x7bd, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0x25d, 0xf90, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xaf8, 0xffff, 0xffff, 0x11f3, 0x859, 
0xffff, 0xffff, 0x112c, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0x113c, 0xffff, 0xffff, 0xffff, 0x631, 0xf34, 0x11d3, 
0xf8c, 0x851, 0x11f5, 0xf01, 0xa3d, 0xf30, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xf84, 0xffff, 0xffff, 0x341, 
0x646, 0xffff, 0xad2, 0x79b, 0x814, 0x68e, 0x6c9, 0xffff, 
0x450, 0xffff, 0xffff, 0x3f7, 0xffff, 0xffff, 0xffff, 0x83c, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x62e, 0xffff, 0xffff, 
0xffff, 0x325, 0xffff, 0xffff, 0xd9d, 0x43, 0x7f3, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x72c, 
0xffff, 0xeb, 0x1237, 0xffff, 0xffff, 0x95, 0xe34, 0xffff, 
0xffff, 0xffff, 0x267, 0xffff, 0xffff, 0xffff, 0xffff, 0xaa6, 
0xffff, 0xffff, 0xffff, 0x964, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0x250, 0xffff, 0xffff, 0xffff, 0x264, 0xa66, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xe65, 0xffff, 0xffff, 
0xb11, 0xffff, 0xffff, 0xa6a, 0xffff, 0xffff, 0xffff, 0x7ed, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xd1d, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xa97, 0xffff, 
0xffff, 0x1095, 0xffff, 0xdd3, 0xffff, 0xffff, 0xffff, 0xffff, 
0xee9, 0xffff, 0xffff, 0xffff, 0xc5, 0xf4d, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0x8df, 0x78e, 0xffff, 0x3fe, 
0xffff, 0xffff, 0xffff, 0x768, 0xffff, 0xdbe, 0xffff, 0xffff, 
0xffff, 0xaa0, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0x3e4, 0xffff, 0xffff, 0x612, 0xffff, 
0xffff, 0xffff, 0xb4f, 0xcf0, 0xffff, 0x251, 0x40c, 0x75e, 
0xbbe, 0x13b, 0x952, 0xffff, 0xffff, 0xffff, 0x7d8, 0xffff, 
0xffff, 0xffff, 0x639, 0xb03, 0xe66, 0xffff, 0xffff, 0xffff, 
0xffff, 0xdda, 0x5b0, 0xffff, 0x931, 0xffff, 0xffff, 0xffff, 
0x5bf, 0x10ac, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x11ec, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x559, 0x1c0, 0xffff, 
0xffff, 0xffff, 0xe5a, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xbef, 0x4f2, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xd05, 0xffff, 0xffff, 
0x8bb, 0x1df, 0xec4, 0xffff, 0x697, 0xd60, 0xffff, 0xffff, 
0xffff, 0xffff, 0x713, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0x5ce, 0xffff, 0xffff, 0xffff, 0x1253, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xc41, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xb4c, 0xcac, 0xffff, 
0x19f, 0xfd1, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xcde, 0xffff, 0xffff, 0xffff, 0x77a, 
0xffff, 0x6a0, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0x922, 0xffff, 0xffff, 0xffff, 0xffff, 0xbd2, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0x5ca, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0x1021, 0x52, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0x8ae, 0xffff, 0x134, 0xffff, 0xffff, 0xffff, 0x69a, 
0x1b3, 0xa79, 0xffff, 0x10c, 0xffff, 0x5fb, 0xffff, 0x352, 
0x172, 0x3e2, 0xffff, 0xffff, 0xffff, 0xffff, 0x520, 0xb52, 
0xffff, 0xffff, 0x46b, 0x1281, 0xaed, 0xffff, 0x9e6, 0xffff, 
0xffff, 0xf74, 0xffff, 0x5fe, 0xffff, 0x12ac, 0xffff, 0xffff, 
0xa6, 0x649, 0x204, 0x59e, 0xa1e, 0xffff, 0xffff, 0xffff, 
0xf0e, 0xffff, 0xffff, 0xffff, 0x512, 0xffff, 0x11b6, 0xffff, 
0x738, 0xffff, 0xffff, 0xf2e, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0x13e, 0x96f, 0x1195, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xac1, 0xa64, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0x199, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0x11a5, 0xffff, 0xffff, 0x6e9, 0x8a7, 0xffff, 0xffff, 0xbe1, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xdb6, 0xffff, 0xba4, 0xffff, 0xa62, 0xffff, 0x10bb, 0xae5, 
0xffff, 0xffff, 0xffff, 0xb3, 0xd36, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xf5e, 0xffff, 0xa0d, 0xffff, 
0xffff, 0xffff, 0xffff, 0x553, 0x1098, 0xffff, 0xffff, 0x22e, 
0x107c, 0xffff, 0xffff, 0xffff, 0xffff, 0x3d0, 0xffff, 0xffff, 
0xffff, 0xffff, 0x6d8, 0xffff, 0xd97, 0x11cf, 0xffff, 0xffff, 
0xffff, 0x108e, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0x10b5, 0xffff, 0xffff, 0xffff, 0xffff, 
0xe1e, 0x110a, 0xffff, 0xffff, 0xffff, 0x869, 0x1285, 0xffff, 
0xffff, 0xffff, 0x83, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x1ae, 
0xea2, 0xffff, 0xdb1, 0x5c8, 0x1277, 0xffff, 0xffff, 0x486, 
0xffff, 0xffff, 0xffff, 0xffff, 0x534, 0xffff, 0xc2c, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x3cb, 0xb2e, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xf0f, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xbd3, 0xffff, 0xffff, 0xffff, 
0xffff, 0xd11, 0xffff, 0xffff, 0xffff, 0x37e, 0x5f7, 0x6ae, 
0xffff, 0xffff, 0xffff, 0xffff, 0x11a1, 0xffff, 0x966, 0xffff, 
0xe28, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xb49, 0xffff, 0xffff, 0xffff, 0x1284, 0xffff, 0xef5, 
0xffff, 0x882, 0xffff, 0xffff, 0xffff, 0xcdb, 0xec6, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x47a, 0x965, 
0xffff, 0xffff, 0xffff, 0xffff, 0x4b, 0xe0c, 0xffff, 0xffff, 
0x98b, 0xd09, 0xffff, 0xfd6, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xbd8, 0x3e8, 0xb38, 0x878, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x26c, 0xf9c, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0x443, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0x12af, 0xffff, 0x622, 0x430, 
0x703, 0xffff, 0xffff, 0x1217, 0xa95, 0x1030, 0xffff, 0x269, 
0xffff, 0xffff, 0xffff, 0xf4, 0xffff, 0xffff, 0xffff, 0xffff, 
0x4ed, 0xffff, 0xffff, 0xffff, 0xffff, 0xb23, 0xffff, 0xffff, 
0xffff, 0xffff, 0x98e, 0xffff, 0x1068, 0xffff, 0xffff, 0xffff, 
0xffff, 0x359, 0x17e, 0x900, 0xffff, 0xffff, 0xffff, 0xffff, 
0xd7c, 0xa4a, 0xffff, 0xffff, 0x270, 0xc36, 0x129d, 0xffff, 
0xffff, 0xffff, 0xffff, 0x5d9, 0xef9, 0xffff, 0x316, 0xffff, 
0xffff, 0x4ff, 0xffff, 0x9f0, 0x101e, 0xffff, 0xffff, 0x96d, 
0xffff, 0xffff, 0x1104, 0xa35, 0xffff, 0xffff, 0x106a, 0xffff, 
0xffff, 0xffff, 0xffff, 0x2db, 0x1268, 0xffff, 0xffff, 0xe7c, 
0xffff, 0x44c, 0x10e3, 0xffff, 0xffff, 0xffff, 0xffff, 0xad6, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xf75, 0xffff, 0x3b4, 
0xffff, 0x794, 0xe43, 0x3a0, 0x37a, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x38, 0xc84, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x581, 
0xffff, 0xffff, 0xffff, 0x248, 0x290, 0xb61, 0xd62, 0x775, 
0x4c6, 0xffff, 0xffff, 0xffff, 0xffff, 0xae, 0xb50, 0xffff, 
0x2cb, 0x866, 0xcf9, 0x4a7, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xa6d, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0x471, 0x72b, 0x1264, 0xffff, 
0xffff, 0x7ea, 0xfe0, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xb3d, 0xffff, 0x1112, 0xc8b, 0xffff, 0xffff, 0xffff, 0xffff, 
0xa4, 0x130, 0x90d, 0xffff, 0xffff, 0xffff, 0x5e7, 0x95a, 
0xd50, 0xffff, 0xffff, 0xffff, 0xffff, 0x7fb, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xd16, 0xffff, 0x75b, 
0xffff, 0xffff, 0xffff, 0x12b9, 0xffff, 0x5eb, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x539, 
0xffff, 0x7c7, 0xffff, 0xffff, 0xffff, 0x829, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x546, 
0x40a, 0x903, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xcf5, 0xffff, 0x1113, 0xffff, 0xffff, 0xab3, 0x1221, 
0x50c, 0x158, 0xa5b, 0xcfe, 0x458, 0x63a, 0xca6, 0xffff, 
0xb9c, 0xffff, 0xffff, 0xffff, 0xdd4, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xb70, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0x166, 0x8c2, 0xa51, 0xb5b, 0xdb0, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x405, 
0x1279, 0x100c, 0xffff, 0xffff, 0xffff, 0x1ee, 0xffff, 0xffff, 
0xffff, 0xffff, 0x4bd, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0x521, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0x69e, 0xe9c, 0xffff, 0xffff, 0xffff, 0xffff, 0x31a, 0x74d, 
0x8bf, 0x98c, 0xffff, 0xffff, 0xee0, 0xf9a, 0x7ce, 0xffff, 
0xbe5, 0xffff, 0xffff, 0xc76, 0x70b, 0xe6e, 0x1015, 0x112d, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x857, 
0xffff, 0xffff, 0xdaf, 0xffff, 0x6c6, 0x116b, 0x6b6, 0xaf6, 
0x2d9, 0xffff, 0xffff, 0x47, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xdbc, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0x7a6, 0xffff, 0x11da, 0xffff, 0x19d, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x6f, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xfbc, 0xffff, 0x237, 0x10e0, 
0xffff, 0xffff, 0x19b, 0xffff, 0xffff, 0xffff, 0xe72, 0x41d, 
0xcaf, 0xcf7, 0x105b, 0xffff, 0xffff, 0x1a7, 0xffff, 0xffff, 
0xffff, 0xa5, 0xc45, 0xffff, 0xa58, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xc5d, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0x58a, 0x7a7, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0x9ee, 0xe73, 0xffff, 0x719, 0xffff, 
0xd2d, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xda7, 
0xd8a, 0xffff, 0xffff, 0xffff, 0xffff, 0x7b3, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xc78, 0xffff, 0xffff, 0xffff, 0xb1a, 
0xffff, 0xffff, 0xffff, 0xffff, 0x9a7, 0xffff, 0x3a5, 0xffff, 
0xffff, 0xffff, 0xffff, 0x8d5, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xf5f, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xe0f, 0xffff, 0xffff, 0xffff, 
0x298, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xa86, 0xffff, 
0xffff, 0x3be, 0xffff, 0xffff, 0xffff, 0xcbd, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xef3, 0xffff, 
0x8a9, 0xffff, 0x1235, 0xffff, 0xffff, 0xffff, 0x731, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x1187, 0xffff, 0x61d, 
0x979, 0xffff, 0xffff, 0xffff, 0x57b, 0x11a6, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x6ff, 0xffff, 0xffff, 
0xffff, 0xffff, 0x10a6, 0xffff, 0x4e6, 0xffff, 0xffff, 0xffff, 
0xffff, 0x949, 0xffff, 0xffff, 0x5ab, 0xffff, 0xffff, 0xffff, 
0x6d, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0x416, 0x5e5, 0x836, 0xb26, 
0xd47, 0xffff, 0x516, 0x1105, 0xffff, 0x212, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0x80d, 0xffff, 0xffff, 0xffff, 0x2ca, 
0xffff, 0x5e2, 0xffff, 0x699, 0xffff, 0xffff, 0xffff, 0x9af, 
0xffff, 0xffff, 0xffff, 0xc4c, 0xffff, 0xffff, 0x40, 0xd8f, 
0x45c, 0xb29, 0xce, 0xffff, 0xffff, 0xffff, 0x81c, 0x60b, 
0x892, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0x877, 0xea8, 0xffff, 0xffff, 0xb0f, 0x5dd, 0x84a, 0xffff, 
0xffff, 0x127a, 0xffff, 0xffff, 0xffff, 0xffff, 0x7b6, 0x816, 
0xffff, 0xffff, 0xffff, 0xffff, 0xbff, 0xc27, 0xffff, 0xe1c, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xf22, 0xffff, 
0x2a8, 0x85, 0x386, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0x1022, 0x27b, 0xffff, 0xeb6, 0xffff, 0xffff, 0xffff, 
0x3ce, 0xaca, 0xd8d, 0xffff, 0xffff, 0xffff, 0xffff, 0x5a8, 
0x28f, 0xffff, 0xffff, 0xffff, 0x788, 0x105c, 0xffff, 0xffff, 
0xffff, 0xffff, 0x9a0, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0x556, 0xaa5, 0xffff, 0xf49, 0xf97, 0x10a8, 0xffff, 0xffff, 
0x1180, 0xffff, 0xffff, 0xffff, 0xffff, 0x940, 0xffff, 0x90f, 
0xa84, 0xeca, 0xffff, 0xffff, 0x51b, 0xffff, 0xffff, 0x34, 
0x82e, 0xffff, 0xaf1, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0x12ad, 0xffff, 0xffff, 0xffff, 0xffff, 0xc32, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0x700, 0xffff, 0x8af, 0xffff, 
0xffff, 0x1094, 0xffff, 0xffff, 0x11b3, 0xffff, 0xce2, 0x60f, 
0xffff, 0xffff, 0x11e8, 0xffff, 0xffff, 0xffff, 0xffff, 0x6e0, 
0x911, 0xffff, 0xffff, 0xbfd, 0xffff, 0x4d5, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x7df, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0x8bc, 0xffff, 0xffff, 0x796, 0xffff, 
0x5fa, 0xffff, 0xffff, 0x120b, 0xa61, 0xffff, 0xffff, 0xffff, 
0xba8, 0xffff, 0xffff, 0x909, 0xffff, 0xffff, 0xd39, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0x7de, 0xffff, 0x715, 0xe60, 0xffff, 
0xffff, 0x1046, 0xec2, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0x67b, 0xffff, 0xffff, 0x25c, 0xb2a, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0x2e6, 0xffff, 0xffff, 0x8dd, 
0xffff, 0x71d, 0xe49, 0xffff, 0xffff, 0xffff, 0xffff, 0x29d, 
0x809, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x1239, 
0x21, 0xffff, 0xffff, 0xffff, 0xe4a, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0x57e, 0xffff, 0xffff, 0xffff, 0xffff, 0x1b4, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x2f9, 0xffff, 0xffff, 
0xffff, 0xffff, 0xd20, 0xb48, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xa4c, 0xdfc, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0x4c4, 0x5c1, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x113d, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xc97, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0x68a, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0x7a1, 0xcfd, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0x92d, 0xffff, 0xffff, 0xffff, 0xd94, 0xb02, 
0x127, 0xac7, 0x37f, 0xf5c, 0xf1f, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xc3a, 0xffff, 0xf86, 0xdde, 0x98, 0x9e5, 
0xffff, 0xffff, 0xffff, 0xffff, 0x8cb, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0x338, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0x933, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xf4c, 0xffff, 0xffff, 0xffff, 0xfa5, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xb5a, 0xffff, 0xef0, 0xffff, 0x801, 0xffff, 
0xffff, 0xffff, 0x404, 0xffff, 0x1e3, 0xadb, 0xde4, 0xffff, 
0xffff, 0x1106, 0xffff, 0xffff, 0xffff, 0xd6e, 0xffff, 0xffff, 
0xffff, 0xffff, 0x11f, 0xffff, 0xffff, 0xa1, 0xffff, 0xffff, 
0x123, 0x321, 0xffff, 0xffff, 0x7ba, 0x10f8, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x127b, 0x311, 0xffff, 
0x12f4, 0x1243, 0xffff, 0xffff, 0x65f, 0xdb4, 0x1170, 0x93b, 
0x12e4, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0x243, 0xffff, 0xffff, 0xffff, 0xffff, 0xd99, 
0xffff, 0xffff, 0x5af, 0xffff, 0xffff, 0xffff, 0x6e3, 0x107f, 
0xffff, 0xffff, 0xffff, 0x2ba, 0xffff, 0xffff, 0x1290, 0xffff, 
0xffff, 0x880, 0xffff, 0xffff, 0xffff, 0x12d9, 0xffff, 0xffff, 
0xffff, 0xffff, 0x660, 0xffff, 0xffff, 0xffff, 0x12f9, 0xffff, 
0xddf, 0x598, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xfd2, 0xffff, 0xffff, 0xffff, 0xffff, 0x49, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xd2a, 0xffff, 0x1260, 0xfc5, 
0xffff, 0xd72, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xc3c, 0xffff, 0x64a, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xb57, 0xffff, 0x9b0, 0x11de, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x9ff, 0x9a4, 
0xffff, 0xffff, 0xffff, 0xffff, 0xb6b, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0x410, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x16f, 0xffff, 
0x27c, 0xffff, 0xffff, 0xdcc, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0x1084, 0xffff, 0x605, 0xffff, 0xe04, 0xa27, 
0xffff, 0xdf8, 0xffff, 0xffff, 0x9c9, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0x2ab, 0x181, 0x2bf, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x1305, 
0x38c, 0xffff, 0xffff, 0x8f0, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0x103a, 0xffff, 0xffff, 0xffff, 0xffff, 0xc88, 0xffff, 
0xffff, 0x2fe, 0xffff, 0xffff, 0xffff, 0xaf9, 0x144, 0x12ba, 
0xffff, 0xffff, 0xffff, 0xffff, 0x6e5, 0x838, 0xffff, 0x31b, 
0x1301, 0xffff, 0x9cb, 0xffff, 0xffff, 0x9f3, 0xce7, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xc12, 0x1163, 0x10d1, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0x1af, 0x1c2, 0xffff, 0x63e, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x4c1, 0xffff, 0xffff, 
0xffff, 0x35a, 0x1204, 0xffff, 0xffff, 0x10c4, 0xffff, 0xf2b, 
0xb65, 0x62a, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0x508, 0xffff, 0xffff, 0xffff, 
0x38e, 0xffff, 0xffff, 0x2d4, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0x293, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0x46a, 0xffff, 0xffff, 0x10ec, 0xffff, 0xffff, 0x8b0, 0xffff, 
0xffff, 0xffff, 0xffff, 0xf17, 0xffff, 0xffff, 0x1f3, 0xa36, 
0xc73, 0x119a, 0xb0, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0x32f, 0x9ce, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xf91, 0xffff, 0xffff, 
0x7a8, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0x73d, 0xffff, 0xffff, 0xccb, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xb18, 0xffff, 
0xffff, 0x664, 0xd1f, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x54, 0xffff, 0xffff, 
0xffff, 0xc0e, 0x122, 0xffff, 0xed9, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xb96, 0x997, 0x69d, 
0xffff, 0x9b6, 0x1ba, 0x240, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x689, 0xaf0, 0xe42, 
0xffff, 0xffff, 0xffff, 0xffff, 0x188, 0xe19, 0xffff, 0xffff, 
0xffff, 0x371, 0xffff, 0xffff, 0xffff, 0xffff, 0xa1d, 0xca3, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x10af, 0x7fd, 
0xffff, 0xffff, 0xf6e, 0xd35, 0x8ef, 0xffff, 0xffff, 0xffff, 
0xffff, 0xc3, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0x522, 0xffff, 0x124e, 0x319, 
0xcfb, 0x1197, 0x10d6, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0x4da, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0x111f, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x22, 
0xb4a, 0xc8, 0xffff, 0x51, 0xffff, 0x500, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xbb7, 0xffff, 0xffff, 
0x46c, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0x6b3, 0xb75, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xe9e, 0xffff, 0xffff, 0x9f2, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xc4a, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0x2d8, 0xffff, 0x63c, 0xffff, 0xffff, 0x45d, 
0x808, 0xffff, 0x1013, 0xffff, 0xffff, 0x5d6, 0xffff, 0xffff, 
0x3c, 0xd84, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0x4cb, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0x1118, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0x75f, 0x98f, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0x654, 0x932, 0xffff, 0xffff, 
0xffff, 0xac2, 0xffff, 0xffff, 0xffff, 0xbbd, 0x2dd, 0xa4e, 
0xc6c, 0xffff, 0x11df, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0x1251, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x2d1, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0x1000, 0xffff, 0xffff, 0x7ab, 0x999, 
0xffff, 0xffff, 0x3c5, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x9ae, 0x732, 0xffff, 
0xffff, 0x43d, 0xffff, 0xffff, 0xffff, 0x35d, 0xaaf, 0xe39, 
0xf9b, 0x1043, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xdc6, 
0xffff, 0xffff, 0xffff, 0xffff, 0xdfd, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0x7b, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0x79, 0xa1a, 0xdc8, 0xffff, 0xffff, 0x5b4, 0x723, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0x9e0, 0xffff, 0x1cf, 0xffff, 
0x12a6, 0xffff, 0xffff, 0x31, 0x702, 0x1054, 0xffff, 0xffff, 
0x4a1, 0xffff, 0xffff, 0x8b9, 0x10c0, 0xffff, 0xffff, 0xb59, 
0xffff, 0xffff, 0xffff, 0xf76, 0x677, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xb0c, 0x1319, 0x12ce, 0x5a6, 0xffff, 0xc46, 
0xd3d, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0x13f, 0xe00, 0xc17, 0xf43, 0xf54, 0x1ac, 0x10f, 0x304, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x261, 0xf20, 0xffff, 
0xffff, 0xffff, 0xffff, 0x1291, 0xffff, 0x9df, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x2de, 
0xffff, 0xffff, 0x11a2, 0xffff, 0x1035, 0x3d2, 0xffff, 0xffff, 
0xffff, 0x4b6, 0xffff, 0xa82, 0xd92, 0xffff, 0xffff, 0xffff, 
0xa9b, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x9a3, 0xffff, 
0x292, 0xffff, 0xffff, 0xffff, 0xffff, 0x3e5, 0xffff, 0x149, 
0xc0, 0x10a3, 0x9bd, 0xffff, 0xffff, 0xffff, 0xffff, 0xa22, 
0xd1c, 0xdc4, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0x24f, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x1308, 0xc24, 
0xffff, 0xffff, 0xffff, 0xffff, 0x1a3, 0xe0b, 0xffff, 0x6ea, 
0xffff, 0x6da, 0x1091, 0xffff, 0xffff, 0xedf, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xcb0, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0x238, 0xad3, 0xffff, 0xffff, 0xffff, 
0x12cf, 0x102b, 0x364, 0xffff, 0xffff, 0xffff, 0xffff, 0xda8, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xa7d, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0x764, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0x7f5, 0xffff, 0x14, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0x4de, 0xcc1, 0xffff, 0xffff, 0xffff, 0x6c, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0x10ad, 0x370, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0x11ce, 0xffff, 0x130c, 0xffff, 0x896, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xd67, 0xfc3, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xf04, 0xffff, 0xb12, 0xffff, 
0xffff, 0x26b, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0x11b1, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xfac, 0xffff, 0x987, 0xffff, 0xffff, 
0x10fc, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0x13, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xe40, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xcd5, 0xffff, 
0xffff, 0xffff, 0xc01, 0xffff, 0xffff, 0xc26, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x2e8, 0xffff, 
0x1175, 0xffff, 0xffff, 0xffff, 0x141, 0xf12, 0xffff, 0xdb2, 
0xffff, 0xffff, 0x17b, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0x10cf, 0xffff, 0xc5b, 0xffff, 0xffff, 
0xffff, 0x11bd, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0x129e, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xb79, 
0xffff, 0x409, 0xffff, 0x2ec, 0xffff, 0xffff, 0xc63, 0xffff, 
0xffff, 0xffff, 0xa0b, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xbae, 0xeb5, 0x103f, 
0x116c, 0x8e8, 0xffff, 0x2a5, 0xf0c, 0xffff, 0xffff, 0xffff, 
0x1dd, 0xffff, 0xffff, 0xffff, 0xd37, 0xffff, 0xffff, 0xffff, 
0xffff, 0xd30, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0x116e, 0xffff, 0xffff, 0xffff, 0xa75, 0xffff, 0x70e, 0xffff, 
0x8f4, 0xffff, 0xa09, 0xc4e, 0xffff, 0xcab, 0xa5d, 0xffff, 
0x485, 0xffff, 0xa6e, 0xffff, 0xffff, 0xffff, 0xffff, 0xb98, 
0x29, 0xffff, 0x4f3, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xbc1, 0x993, 
0xb4e, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xef, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xf05, 0xffff, 0xcb8, 0xffff, 0x7eb, 0xffff, 
0xba2, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xe2c, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xcd, 0xffff, 0xffff, 0x3a8, 0x41f, 
0xffff, 0x3ab, 0xffff, 0x11e3, 0xffff, 0x11d8, 0x297, 0xffff, 
0xffff, 0xffff, 0x385, 0xcc0, 0xffff, 0xffff, 0xffff, 0x1148, 
0xffff, 0xffff, 0xffff, 0xe8a, 0xcb5, 0xffff, 0xffff, 0x5ac, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x10da, 0xffff, 
0x11eb, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xd33, 0xffff, 0x590, 0xffff, 0x53, 0x4fe, 0xffff, 0xffff, 
0xffff, 0x1234, 0x12fd, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0x21d, 0x7c3, 0xfec, 0xffff, 0xd19, 0xffff, 0x6a1, 
0xdf9, 0xffff, 0x78c, 0xffff, 0xffff, 0x101d, 0x6d6, 0x1052, 
0x1226, 0xffff, 0xffff, 0xf3b, 0xffff, 0xffff, 0xf78, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0x1176, 0x7d4, 0xe55, 0xffff, 
0xffff, 0xffff, 0xffff, 0x5d8, 0xffff, 0x12d0, 0xffff, 0x122a, 
0xffff, 0x50, 0x318, 0x1225, 0x122f, 0xffff, 0xffff, 0xefd, 
0xa2a, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x799, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0x872, 0xffff, 0x91c, 0x555, 
0xfb3, 0x115e, 0xffff, 0xffff, 0xffff, 0xffff, 0xada, 0xffff, 
0x1153, 0xffff, 0x18c, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0x11c8, 0xffff, 0xffff, 0xffff, 0x10ab, 0xed6, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x63d, 0xffff, 
0xffff, 0xffff, 0xffff, 0xeed, 0xffff, 0xffff, 0x795, 0xffff, 
0x79e, 0xffff, 0xa78, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0x284, 0xffff, 0x86c, 0x1310, 0xffff, 
0xffff, 0xffff, 0x129b, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0x116, 0xffff, 0xffff, 0x2b6, 0x103e, 0x35f, 
0x7e6, 0x937, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x3c9, 0xffff, 
0xffff, 0xffff, 0xffff, 0xb47, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0x7c2, 0xffff, 0xffff, 0xffff, 
0xffff, 0x833, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0x1eb, 0xffff, 0xffff, 0x2b7, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0x202, 0xd0e, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0x6f7, 0xffff, 0xf69, 0xffff, 0x6a7, 
0x125e, 0x380, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xbb3, 0xffff, 0xffff, 0xffff, 0xc82, 0xd98, 0xe4c, 0x1041, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0x8c, 0xffff, 0xffff, 0xffff, 0xffff, 0x78a, 
0x1271, 0xffff, 0x860, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0x6ad, 0x8f7, 0x92f, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0x106, 0xffff, 0xffff, 0xffff, 0xffff, 0x59b, 
0xffff, 0xffff, 0x970, 0xe02, 0xf71, 0xffff, 0xffff, 0xffff, 
0xffff, 0x9de, 0xffff, 0xffff, 0x388, 0xffff, 0xffff, 0xffff, 
0xffff, 0xc40, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x11a4, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0x821, 0xffff, 0x15f, 0x80b, 0xffff, 0xffff, 0xffff, 0xffff, 
0xaa4, 0xffff, 0xffff, 0x12dc, 0xe52, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0x122e, 0xffff, 0x4ee, 0x153, 0xb00, 0x1d, 0xc0d, 0x56e, 
0x103c, 0xffff, 0xffff, 0xffff, 0xffff, 0x537, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xbcb, 0xffff, 0xef1, 0xa3, 
0x115f, 0xffff, 0x800, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xab, 0xab7, 0xb1d, 0xffff, 0xffff, 0xffff, 0xe8, 0xffff, 
0xffff, 0xffff, 0x635, 0xf56, 0xf5b, 0x12e0, 0xffff, 0xffff, 
0xffff, 0x1306, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0x29a, 0x381, 0xdef, 0xffff, 0xffff, 0xffff, 
0x474, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0x7cb, 0x96e, 0xffff, 0xffff, 0x11b2, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xe5f, 0xb54, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0x74b, 0xffff, 0x5b7, 0x11fd, 0xffff, 0xffff, 
0x12d5, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0x10dc, 0xffff, 0x170, 0x958, 0xffff, 0xffff, 0xdad, 0xffff, 
0xffff, 0xffff, 0xffff, 0x8dc, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xc2d, 0xffff, 0xffff, 
0xffff, 0xffff, 0xdc9, 0xb8c, 0x12c, 0xf1c, 0x81d, 0xf98, 
0xffff, 0xffff, 0x478, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xd85, 0x1149, 0xffff, 0xffff, 0xffff, 0xffff, 0xca9, 0xffff, 
0xffff, 0xffff, 0xffff, 0x35e, 0xffff, 0xffff, 0x8cf, 0xbba, 
0xe9f, 0xffff, 0xffff, 0xffff, 0x4a2, 0xe23, 0xffff, 0xad9, 
0xffff, 0x2b4, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xe9, 
0xf25, 0xa76, 0xfc2, 0xffff, 0xffff, 0xf8, 0x73b, 0xbed, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xc81, 0x8a4, 
0x1139, 0xffff, 0xffff, 0x679, 0x92a, 0xffff, 0xffff, 0xffff, 
0xffff, 0x1300, 0xa68, 0xffff, 0xffff, 0x483, 0xae1, 0xe88, 
0x4c0, 0x12b8, 0xffff, 0xffff, 0x82f, 0xffff, 0xffff, 0xffff, 
0x575, 0xbc7, 0xffff, 0xd10, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0x529, 0x1151, 0xffff, 0xffff, 
0xffff, 0x11d6, 0xffff, 0xffff, 0xcdf, 0xffff, 0xffff, 0xffff, 
0xf2f, 0xc48, 0xffff, 0xffff, 0xffff, 0x12f2, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0x1004, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x1126, 
0x5d0, 0xffff, 0x28, 0xffff, 0xffff, 0xffff, 0xc3d, 0xffff, 
0xffff, 0x9d6, 0xffff, 0xffff, 0xffff, 0xffff, 0xa6c, 0xffff, 
0x2e2, 0xffff, 0xa46, 0xffff, 0x11d7, 0xffff, 0x23b, 0xad4, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x3b, 0xcb4, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x5f3, 0xffff, 
0xffff, 0x4d3, 0x777, 0xffff, 0xffff, 0xffff, 0xffff, 0x70d, 
0x105a, 0xffff, 0x130a, 0xffff, 0xffff, 0x1bd, 0x94c, 0xffff, 
0x9d8, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0x886, 0xffff, 0xce8, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x27f, 0xffff, 0xffff, 
0xcbf, 0xffff, 0xffff, 0x11bc, 0xffff, 0xffff, 0x5bd, 0xffff, 
0xffff, 0xffff, 0x71b, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x594, 0x850, 
0x118c, 0x12ea, 0xffff, 0xffff, 0xad8, 0xfe8, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x1228, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x8a, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x1313, 0xffff, 0xffff, 
0x46d, 0x636, 0x708, 0xd31, 0xda9, 0xffff, 0xffff, 0xffff, 
0x226, 0xbfb, 0xffff, 0xffff, 0xffff, 0xffff, 0x1066, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x2bd, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x842, 0xffff, 0x2e9, 
0x839, 0xffff, 0xffff, 0xf21, 0xdb3, 0xffff, 0xffff, 0xffff, 
0x767, 0xffff, 0xffff, 0x1132, 0xffff, 0x6f5, 0x10e2, 0xffff, 
0xf79, 0x929, 0xffff, 0xab2, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xd4f, 0x65e, 0xffff, 0xffff, 
0x2f6, 0x401, 0x90e, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xcf4, 0xffff, 0xc56, 0xffff, 0xffff, 
0xa30, 0xffff, 0xffff, 0x9cf, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x1125, 
0xffff, 0xbe7, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0x6d3, 0xffff, 0xebb, 0xa38, 0xffff, 0xffff, 0xffff, 0x86, 
0xffff, 0xbe2, 0xe92, 0x1266, 0xffff, 0xffff, 0xc0c, 0x10ca, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x57c, 0xffff, 0xffff, 
0xffff, 0x517, 0xbf9, 0xffff, 0xffff, 0xffff, 0xffff, 0x625, 
0xffff, 0xffff, 0xffff, 0xffff, 0x5f6, 0xb88, 0x7f6, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xa08, 0xffff, 
0xffff, 0xee1, 0xffff, 0xa6f, 0xffff, 0x984, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xb, 0x375, 0xffff, 
0xffff, 0xffff, 0x79f, 0x1053, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0x351, 0xbda, 0xffff, 0xffff, 0x577, 0xca4, 
0xffff, 0x6a, 0xffff, 0x16b, 0x2e0, 0xffff, 0xffff, 0x6fe, 
0x9db, 0xb94, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0x37d, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x1109, 
0x5f1, 0xffff, 0xffff, 0x11e4, 0xffff, 0xab9, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xb31, 0x75d, 0x129a, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0x45f, 0x123a, 0xffff, 0x885, 
0xb72, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0x104, 0x6ca, 0xc90, 0x36, 0xfdc, 0x20a, 0xffff, 0xffff, 
0xba9, 0xffff, 0x943, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x695, 0x8fa, 
0xffff, 0xffff, 0xffff, 0xffff, 0x8ba, 0xffff, 0xffff, 0xffff, 
0xffff, 0x1031, 0xffff, 0xe26, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xef4, 0xbb5, 0xffff, 0x1d7, 0xffff, 0x71, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0x31c, 0x11a9, 0xffff, 0xffff, 0xffff, 0xab4, 0xc4, 0xa04, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x3f0, 
0xffff, 0xffff, 0xffff, 0x222, 0xffff, 0x10e4, 0xffff, 0x1063, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x1134, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0x91e, 0xffff, 0xffff, 0xffff, 0xffff, 0xfa0, 0x567, 
0xffff, 0xffff, 0xffff, 0x117a, 0x307, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0x390, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xfb0, 
0xffff, 0xffff, 0xa69, 0xe3f, 0xffff, 0x1049, 0x873, 0xffff, 
0xbe6, 0xffff, 0xffff, 0xffff, 0x39, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xbc9, 0xffff, 0xd6c, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x883, 0x9a5, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x35, 0x771, 0xffff, 
0xffff, 0x924, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xd1e, 0xffff, 0xffff, 0xffff, 0xba7, 0xffff, 0xffff, 0xffff, 
0x5b3, 0x6e4, 0xffff, 0xffff, 0x92b, 0xffff, 0xffff, 0xb77, 
0xffff, 0xffff, 0xffff, 0xffff, 0x4ab, 0xffff, 0xffff, 0xffff, 
0x3c8, 0xffff, 0xffff, 0xffff, 0xffff, 0xfd7, 0xffff, 0xffff, 
0xffff, 0x11f1, 0x105d, 0x11e9, 0xffff, 0x12cb, 0xffff, 0xe20, 
0xffff, 0x12d, 0xf47, 0xffff, 0xb2, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0x1210, 0xffff, 0x47b, 0x8da, 0x77f, 
0x113a, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0x25f, 0xb04, 0x5bc, 0x38b, 0x4e1, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x582, 
0xffff, 0xffff, 0xdea, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xcdc, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0x278, 0xffff, 0xffff, 0x333, 0xffff, 0x7aa, 0xd17, 
0x1c9, 0x8b6, 0x95e, 0xffff, 0xffff, 0xffff, 0x607, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xed8, 0xffff, 0x262, 0xa3b, 0xffff, 0xffff, 
0xffff, 0xffff, 0x8cc, 0xffff, 0xffff, 0xffff, 0xffff, 0x57d, 
0xffff, 0xffff, 0xffff, 0xffff, 0xd4a, 0xffff, 0x39c, 0xfbb, 
0x905, 0xffff, 0xffff, 0xffff, 0x109d, 0xffff, 0xffff, 0x8cd, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xdae, 0xffff, 0xffff, 
0x8a1, 0xb15, 0x107, 0x7af, 0x1185, 0x12fc, 0xffff, 0xffff, 
0xffff, 0xffff, 0x4d1, 0x5e1, 0x961, 0xffff, 0xffff, 0xbc4, 
0x1155, 0x119e, 0xffff, 0xffff, 0x208, 0x1ab, 0xffff, 0xffff, 
0x119b, 0xffff, 0x525, 0xffff, 0xffff, 0xffff, 0xffff, 0x75, 
0xb7, 0xbde, 0xa98, 0xffff, 0xffff, 0xffff, 0xffff, 0x1231, 
0xffff, 0x9d, 0xffff, 0x1240, 0xffff, 0xffff, 0xffff, 0xd21, 
0xffff, 0xffff, 0xffff, 0xffff, 0xff1, 0x82c, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xfa9, 0xffff, 0x7a0, 
0xffff, 0x2c, 0xffff, 0xffff, 0x601, 0x630, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0x5e0, 0xffff, 0x8b5, 0xf19, 0xffff, 0x542, 0xffff, 
0x632, 0x754, 0xffff, 0xffff, 0xffff, 0xffff, 0x1079, 0xffff, 
0x72f, 0xffff, 0x772, 0xffff, 0xe57, 0xffff, 0x7d3, 0xffff, 
0xffff, 0x9ea, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0x46f, 0x9c8, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0x541, 0xffff, 0x953, 0xffff, 0x7e7, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0x130d, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0x9e1, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0x205, 0xd0c, 0xffff, 0xffff, 0xffff, 0x8b3, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xe61, 0xffff, 0xf00, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x12c6, 0xffff, 0xffff, 
0x876, 0xffff, 0xffff, 0x195, 0xffff, 0xffff, 0xffff, 0xffff, 
0x95f, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x234, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0x5b5, 0xffff, 0xffff, 0x1317, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x1076, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0x1f7, 0x1d8, 0xffff, 0xffff, 
0xffff, 0xffff, 0xf9d, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0x117f, 0xffff, 0xffff, 0xffff, 0x841, 
0xffff, 0x6bf, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xe3e, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xc02, 0x948, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xf66, 
0x55e, 0xffff, 0xffff, 0xffff, 0xffff, 0x10e9, 0xffff, 0x985, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xd54, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x54c, 
0x6a3, 0xffff, 0xb24, 0x1e8, 0x552, 0xabc, 0xffff, 0xffff, 
0xffff, 0xffff, 0x6e, 0x856, 0x194, 0xacb, 0x1184, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0x101f, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xce1, 0xffff, 0xffff, 0xffff, 0xffff, 
0x346, 0xffff, 0xffff, 0xb14, 0xffff, 0x10b9, 0xffff, 0xffff, 
0x122d, 0xffff, 0xffff, 0xffff, 0x7fa, 0x11aa, 0xffff, 0xffff, 
0x14d, 0xffff, 0xffff, 0x191, 0xffff, 0xffff, 0xffff, 0xb2c, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0x8be, 0x10dd, 0x438, 0xffff, 0x147, 0x1019, 0x76a, 0x8c4, 
0xc06, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0x107b, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0x34f, 0x9a6, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0x2fb, 0x8c7, 0xffff, 0xffff, 
0xffff, 0xffff, 0x5c3, 0xffff, 0xffff, 0x113b, 0xffff, 0xffff, 
0xaa7, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x725, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x8c0, 0xffff, 0xffff, 
0xffff, 0xffff, 0x56c, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xa7f, 0x49e, 0xffff, 0xffff, 0xffff, 0x10f9, 
0xffff, 0xffff, 0xffff, 0xdd9, 0xffff, 0xffff, 0xffff, 0x24e, 
0xffff, 0xffff, 0x10a5, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0x19a, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x10f6, 0xffff, 
0xffff, 0xffff, 0x914, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0x1f6, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x526, 
0xffff, 0xffff, 0x19, 0xfc1, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xd6, 0xffff, 0xffff, 0xffff, 0x7e8, 
0x977, 0x106f, 0xffff, 0xffff, 0x822, 0xffff, 0x2e, 0xffff, 
0xffff, 0x9c3, 0x12df, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0x88f, 0x1f2, 0xe7a, 0xffff, 0x7e5, 0xffff, 0xffff, 0x482, 
0xffff, 0xffff, 0x306, 0xfce, 0x121d, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0x706, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xd38, 
0x39d, 0xffff, 0xffff, 0x9d7, 0xffff, 0xffff, 0xffff, 0xffff, 
0x3d8, 0x66d, 0xffff, 0xa81, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0x3a2, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xd55, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xa52, 0xe45, 0xcad, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0x2f7, 0xffff, 0xffff, 0xffff, 
0xffff, 0x1209, 0xffff, 0x705, 0xffff, 0xffff, 0xffff, 0x9e7, 
0xffff, 0xffff, 0xffff, 0xb9, 0xffff, 0x10d8, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xf95, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0x10e8, 0xffff, 0x3b6, 0x3f9, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xad7, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xa4d, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xfdd, 
0xffff, 0xffff, 0x220, 0x726, 0xc49, 0xffff, 0x2bc, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0x350, 0xffff, 0xffff, 0x52d, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xe7e, 0xd08, 0xffff, 
0x847, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0x9fb, 0xffff, 0x5ff, 0xffff, 0xffff, 
0xffff, 0x1123, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0x925, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xd3c, 0xffff, 0x2dc, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0x449, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x6f6, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0x77e, 0xffff, 0x1e4, 0xffff, 0xb6d, 0xffff, 0xffff, 0x6aa, 
0xefe, 0xffff, 0xffff, 0xffff, 0xdb8, 0xc9a, 0xd13, 0xffff, 
0xffff, 0xffff, 0xfa2, 0xffff, 0xffff, 0xc57, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0x57a, 0xffff, 0xffff, 0xffff, 0x155, 0xffff, 0x960, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xcc7, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0x470, 0x102d, 0xffff, 0xffff, 0xffff, 
0xffff, 0x1215, 0xffff, 0xffff, 0x84b, 0xffff, 0xffff, 0xffff, 
0x111d, 0xffff, 0xffff, 0xffff, 0xe21, 0xfe7, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xd46, 0xffff, 0xffff, 0xffff, 0xb42, 
0xffff, 0xffff, 0xffff, 0xffff, 0x989, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x2a9, 0xffff, 0xffff, 
0x15a, 0xffff, 0xdd7, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xc6b, 0x12d2, 0xffff, 0xffff, 0xe9d, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0x1aa, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0x3fa, 0xffff, 0xffff, 0xffff, 0x9a8, 0xffff, 0xffff, 
0xc15, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xfee, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0x6b7, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0x42e, 0xffff, 0xc0f, 0xffff, 0xffff, 0xb6e, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xdc7, 0xbc8, 
0xffff, 0xffff, 0xffff, 0xffff, 0x96, 0x108b, 0xc8c, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xe2f, 0xffff, 
0xffff, 0xffff, 0xffff, 0x3aa, 0x12ab, 0x65a, 0xca5, 0xffff, 
0xffff, 0xffff, 0xf72, 0xffff, 0xffff, 0xffff, 0xffff, 0x3b3, 
0xffff, 0xffff, 0xffff, 0xffff, 0xcb3, 0x275, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x3b7, 
0xffff, 0xffff, 0xffff, 0x31d, 0xe2b, 0xa3c, 0x12a3, 0x121f, 
0xffff, 0xffff, 0xffff, 0x621, 0x741, 0x10d2, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x7b5, 0xffff, 0xc42, 
0xffff, 0xffff, 0x5b1, 0x139, 0x236, 0xffff, 0xece, 0xffff, 
0x945, 0xd3, 0x343, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0x4dc, 0xffff, 0xffff, 0xffff, 
0x9c6, 0xffff, 0xffff, 0x3ef, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x8fd, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x24d, 
0xffff, 0xffff, 0xffff, 0x14e, 0xffff, 0xe3c, 0xffff, 0xffff, 
0xffff, 0x4b0, 0xb2b, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xbf7, 0xffff, 0x418, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xf87, 0xffff, 0xffff, 0x1244, 0xffff, 0xffff, 0xffff, 
0xfc6, 0x9b7, 0xdc2, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0x22f, 0xffff, 0x44a, 0xffff, 0x5e8, 0xffff, 0xffff, 
0xffff, 0xffff, 0x255, 0x7c1, 0xffff, 0xeee, 0x112b, 0x1275, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x34d, 0xffff, 0xffff, 
0xffff, 0xffff, 0x90, 0xffff, 0xffff, 0xffff, 0x2f3, 0xffff, 
0xffff, 0xffff, 0xffff, 0x10ae, 0xffff, 0x53e, 0x12c9, 0x101a, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0x986, 0xb08, 0xffff, 0xffff, 0xffff, 0x5d1, 0xffff, 
0x5da, 0xffff, 0xffff, 0xffff, 0x17f, 0xffff, 0xffff, 0xffff, 
0x19c, 0xffff, 0x696, 0x1a2, 0x4e5, 0x1083, 0xffff, 0xb37, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x1110, 0xffff, 
0xffff, 0xffff, 0x604, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xab6, 0xae3, 0x976, 0xffff, 0xfd4, 0x5a, 0xffff, 
0xa0e, 0xffff, 0xffff, 0xee2, 0xffff, 0xffff, 0xffff, 0x515, 
0xffff, 0xffff, 0xcbb, 0xffff, 0xffff, 0xffff, 0xffff, 0xc9f, 
0xffff, 0xffff, 0x119d, 0xffff, 0xffff, 0xffff, 0xa80, 0xffff, 
0x10ea, 0xffff, 0xffff, 0xffff, 0x6ab, 0xffff, 0xffff, 0xffff, 
0xffff, 0xdd2, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0x1274, 0xffff, 0xffff, 0x9d3, 0x207, 0xffff, 0xed4, 
0xffff, 0xffff, 0x499, 0xb0a, 0x4ec, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0x271, 0xffff, 0x104c, 0xffff, 0xffff, 
0xffff, 0x7f7, 0xdff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xdfa, 0xffff, 0xffff, 0xffff, 0x1188, 0xffff, 0x104e, 
0x830, 0xffff, 0xffff, 0x24c, 0x942, 0xa0, 0xa2f, 0xffff, 
0xffff, 0xffff, 0xa8c, 0xffff, 0x1307, 0xffff, 0xf6, 0xffff, 
0xffff, 0xffff, 0xffff, 0x108, 0xffff, 0x8f3, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x27d, 0xffff, 
0xffff, 0xffff, 0xffff, 0x254, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0x5b8, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x545, 
0xffff, 0xffff, 0xffff, 0xffff, 0x30d, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xf2d, 0xffff, 0xffff, 0x9f, 0x2c2, 0x295, 
0x12c3, 0xffff, 0xffff, 0x38f, 0x669, 0xffff, 0xffff, 0xffff, 
0x6fa, 0xffff, 0xffff, 0xffff, 0xffff, 0x4f7, 0xffff, 0xffff, 
0xffff, 0x34b, 0x6b4, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xd5f, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xf44, 0xde5, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xf99, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x1099, 0xffff, 
0xffff, 0xffff, 0xffff, 0x728, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0x1033, 0xffff, 0x99e, 0x18, 
0x163, 0x1201, 0xffff, 0xffff, 0x55a, 0x43a, 0xa10, 0xffff, 
0xffff, 0xffff, 0xdcd, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0x113e, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0x510, 0x39f, 0xffff, 0xffff, 0xffff, 0x100a, 0xf, 0xffff, 
0xffff, 0xffff, 0x760, 0xffff, 0xffff, 0xffff, 0xb63, 0xffff, 
0xffff, 0xffff, 0x6bc, 0x1101, 0xffff, 0xffff, 0xffff, 0xd53, 
0xffff, 0xffff, 0xffff, 0x3d7, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x107e, 0xffff, 
0xffff, 0xffff, 0xffff, 0x210, 0xffff, 0x4e8, 0xaf, 0xb7d, 
0x507, 0x946, 0xc39, 0xe32, 0xf93, 0x10a9, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xb8f, 
0xffff, 0x627, 0x661, 0xffff, 0xffff, 0xffff, 0x2ee, 0xffff, 
0xffff, 0xeba, 0x563, 0xffff, 0x417, 0xdb5, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x1179, 0xffff, 
0xffff, 0x41b, 0xffff, 0xffff, 0xffff, 0x1232, 0x6be, 0xffff, 
0xffff, 0xffff, 0xffff, 0x9dc, 0x1242, 0xd95, 0xffff, 0xffff, 
0xffff, 0x10b6, 0xffff, 0x7d6, 0xffff, 0xffff, 0x45, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0x88, 0xffff, 0xffff, 0xffff, 
0xffff, 0x6e1, 0xc14, 0xffff, 0xffff, 0xffff, 0x50d, 0xaa1, 
0xae9, 0x1293, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0x12c4, 0x126a, 0x12d8, 0x67f, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0x461, 0xffff, 0x638, 0xffff, 0x1bc, 0x8bd, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0x5e, 0xffff, 0x105f, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xd5b, 0x10a4, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0x1050, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xc5f, 0xffff, 0xffff, 0xffff, 0xf9, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xc28, 0xffff, 0xffff, 0x12d4, 0xe7, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xa7, 0x413, 0xbab, 0xffff, 0xffff, 0x91, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xc1e, 0xffff, 0xffff, 0xffff, 0xffff, 0xd0, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x197, 0x69c, 0xffff, 
0xffff, 0xcca, 0x86b, 0xffff, 0x1216, 0x12d3, 0xffff, 0xffff, 
0x6dd, 0x8d8, 0xceb, 0x3bf, 0x257, 0x12a1, 0xffff, 0xffff, 
0x8de, 0xffff, 0xffff, 0x1119, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x69b, 
0xbfe, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x1f, 0xffff, 
0xffff, 0xd8b, 0xffff, 0xffff, 0xffff, 0xe5e, 0xffff, 0xffff, 
0x11f0, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x1316, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0x303, 0xb68, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xda1, 0xffff, 0xffff, 0x162, 0xd88, 0x5ad, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0x2d6, 0x729, 0x11fe, 0xffff, 0xffff, 0x30b, 
0x1295, 0x12e7, 0xffff, 0xfb6, 0xffff, 0x8e5, 0xffff, 0x13c, 
0xffff, 0xffff, 0xffff, 0xffff, 0xcf8, 0xffff, 0xffff, 0xd56, 
0x97c, 0x1103, 0xffff, 0xc2b, 0x12, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0x125, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0x457, 0x8e0, 0x6d2, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x1296, 0xe6b, 
0x1261, 0xffff, 0xe1d, 0xffff, 0xae8, 0xffff, 0xffff, 0xffff, 
0x2a0, 0xffff, 0xffff, 0xffff, 0xffff, 0x63b, 0xffff, 0xffff, 
0xffff, 0xffff, 0x277, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xd25, 0xffff, 0xffff, 0x107a, 0xffff, 0x40d, 0x10bd, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0x5c4, 0xffff, 0xffff, 0xc62, 
0xffb, 0x10f2, 0xdb7, 0xffff, 0xffff, 0xffff, 0xffff, 0x462, 
0xffff, 0x98a, 0x184, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xbb1, 0x2e3, 0x765, 0x884, 0xffff, 0xffff, 
0xffff, 0xbb8, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xc83, 
0xe81, 0xffff, 0x90c, 0xd0d, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0x95d, 0x7d7, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0x1da, 0xffff, 0xcb6, 0x123f, 0x108d, 0xffff, 
0xffff, 0xffff, 0xffff, 0x6a6, 0xc95, 0xffff, 0xd3e, 0xffff, 
0xffff, 0xffff, 0x608, 0xffff, 0x10ce, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0x1007, 0x5, 0x52e, 0xffff, 0xd4e, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xf29, 0xffff, 
0xffff, 0xffff, 0xffff, 0x258, 0x12fb, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xe87, 
0xf80, 0x12bd, 0x828, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xcd4, 0xffff, 0x274, 0xffff, 0xffff, 0xc1d, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x4fc, 0xb27, 0xffff, 
0xffff, 0xffff, 0xffff, 0x17a, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xff8, 0xffff, 0xffff, 0xffff, 0xffff, 0x957, 
0x9ef, 0x80f, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x88a, 0x653, 0x938, 
0xbe8, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xfe, 
0xffff, 0xc64, 0xffff, 0x501, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xb55, 0xffff, 0x11db, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xa13, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0x25e, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0x2a6, 0xffff, 0xffff, 0xffff, 0x1a6, 0xe6f, 0x10d9, 
0xffff, 0xffff, 0xff7, 0x11be, 0xffff, 0xffff, 0xffff, 0xe8d, 
0xffff, 0x5d3, 0x101c, 0xffff, 0x369, 0xffff, 0xffff, 0xffff, 
0xffff, 0xaf2, 0xffff, 0xffff, 0xffff, 0xc23, 0x55c, 0xffff, 
0xffff, 0x1ef, 0xffff, 0x629, 0xffff, 0xffff, 0x183, 0xffff, 
0xffff, 0x8ab, 0xffff, 0xffff, 0xffff, 0xffff, 0x3a3, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xc30, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0x109e, 0xffff, 0xffff, 0xffff, 0xc19, 0xffff, 0xffff, 
0xffff, 0x113, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0x472, 0x89b, 0xce0, 0xffff, 0xffff, 0x43c, 0xffff, 0xde2, 
0x1256, 0xffff, 0xb3b, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0x798, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x714, 
0xffff, 0xffff, 0x67d, 0xffff, 0xffff, 0xffff, 0x407, 0xffff, 
0xffff, 0xafd, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xf14, 
0xffff, 0xffff, 0x102, 0x11d1, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0x52f, 0x9fe, 0x3b9, 0xffff, 0x69f, 0xffff, 
0x12b6, 0xffff, 0xffff, 0xffff, 0x1092, 0xffff, 0xc11, 0xffff, 
0xffff, 0xade, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xc3b, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x96b, 
0xa94, 0x11f6, 0x466, 0x59d, 0xf08, 0xffff, 0x2f8, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x650, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xe47, 0x10c2, 0xb36, 0xffff, 0xffff, 0xbb2, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x951, 0xcc5, 
0xffff, 0xffff, 0x104f, 0xffff, 0xffff, 0x950, 0xffff, 0xffff, 
0x3a1, 0x497, 0xffff, 0xea0, 0xffff, 0x1a, 0xffff, 0xffff, 
0x11e0, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xa5e, 0xffff, 0xffff, 0x923, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0x4d8, 0x51c, 0xc72, 0xffff, 
0xffff, 0xffff, 0xffff, 0x7dc, 0x12de, 0x5ae, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xa7c, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x5fc, 0x11a8, 0xffff, 
0x112a, 0xffff, 0xffff, 0x4b3, 0xffff, 0xffff, 0x117c, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0x536, 0x670, 0xffff, 0x753, 0x283, 0xefc, 0x127d, 0x324, 
0xffff, 0xffff, 0xffff, 0xffff, 0x692, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0x6e8, 0xfca, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x99a, 0xffff, 
0xd5, 0xffff, 0xd27, 0xd40, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0x58d, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xbf1, 0x610, 0xda4, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xbe3, 0xffff, 0xffff, 0xffff, 0xffff, 
0x33a, 0xffff, 0xffff, 0x4a0, 0xffff, 0xffff, 0xffff, 0xa70, 
0x26d, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0x3c2, 0xffff, 0x490, 0xcff, 0xffff, 0xffff, 
0x7b8, 0xffff, 0xe31, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0x440, 0xffff, 0xd91, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0x480, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0x1a8, 0x55, 0xedc, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xbf3, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0x12b2, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xfe5, 0x1241, 0xffff, 0xffff, 0x137, 0x13a, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x1147, 0xffff, 
0xffff, 0x25, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0x3f2, 0xffff, 0xe67, 0xffff, 0xffff, 0xffff, 0xffff, 0x8, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0x124f, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x360, 0x1222, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0x1be, 0xdc, 0x76c, 0xd70, 0xffff, 
0x6e2, 0xffff, 0xffff, 0xffff, 0x55d, 0xffff, 0xffff, 0xffff, 
0xf7, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xdc0, 
0xffff, 0xba3, 0xffff, 0x2e1, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xea6, 0x126e, 0x70a, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0x597, 0xffff, 0xffff, 0xffff, 0xffff, 0xa57, 0x615, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x12d1, 0xa8a, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xb8b, 
0xffff, 0xffff, 0x28c, 0x288, 0xffff, 0xffff, 0xffff, 0x329, 
0xffff, 0xffff, 0xffff, 0xffff, 0x3bd, 0xffff, 0x340, 0xffff, 
0x63f, 0xffff, 0xffff, 0x1129, 0x152, 0xbdd, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xf88, 0xffff, 0x9e2, 0xffff, 
0xffff, 0xffff, 0xffff, 0xd0f, 0x1304, 0xc7f, 0xffff, 0xffff, 
0xffff, 0xffff, 0x740, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xd86, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xf6b, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xf96, 0xffff, 
0x5ea, 0xcc3, 0xffff, 0xffff, 0xffff, 0xffff, 0x746, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xb5, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0x94, 0xffff, 0xffff, 0xffff, 0x12f0, 0xffff, 0xffff, 0xffff, 
0xffff, 0x54e, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xdf6, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0x88e, 0xffff, 0xffff, 0xffff, 0xffff, 0x99b, 
0xffff, 0x1186, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x815, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0x1d9, 0xffff, 0x323, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0x843, 0xffff, 0xd8, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0x23a, 0xffff, 0xffff, 0x10f4, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x253, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0x1203, 0xee6, 0x128d, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0x12a5, 0xe54, 0xffff, 0xffff, 0xffff, 
0x10d3, 0xffff, 0x16d, 0xffff, 0xffff, 0x2be, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0x60d, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xa65, 0xffff, 0xbdc, 0xffff, 0xffff, 
0xffff, 0x550, 0xffff, 0xffff, 0xffff, 0xffff, 0x59a, 0x12f6, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xb35, 0x39e, 0xffff, 0xffff, 0x5ee, 0xffff, 0xa31, 0x62, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0x996, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xec3, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xc9, 0xfab, 0xffff, 0xd28, 0x1ea, 0xf16, 
0xd01, 0xffff, 0xffff, 0xffff, 0xffff, 0xda3, 0xffff, 0xffff, 
0xffff, 0x2, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0x4c3, 0xffff, 0xffff, 0x628, 0xb84, 0x51d, 0x118, 0xffff, 
0xffff, 0xffff, 0x7b7, 0xffff, 0xffff, 0x106e, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0x641, 0xffff, 0xffff, 0xdeb, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0x17, 0xffff, 0xffff, 0xffff, 
0xffff, 0x1d3, 0x853, 0x8ea, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x939, 
0x755, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xa25, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0x22d, 0x85a, 0xffff, 0xffff, 0xffff, 0xffff, 0xca, 
0xffff, 0xb64, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0x906, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0x3f1, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x5f, 0x614, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xbdb, 
0xffff, 0xffff, 0x82d, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0x62d, 0xe4d, 0x136, 0xffff, 0xffff, 0x70f, 0x5dc, 0xffff, 
0xffff, 0xffff, 0xffff, 0x10ff, 0xffff, 0xffff, 0x519, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x595, 0xffff, 
0xf8f, 0xffff, 0xffff, 0xffff, 0xb3e, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0x75a, 0xffff, 0x9ba, 0xffff, 0xffff, 0xffff, 0xe3b, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x23e, 0xffff, 0xcf2, 
0xffff, 0x1034, 0xffff, 0xffff, 0x3b2, 0xffff, 0xffff, 0x349, 
0xffff, 0xffff, 0x9f1, 0x11ef, 0xffff, 0x36c, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x967, 0xffff, 0xffff, 
0xffff, 0xb25, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xfeb, 0xffff, 0xffff, 0xffff, 0xffff, 0xbeb, 0x1315, 0xffff, 
0x21a, 0xffff, 0xffff, 0x86f, 0xd81, 0xffff, 0xd22, 0xffff, 
0xb67, 0xffff, 0xba1, 0xffff, 0x9f8, 0xffff, 0x6f1, 0x7, 
0xffff, 0xffff, 0xcfc, 0x991, 0xc1c, 0x912, 0x12ee, 0xffff, 
0x4fa, 0xf94, 0x2a2, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0x14f, 0x2c5, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0x382, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xf6c, 0xffff, 
0xffff, 0x9be, 0xffff, 0xffff, 0xffff, 0x1294, 0xffff, 0xc13, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x930, 0xffff, 0xffff, 
0xffff, 0xffff, 0x782, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xbad, 0xffff, 0xffff, 0x2b1, 0xffff, 0x826, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xf51, 0x10d5, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0x596, 0xac6, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x82a, 0xffff, 0xffff, 
0xffff, 0xcdd, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xa1c, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0x397, 0xabb, 0xf23, 0xffff, 0xffff, 0xa06, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x33c, 0xffff, 0x1b2, 
0xd96, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0x414, 0xbd0, 0xffff, 0xffff, 0xffff, 0xd64, 0xffff, 
0xffff, 0x11cc, 0xffff, 0x211, 0x658, 0xf7c, 0xffff, 0xffff, 
0x36f, 0xffff, 0x43f, 0x4cf, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xcee, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0x803, 0xffff, 0xffff, 0x1e7, 0xffff, 0xffff, 0xffff, 
0xffff, 0x514, 0xffff, 0xffff, 0xffff, 0x8b2, 0xffff, 0x8ec, 
0xffff, 0xffff, 0xffff, 0xffff, 0xc91, 0xffff, 0xc2, 0xffff, 
0x1e9, 0xffff, 0x47f, 0xffff, 0xffff, 0xffff, 0x11d2, 0xffff, 
0xffff, 0xffff, 0x4bf, 0xe09, 0x73e, 0x12c5, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x383, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0x310, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0x37c, 0xffff, 0x30, 0xcce, 0xffff, 0xffff, 
0xffff, 0x1b, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0x968, 0xffff, 0x1065, 0xffff, 0xffff, 0xebe, 
0xffff, 0xffff, 0x6c4, 0x862, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0x1168, 0xffff, 0x8ff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x7cc, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0x281, 0xffff, 0xffff, 0xffff, 0xffff, 
0xe90, 0xffff, 0xffff, 0xffff, 0xffff, 0x31e, 0x74f, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xd06, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xf67, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x1f8, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xcc6, 0xffff, 0xffff, 
0xffff, 0xffff, 0xa43, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0x1032, 0xb10, 0x1167, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0x704, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xea, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xb99, 0xffff, 0xffff, 0xffff, 0xffff, 0x918, 
0xb06, 0xd04, 0xba6, 0xf2c, 0x1169, 0x1238, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0x7f0, 0xc5a, 0xffff, 0xffff, 
0x19e, 0xffff, 0xffff, 0xffff, 0xa39, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0x711, 0xffff, 0xffff, 0xffff, 0xc3f, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x384, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0x129, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x9a1, 
0xffff, 0x1280, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0x100, 0xffff, 0x10f0, 0xd5a, 0xffff, 0x1248, 0xffff, 0x9fa, 
0xb6c, 0x693, 0xffff, 0xffff, 0xffff, 0xffff, 0x115d, 0xffff, 
0xffff, 0xffff, 0xd42, 0xffff, 0xffff, 0xffff, 0xdee, 0xffff, 
0xffff, 0xffff, 0xffff, 0xe8b, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xa60, 0xffff, 0xffff, 0xfcc, 0x1254, 0x4dd, 0x10cd, 
0x4f5, 0xb09, 0x29f, 0xffff, 0xffff, 0xffff, 0xffff, 0x5b6, 
0x55f, 0xffff, 0xffff, 0x15c, 0xffff, 0xffff, 0xd66, 0xffff, 
0xffff, 0xffff, 0xffff, 0xcb9, 0xffff, 0xffff, 0x920, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x6a9, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0x806, 0xffff, 0xa73, 0xffff, 0xffff, 
0xe93, 0xffff, 0x6ef, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0x10fd, 0xffff, 0xffff, 0x10cc, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xa34, 0xffff, 0x97f, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0x1127, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xc1f, 
0xffff, 0xf15, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xd49, 
0xffff, 0xffff, 0xffff, 0x51e, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0x3c4, 0xffff, 0xffff, 0xffff, 0xd57, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0x11d0, 0x1077, 0x233, 0xa9c, 0x907, 
0xc7a, 0xbaa, 0xd76, 0xffff, 0xffff, 0xffff, 0xffff, 0xf24, 
0xffff, 0xffff, 0xffff, 0xffff, 0xfa8, 0xffff, 0xffff, 0xffff, 
0x6c5, 0x73a, 0xffff, 0x1211, 0xffff, 0xffff, 0xffff, 0xffff, 
0x1205, 0xffff, 0xffff, 0xffff, 0xffff, 0x6df, 0x793, 0xffff, 
0xffff, 0xffff, 0xffff, 0xe89, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0x35b, 0x7a2, 0xffff, 0xffff, 0xffff, 
0xffff, 0x6a5, 0xffff, 0xffff, 0xffff, 0xc6, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xc70, 0x1047, 0xf06, 0xffff, 
0xffff, 0xffff, 0xffff, 0x42, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0x82b, 0xffff, 0xffff, 0xffff, 0x589, 0x533, 0xffff, 
0xffff, 0xffff, 0xe82, 0xffff, 0xffff, 0xffff, 0xffff, 0x272, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0x7cd, 0xffff, 0xffff, 0xffff, 0xffff, 0xa42, 0xffff, 0xffff, 
0x146, 0xffff, 0xffff, 0xeb7, 0xffff, 0xffff, 0x1102, 0xffff, 
0xffff, 0xffff, 0xff2, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0x2a7, 0xffff, 0xffff, 0xffff, 0xffff, 0x620, 0xffff, 0xffff, 
0xf81, 0xffff, 0xffff, 0x140, 0xffff, 0xde6, 0xc4f, 0x65d, 
0x819, 0xffff, 0x902, 0xffff, 0xffff, 0xffff, 0xffff, 0x12dd, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xae7, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xa14, 0xffff, 0x549, 
0xffff, 0xffff, 0x5f0, 0xffff, 0xffff, 0xffff, 0xffff, 0x1002, 
0xffff, 0xa18, 0xb73, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0x348, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0x651, 0xffff, 0xffff, 0xd93, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0x87c, 0xb6a, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xa59, 0xc9d, 0x100b, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x1d4, 0xffff, 
0xffff, 0xffff, 0x4ac, 0xd44, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0x133, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x126f, 0xffff, 
0xffff, 0x55b, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0x406, 0x1265, 0xffff, 0xffff, 0xffff, 0xffff, 
0xe76, 0x10b0, 0xfb9, 0xffff, 0xff9, 0xffff, 0x7ff, 0x1312, 
0xffff, 0xffff, 0xffff, 0xffff, 0x16c, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0x53f, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xd89, 0xffff, 0x663, 0xffff, 0x881, 0xffff, 0xffff, 
0xffff, 0xffff, 0xcea, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0x6ce, 0xffff, 0x981, 0xf45, 0xffff, 0xffff, 0xffff, 
0xffff, 0xfcb, 0xdd0, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0x244, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0x1157, 0xffff, 0xffff, 0x97, 
0x4f0, 0x11cb, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xd1b, 0xffff, 0x8f5, 0xffff, 0xe13, 0xc18, 
0xffff, 0xffff, 0xffff, 0xa44, 0xffff, 0x396, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xb17, 0xc9e, 0xffff, 
0xffff, 0xffff, 0x3e9, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xfa, 
0xaae, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x9e3, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xb83, 
0xa0c, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xb46, 0xffff, 0xffff, 0x1fa, 
0x32e, 0xffff, 0xffff, 0xffff, 0x969, 0xffff, 0xffff, 0xffff, 
0xffff, 0x8d7, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xf70, 0xf8a, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0x4af, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xe0, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x11f9, 0x4d4, 
0x7cf, 0x1d0, 0xffff, 0xffff, 0x279, 0x1012, 0xffff, 0xffff, 
0xffff, 0xffff, 0xaa2, 0x15e, 0x179, 0xd4b, 0xffff, 0xffff, 
0xffff, 0xffff, 0xc04, 0x3e6, 0xffff, 0xffff, 0xe33, 0xffff, 
0xffff, 0x4be, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x895, 0x18f, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x89f, 
0xf28, 0xffff, 0x85b, 0xde, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0x4f8, 0x1061, 0xffff, 0xffff, 0xffff, 0x435, 0xffff, 
0xffff, 0xf1b, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x5c2, 
0x811, 0xbfc, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x64d, 0xffff, 
0x44, 0xffff, 0xffff, 0xd73, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0x10b4, 0xffff, 0xffff, 0x40f, 0xffff, 0xffff, 0xffff, 
0xbaf, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0x1128, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0x7da, 0xffff, 0xffff, 0xffff, 0xffff, 0x3da, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xf85, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xdd, 0x5a1, 0xffff, 0xffff, 0xffff, 0xffff, 0x335, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x3d3, 0xe77, 
0xffff, 0xffff, 0xffff, 0x6f9, 0xffff, 0xffff, 0xca0, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xc25, 0xffff, 0xffff, 
0x448, 0xffff, 0xe4b, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0x3de, 0xffff, 0xffff, 0x58c, 0xffff, 0xffff, 0x3d, 
0x7ca, 0xffff, 0xffff, 0x1028, 0xffff, 0xffff, 0x60c, 0xffff, 
0xffff, 0xf60, 0x1036, 0xffff, 0xffff, 0xffff, 0xffff, 0x807, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xd2e, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xdc5, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x84f, 
0xffff, 0xffff, 0x1a0, 0xffff, 0x1144, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xbbc, 0x280, 0xffff, 
0x361, 0x1059, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x823, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0x745, 0x10db, 0xffff, 0xffff, 0x3fd, 0xffff, 
0xffff, 0xffff, 0x4d6, 0xffff, 0xffff, 0xeda, 0xffff, 0xffff, 
0x4e0, 0xffff, 0xffff, 0xbac, 0xffff, 0x313, 0xffff, 0x5b2, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x412, 0xffff, 
0xffff, 0xffff, 0xc09, 0xffff, 0xffff, 0xffff, 0xffff, 0x393, 
0xd7d, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x7dd, 0xffff, 
0xffff, 0x6c3, 0xffff, 0x64b, 0xff, 0xffff, 0xffff, 0xffff, 
0x56, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0x637, 0xffff, 0xc77, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0x4db, 0x87f, 0xffff, 0xffff, 
0x276, 0x14a, 0xffff, 0x69, 0xffff, 0x49a, 0xf1, 0xc51, 
0xffff, 0xffff, 0xafc, 0x1252, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0x12b4, 0xffff, 0xffff, 0xffff, 0xffff, 
0x22a, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0x1111, 0xffff, 0xffff, 0xffff, 0xecc, 0xffff, 0xffff, 
0x8a5, 0x12ec, 0x1146, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0x10f5, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0x6d4, 0xb62, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0x1fb, 0x24, 0xffff, 0x776, 
0xffff, 0xffff, 0xa16, 0xffff, 0xffff, 0xffff, 0xffff, 0x185, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xacf, 0xffff, 0xffff, 0xeb2, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x717, 0xffff, 
0xffff, 0xb91, 0xffff, 0x289, 0x91f, 0x72d, 0xc92, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xb7e, 0xffff, 0x252, 0x95c, 0x100f, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0x562, 0xaa8, 0xf18, 0x111b, 0x1298, 0xffff, 0xffff, 0xffff, 
0xffff, 0x11c4, 0x674, 0xffff, 0xffff, 0xffff, 0x336, 0xffff, 
0xffff, 0xba5, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0x61a, 0xacd, 0xf5d, 0xffff, 0xffff, 0xd7e, 0xddd, 0xffff, 
0xffff, 0xbe, 0x68f, 0x1160, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0x1e2, 0xffff, 0x3fc, 0x65c, 0x1259, 0xf4a, 0xffff, 
0x328, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0x1247, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x109c, 
0xffff, 0x1055, 0xffff, 0xffff, 0x11e7, 0xc9b, 0x113f, 0xffff, 
0xffff, 0x42c, 0xffff, 0xffff, 0x11dc, 0xffff, 0xe38, 0xffff, 
0xffff, 0xff0, 0x10b3, 0x1314, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xa11, 0x97a, 0xfff, 
0x1269, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xc89, 0xffff, 
0xffff, 0x395, 0xbca, 0xffff, 0x12a7, 0x757, 0xeae, 0x119c, 
0xffff, 0x805, 0x645, 0x105, 0xffff, 0xffff, 0x1309, 0xffff, 
0xffff, 0xbe4, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x1137, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xd34, 0xffff, 0x26, 0xffff, 0xffff, 0x61c, 0xffff, 
0xffff, 0xffff, 0x543, 0xeeb, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0x4aa, 0xffff, 0xffff, 0x4e2, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0x83f, 0x66b, 0xffff, 0xe64, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x489, 0x7f8, 0xffff, 
0xffff, 0xffff, 0x7f, 0xffff, 0xffff, 0xffff, 0xc38, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xac, 0x1289, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0x1120, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xa9e, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0x785, 0xffff, 0x2d3, 0x551, 0x94e, 0xffff, 0xf68, 
0xddc, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xca8, 0xffff, 0xffff, 0x0, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0x353, 0xffff, 0xffff, 0xabd, 0x6db, 0xffff, 
0xffff, 0x3dc, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xcc8, 0xf92, 0x223, 0xffff, 0x7f1, 
0xffff, 0xffff, 0xffff, 0xffff, 0xcd3, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xf41, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xec5, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0x824, 0x101b, 0xffff, 0xffff, 0xffff, 0xb6f, 0xcd2, 0x347, 
0x357, 0xffff, 0xffff, 0xc58, 0xffff, 0xffff, 0x427, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0x681, 0x34c, 0x7db, 0x2a1, 
0xf50, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x7fc, 0xffff, 
0x128, 0xffff, 0x6c1, 0xa49, 0x28a, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xfbe, 0xffff, 0x9b4, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0x1fc, 0xffff, 0xffff, 0x1093, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xed, 0xffff, 0xaee, 0xffff, 0xffff, 0x57, 0xffff, 
0xffff, 0xa4b, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0x126b, 0xff5, 0x504, 0x1258, 0xf7d, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x84c, 
0xffff, 0xffff, 0xffff, 0xffff, 0x6fb, 0xb93, 0x1194, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0x3c6, 0xffff, 0x52b, 0x5ed, 0x6ed, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x6e7, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xed7, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0x453, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xe7b, 0x910, 0xffff, 0xffff, 0xffff, 0xffff, 
0xc0a, 0xdd1, 0xffff, 0x743, 0xffff, 0xe6a, 0x613, 0xa71, 
0xaac, 0xfea, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xc54, 0xffff, 0xe08, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xf62, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0x5c, 0x1a1, 0x105e, 0x10fa, 0x97d, 0xf5a, 0x76f, 0x12ed, 
0x724, 0xffff, 0xffff, 0xffff, 0x846, 0x9e4, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0x49c, 0xcec, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xce5, 0x6b8, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x268, 0xffff, 
0xffff, 0xffff, 0x972, 0xffff, 0xffff, 0xffff, 0xffff, 0x7c6, 
0xffff, 0xffff, 0x4bc, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0x456, 0x889, 0x5f2, 0xffff, 0x87d, 0xffff, 0xffff, 0xffff, 
0xa55, 0x1067, 0x572, 0x1220, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0x44d, 0xffff, 0xffff, 0x720, 0xffff, 0xffff, 0xffff, 
0xf3e, 0x88b, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0x104d, 0xffff, 0xffff, 0x1fd, 0x9b8, 0x8f, 0x11fb, 0x446, 
0x2eb, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xe22, 0xffff, 0xffff, 0xffff, 0xffff, 0x118f, 
0xffff, 0x138, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0x58f, 0xffff, 0xffff, 0xffff, 0x301, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0x675, 0xffff, 0x640, 0xffff, 0xffff, 
0xffff, 0xffff, 0xbec, 0xffff, 0xffff, 0xf55, 0x50e, 0x1010, 
0xffff, 0xffff, 0xb30, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0x3ae, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xa29, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x1f0, 0x287, 
0x439, 0xd6a, 0x286, 0xffff, 0xffff, 0x9b, 0xffff, 0xffff, 
0xffff, 0x42f, 0xffff, 0xffff, 0xffff, 0xffff, 0x218, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x5aa, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xdce, 0xffff, 0x688, 0xffff, 0xffff, 0x691, 0xffff, 0xffff, 
0xffff, 0xffff, 0xd52, 0xffff, 0xde0, 0xffff, 0xffff, 0xffff, 
0x12ef, 0xffff, 0xffff, 0x619, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0x817, 0xffff, 0xffff, 0x1202, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0x6c0, 0xa63, 0x423, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xc00, 0xffff, 
0x61f, 0xff4, 0xffff, 0xffff, 0xffff, 0xedd, 0xffff, 0xffff, 
0x81b, 0xffff, 0x408, 0x5e6, 0xffff, 0xffff, 0xffff, 0xb33, 
0xffff, 0xffff, 0xffff, 0xffff, 0xde8, 0xffff, 0xffff, 0xffff, 
0xffff, 0xc43, 0x11ca, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0x2c4, 0x1143, 0x7d2, 0x1283, 0x358, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xbf0, 
0x678, 0xffff, 0xffff, 0xffff, 0xd0b, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x3ec, 
0x810, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x44f, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x441, 0x1c8, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xd14, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0x9cc, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0x3ff, 0xffff, 0xf13, 0xffff, 0xc87, 0x116d, 
0xffff, 0xdf0, 0x1288, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0x1005, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x1200, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x263, 0xffff, 0xffff, 
0x1d1, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x54b, 0x2aa, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x78, 0xffff, 
0xf36, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x10bc, 
0xffff, 0x231, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xc47, 0xffff, 0xffff, 
0xffff, 0xffff, 0x855, 0x6d0, 0xffff, 0xffff, 0xffff, 0x171, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0x602, 0xffff, 0x8ad, 0x8eb, 0xffff, 0xffff, 0x1267, 
0xffff, 0xe94, 0xffff, 0xffff, 0xd65, 0xffff, 0xcb, 0xffff, 
0xffff, 0xffff, 0x11b0, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0x373, 0xffff, 0xffff, 0xffff, 0xffff, 0x10aa, 0x2ce, 0xffff, 
0x733, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0x506, 0xfb1, 0xffff, 0xffff, 0xd7, 0xffff, 0x235, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x835, 0xffff, 
0xffff, 0x1ec, 0xffff, 0xffff, 0xffff, 0x282, 0xffff, 0x633, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x442, 0x102a, 
0x870, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0x9c2, 0xffff, 0xffff, 0xffff, 0x232, 
0x2c3, 0xffff, 0xffff, 0x12be, 0xffff, 0xffff, 0xffff, 0xc2e, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x1292, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xccf, 0x867, 
0xfa7, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xee4, 0xffff, 0xffff, 0xffff, 0x1278, 0x676, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0x56b, 0xffff, 0xffff, 0xffff, 0x998, 
0xffff, 0xffff, 0xffff, 0xffff, 0xbea, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xb40, 0xf52, 0xffff, 0x308, 
0x568, 0x2cd, 0xffff, 0x10b8, 0xffff, 0xffff, 0x6ba, 0x130e, 
0xffff, 0xffff, 0xffff, 0xffff, 0xd78, 0xffff, 0xffff, 0x428, 
0x260, 0x5e9, 0x12ff, 0xffff, 0x496, 0xffff, 0xffff, 0xffff, 
0x671, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x93a, 0xffff, 
0x322, 0xffff, 0xffff, 0xffff, 0x32d, 0xffff, 0xffff, 0xffff, 
0xffff, 0x40b, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0x919, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x730, 
0xffff, 0xffff, 0xffff, 0xffff, 0xeec, 0xffff, 0x4e7, 0xffff, 
0xffff, 0xffff, 0xffff, 0x5a9, 0xffff, 0xffff, 0xffff, 0xffff, 
0xdbd, 0xffff, 0xffff, 0xffff, 0x2fc, 0xffff, 0x53d, 0x11a, 
0x978, 0xffff, 0x100e, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xf58, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0x1130, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0x12eb, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xb19, 0xbe9, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xc1, 0x182, 0xffff, 0xd69, 0xffff, 
0x52c, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xecf, 0xffff, 0xffff, 0xb5d, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0x102e, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xc80, 
0x4ad, 0xffff, 0xffff, 0x7d1, 0xffff, 0x535, 0x5be, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0x12cc, 0xffff, 0xe9b, 0x45e, 0xffff, 0xffff, 0x44e, 0xffff, 
0xffff, 0xffff, 0x9f4, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0x109f, 0xffff, 0x82, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xf4b, 
0xffff, 0xffff, 0xffff, 0x12f5, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xc85, 0xffff, 0xffff, 0x64c, 0xab0, 0xffff, 
0xffff, 0x831, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0x94a, 0xb3f, 0xffff, 0xffff, 0xffff, 0x13d, 0xffff, 0xa33, 
0x1014, 0xffff, 0xffff, 0xa7b, 0xffff, 0xffff, 0xffd, 0x46e, 
0xffff, 0xa6b, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x1e5, 
0xffff, 0xffff, 0xffff, 0x840, 0x3cd, 0x64f, 0xa3e, 0xffff, 
0xffff, 0xd8e, 0xffff, 0xfaf, 0xffff, 0xadd, 0xffff, 0xffff, 
0xffff, 0x94b, 0x5f5, 0x962, 0x1e6, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0x673, 0xffff, 0xffff, 0xcbe, 0x1ad, 0xa23, 
0x1060, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x1c5, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0x400, 0x62c, 0x975, 0xffff, 
0xffff, 0xffff, 0xded, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xc60, 
0xfb, 0x11c9, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0x9cd, 0xffff, 0xffff, 0xffff, 0x8e1, 
0xffff, 0xffff, 0x531, 0xffff, 0xffff, 0xffff, 0x1082, 0x123c, 
0xffff, 0xffff, 0xafa, 0xb7c, 0x12c7, 0xffff, 0x30f, 0xffff, 
0xb87, 0xffff, 0xffff, 0xffff, 0xffff, 0x2ad, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x1045, 0x43e, 0xffff, 
0xffff, 0xffff, 0xb80, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xd83, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0x1246, 0xf83, 0x247, 0xffff, 0x22b, 0x121a, 0xffff, 0xffff, 
0xffff, 0xffff, 0x24a, 0xffff, 0xffff, 0x74, 0xffff, 0xffff, 
0xffff, 0x5ba, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0x9b3, 0xffff, 0xffff, 0xffff, 0xffff, 0xd58, 
0x8d9, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0x686, 0xa96, 0xffff, 0xffff, 0x63, 0xffff, 0x60e, 
0x11b9, 0xaeb, 0xffff, 0xffff, 0xffff, 0x5b9, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0x1150, 0x502, 0xffff, 0xffff, 0xb9a, 
0xffff, 0xffff, 0x712, 0xf1d, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0x29b, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0x16e, 0xac4, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0x7bb, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0x10c9, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0x11f4, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0x20e, 0xffff, 0xffff, 0xffff, 0xffff, 0x7e4, 0xef7, 
0xffff, 0xffff, 0xffff, 0x399, 0xffff, 0xffff, 0xffff, 0xffff, 
0xc7b, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xed3, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0x12f1, 0xe51, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x156, 
0xffff, 0xffff, 0xffff, 0xffff, 0xd9c, 0x527, 0x88d, 0xb71, 
0x11c3, 0xffff, 0xffff, 0x11f8, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0x109b, 0xffff, 0xffff, 0x7ad, 0x571, 0xffff, 0xe95, 0x1025, 
0x1070, 0xffff, 0xffff, 0xffff, 0xffff, 0xb28, 0xa99, 0x1171, 
0x927, 0xffff, 0x742, 0xdfe, 0x72a, 0xffff, 0xffff, 0xffff, 
0x331, 0x38d, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0x434, 0xffff, 0xdc1, 0x643, 0xa9a, 
0xffff, 0x157, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0x9f5, 0x1ed, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xfa1, 0xffff, 0xffff, 
0xffff, 0xffff, 0x2ef, 0xffff, 0xffff, 0xffff, 0xffff, 0x37, 
0xffff, 0x2a3, 0xffff, 0xffff, 0x84, 0x687, 0xd15, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0x154, 0x29e, 0xffff, 0xffff, 
0x8a0, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x858, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xa83, 
0x2e5, 0x463, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0x4ba, 0xffff, 0x76, 0x169, 0x934, 0x11d9, 0xffff, 
0xbc0, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x145, 0xffff, 
0xffff, 0x71e, 0xffff, 0xffff, 0xffff, 0xbb4, 0x1212, 0xffff, 
0xffff, 0xca2, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x5d2, 
0x90b, 0x123d, 0xffff, 0xffff, 0xffff, 0x4a4, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x7e0, 
0xffff, 0x50b, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x1f9, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0x245, 0xffff, 0xe58, 0xffff, 
0xffff, 0xffff, 0x2da, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xde1, 0xffff, 0xffff, 0xffff, 0x1181, 0xffff, 0xffff, 0xffff, 
0xd5e, 0xffff, 0x787, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x73f, 
0xffff, 0xffff, 0xffff, 0xffff, 0xebc, 0xffff, 0xffff, 0xffff, 
0xb0e, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0x4b2, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x9f9, 0xffff, 0xffff, 
0xffff, 0xffff, 0x683, 0xffff, 0xffff, 0xffff, 0x227, 0x79a, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x5a2, 0xfe6, 
0x887, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xc08, 
0x1224, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x3ba, 0xffff, 
0xffff, 0xffff, 0xffff, 0x48a, 0xffff, 0x39a, 0x1182, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0x10a0, 0xffff, 0xffff, 0x129f, 
0x5cb, 0x9d1, 0xffff, 0xffff, 0xffff, 0x83e, 0x1249, 0xffff, 
0xffff, 0xffff, 0x8b7, 0x8c9, 0xffff, 0x9eb, 0xffff, 0x588, 
0xe75, 0xffff, 0xffff, 0xffff, 0x11e, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0x9a, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0x8e, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xa8f, 
0xffff, 0xffff, 0xffff, 0x530, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0x10ed, 0xfda, 0xffff, 0xffff, 0xffff, 0x47c, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0x54f, 0xffff, 0xffff, 0xffff, 0xf46, 0x10e6, 
0xffff, 0xffff, 0xffff, 0xffff, 0xb66, 0xffff, 0xffff, 0xffff, 
0xffff, 0xac0, 0xa92, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x802, 
0xe, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xb74, 
0xffff, 0xffff, 0xefa, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xa5a, 0xffff, 
0xffff, 0xffff, 0x10ef, 0xffff, 0xe4f, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x87, 
0xffff, 0xffff, 0xffff, 0xffff, 0xec9, 0xffff, 0xffff, 0xffff, 
0xffff, 0x1245, 0xffff, 0x12e1, 0xffff, 0xffff, 0xffff, 0x11a3, 
0xffff, 0xffff, 0x868, 0xffff, 0xffff, 0xa2c, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x5df, 0x1250, 
0xffff, 0x2b2, 0xffff, 0xffff, 0xffff, 0xb2d, 0xffff, 0xffff, 
0xffff, 0x2ac, 0x21b, 0x344, 0xffff, 0x6a2, 0xa3f, 0xffff, 
0xffff, 0xffff, 0xffff, 0x87b, 0xfd8, 0xffff, 0xffff, 0x1062, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xa17, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xe14, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0x7ef, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0x16a, 0x928, 0xffff, 0xffff, 0xffff, 0xcc, 0xffff, 
0x727, 0xa26, 0xcd7, 0x1297, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xd, 0xffff, 0xffff, 0xffff, 0xffff, 0x5de, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x1069, 
0xffff, 0xffff, 0xb22, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0x460, 0xffff, 0xffff, 0xffff, 
0xcb1, 0xf35, 0xffff, 0xffff, 0xf8d, 0xffff, 0xffff, 0xa5f, 
0xcaa, 0x7e2, 0xffff, 0xffff, 0xffff, 0x6b5, 0xffff, 0xffff, 
0xffff, 0xffff, 0x5db, 0x1115, 0xffff, 0x120f, 0xffff, 0xffff, 
0xffff, 0x7fe, 0xffff, 0xffff, 0x1131, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x3eb, 0x68d, 0xffff, 
0xffff, 0xffff, 0x1114, 0xffff, 0xffff, 0x758, 0xffff, 0xffff, 
0x176, 0xffff, 0xffff, 0xffff, 0xeff, 0x36a, 0xffff, 0xffff, 
0x98d, 0xc7c, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xed0, 
0xffff, 0xffff, 0xffff, 0x196, 0x80c, 0xffff, 0xffff, 0x60, 
0x84e, 0x1282, 0xffff, 0x779, 0xffff, 0x71c, 0xffff, 0xffff, 
0xffff, 0xffff, 0x8fc, 0x479, 0xffff, 0xffff, 0xffff, 0x1214, 
0x6a8, 0xffff, 0xffff, 0xffff, 0xffff, 0x356, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0x175, 0xffff, 0xe6d, 0xffff, 0x68, 
0xffff, 0x115, 0x24b, 0xffff, 0xffff, 0xffff, 0x62f, 0xffff, 
0xffff, 0x12a9, 0xffff, 0xffff, 0xffff, 0xc6e, 0x10c7, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x554, 0xe10, 
0xf37, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x7e1, 0xf39, 
0xffff, 0xffff, 0xffff, 0xd3a, 0xffff, 0xffff, 0xffff, 0xffff, 
0xd79, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xce9, 0xffff, 0x7ae, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xb39, 0xffff, 
0xffff, 0x915, 0xffff, 0xffff, 0xffff, 0x665, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0x14b, 0xffff, 0xe5d, 0xffff, 0xffff, 0xffff, 
0xd4c, 0xe5, 0xecb, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xc75, 0xffff, 
0x10e5, 0x3f6, 0xf48, 0x108c, 0xfd, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xce6, 0xffff, 0xffff, 0xcbc, 0xffff, 
0x77c, 0x11ba, 0xffff, 0x8b4, 0xffff, 0xffff, 0xffff, 0xffff, 
0xa3a, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x721, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x5c9, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xdaa, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x6f2, 
0xffff, 0xe80, 0xffff, 0xffff, 0xffff, 0xffff, 0xb86, 0xffff, 
0x600, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xaf4, 0x1263, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0x879, 0xffff, 0xffff, 0xffff, 0x9ec, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0x111, 0xc9c, 0x1064, 0xffff, 
0xffff, 0x1087, 0xffff, 0xffff, 0x3c7, 0x9e9, 0x299, 0xdab, 
0xe2d, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xaf5, 0xffff, 0xffff, 0xffff, 0xffff, 0xd61, 0xdac, 0x944, 
0xffff, 0xffff, 0x10bf, 0xffff, 0xffff, 0xa, 0x7d9, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x8d3, 0xffff, 0xffff, 
0xffff, 0xffff, 0x5bb, 0xffff, 0xffff, 0xffff, 0xffff, 0x403, 
0xffff, 0xe37, 0xa48, 0xffff, 0x94d, 0xffff, 0x10, 0x192, 
0xffff, 0x736, 0xffff, 0xffff, 0xffff, 0xffff, 0x1088, 0xffff, 
0x124b, 0xc52, 0xf3a, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0x980, 0xffff, 0xffff, 0xffff, 0x77d, 0xffff, 
0xffff, 0x2b5, 0xffff, 0xffff, 0xffff, 0x78b, 0xffff, 0x4ae, 
0x6b, 0xc21, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0x6, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x4a, 
0xffff, 0xffff, 0x15d, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0x524, 0x452, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0x164, 0x118b, 0xffff, 0x7b1, 
0xffff, 0xffff, 0xffff, 0x685, 0xffff, 0xd8c, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xe1b, 0x426, 0x110b, 
0x112f, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xcd6, 
0xffff, 0xffff, 0xffff, 0xffff, 0x10ba, 0xffff, 0xffff, 0xffff, 
0x125b, 0xffff, 0xffff, 0xffff, 0xeea, 0x10b1, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0x1017, 0xffff, 0xffff, 0x305, 0xffff, 0x1072, 0xda, 
0x131, 0x988, 0x46, 0x7c8, 0xd07, 0xffff, 0x11d5, 0xffff, 
0xffff, 0x12c0, 0x916, 0xffff, 0xffff, 0xffff, 0xffff, 0x1121, 
0xef8, 0xe12, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0x76e, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0x1b8, 0xffff, 0xffff, 0xffff, 0x6b1, 0xffff, 0xd2f, 0xffff, 
0xffff, 0x647, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xae4, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0x5a4, 0xa74, 0xffff, 0xffff, 0x121c, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xcd1, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0x1100, 0xffff, 0xffff, 0x10be, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xe2a, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0x3f4, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0x45a, 0xb5e, 0xc6d, 0x12c8, 0xffff, 0x11, 
0xffff, 0xffff, 0xffff, 0xffff, 0x1286, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x28b, 0xffff, 0xffff, 
0xffff, 0x1c, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xfcd, 
0xffff, 0xffff, 0x224, 0x1145, 0x585, 0x117e, 0xffff, 0x1058, 
0xffff, 0xa20, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x229, 
0xffff, 0x2b9, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x4bb, 0xffff, 0xffff, 
0xffff, 0xffff, 0x123e, 0xffff, 0xffff, 0xffff, 0xffff, 0xe1, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x209, 0xffff, 0xffff, 
0xffff, 0xffff, 0x35c, 0x959, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0x2d, 0xffff, 0x4b7, 0x89e, 0xffff, 0xffff, 0xffff, 
0xffff, 0x642, 0xffff, 0xffff, 0x473, 0xffff, 0xffff, 0xffff, 
0xce4, 0x812, 0x7ac, 0xa72, 0xfaa, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0x491, 0xffff, 0xffff, 0xffff, 
0xfd0, 0xffff, 0x445, 0xffff, 0x3e1, 0x2d0, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xe83, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0x104a, 0x493, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0x1bb, 0xffff, 0xffff, 0x81a, 
0xfb2, 0xffff, 0xffff, 0x995, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xf3f, 0xffff, 0xda2, 0xffff, 0xffff, 
0x5ec, 0xffff, 0xffff, 0x5c5, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xb9d, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xc0b, 0x11ab, 
0x72, 0x739, 0xf02, 0xf6f, 0x108f, 0x8ee, 0x1183, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0x18b, 0x770, 0x8e7, 0x921, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0x12a, 0x108a, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xb56, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0x9f7, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xbb, 0x201, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xb41, 0xffff, 0xffff, 0xffff, 0x339, 0xffff, 0xffff, 0x11ea, 
0xffff, 0xffff, 0xffff, 0xffff, 0xdf1, 0xa77, 0x971, 0x3f5, 
0x26f, 0xffff, 0xffff, 0xffff, 0xffff, 0xcd0, 0xffff, 0xffff, 
0x114c, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0x1, 0x50a, 0xffff, 0xc03, 0xffff, 0x4d9, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0x9c, 0x759, 0x904, 0xffff, 0xfd5, 0xffff, 0xffff, 0xffff, 
0x128e, 0xffff, 0xffff, 0xffff, 0x1303, 0xffff, 0xffff, 0xffff, 
0xffff, 0x748, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0x85e, 0xffff, 0x606, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0x334, 0xffff, 0x9c7, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x1b0, 0xffff, 
0xffff, 0x913, 0xffff, 0xffff, 0x752, 0x1154, 0xffff, 0xffff, 
0x564, 0xa9d, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0x1b7, 0x1177, 0x368, 0x509, 0xffff, 0xffff, 0xffff, 
0xffff, 0xa56, 0x135, 0x4f1, 0xffff, 0x762, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x3b0, 0xffff, 
0x64, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x1bf, 0xffff, 0x109a, 
0xffff, 0xc, 0x3d9, 0xffff, 0xf26, 0xffff, 0x187, 0xc5e, 
0xffff, 0xffff, 0xffff, 0xdec, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0x221, 0xffff, 0xffff, 0x4, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xbf4, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0x1008, 0xffff, 0xffff, 0xf11, 0xffff, 
0xbd4, 0xffff, 0xffff, 0xffff, 0x593, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0x10ee, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xcfa, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0x6af, 0xffff, 0xffff, 0xffff, 0xffff, 0x12bf, 0xec1, 0xf65, 
0x5a0, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xc86, 0x378, 0xff3, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x464, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xdb9, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xea4, 0xffff, 
0x3c1, 0xffff, 0xffff, 0xffff, 0xffff, 0x634, 0xffff, 0xffff, 
0xffff, 0xffff, 0x447, 0xffff, 0xc59, 0x50f, 0x5f4, 0xf40, 
0x1090, 0xffff, 0xffff, 0xc33, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xf63, 0xffff, 0xde9, 0xffff, 0x72e, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xdf3, 0xdf2, 0x10a1, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0x576, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0x42d, 0xffff, 0xffff, 0xffff, 0x120a, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xc31, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0x9b1, 0xffff, 0xffff, 0xffff, 0xe71, 0xffff, 0xffff, 0xffff, 
0xffff, 0x9bb, 0xffff, 0x1051, 0x1124, 0xffff, 0xffff, 0x332, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0x528, 0xffff, 0xffff, 0xffff, 0xffff, 0x12b1, 0xffff, 0x173, 
0xffff, 0xffff, 0x48, 0xffff, 0x59, 0xffff, 0xffff, 0xae2, 
0xffff, 0x1078, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xfc, 
0xffff, 0xffff, 0x926, 0xffff, 0xffff, 0xffff, 0xffff, 0xa50, 
0xffff, 0x1156, 0xaef, 0xffff, 0xffff, 0xfa6, 0xffff, 0xb20, 
0x25a, 0xffff, 0x1218, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0x12ae, 0xbf, 0xffff, 0x8c8, 0xf27, 0xffff, 0xffff, 0x77, 
0x309, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x8f8, 0x750, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xa21, 0xffff, 0xffff, 0xffff, 0xffff, 0xbc2, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x4c2, 
0x9ac, 0x9e8, 0xc67, 0x10b7, 0xffff, 0xffff, 0xc8f, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0x1057, 0xffff, 0xffff, 0x256, 0xffff, 0xffff, 0xffff, 
0x34a, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xdd5, 0xc5c, 0xffff, 
0xffff, 0xffff, 0x31f, 0x5c0, 0xdca, 0xffff, 0xffff, 0xffff, 
0x12d7, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x96a, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x1ce, 0xffff, 0xcf1, 
0xffff, 0xffff, 0xffff, 0xffff, 0xf07, 0xffff, 0xffff, 0xffff, 
0x70, 0xffff, 0xffff, 0xffff, 0x845, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xa37, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0x3ac, 0xffff, 0x623, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xaa, 0x77b, 
0x127c, 0xffff, 0xffff, 0xffff, 0xffff, 0xf38, 0xffff, 0xffff, 
0x4e9, 0xffff, 0xffff, 0x314, 0x25b, 0xffff, 0xffff, 0xffff, 
0x402, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xc53, 0x421, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0x312, 0xffff, 0x10fb, 0xffff, 0xffff, 0x5cc, 0xbf6, 0x672, 
0xffff, 0xffff, 0xffff, 0xffff, 0xa00, 0xffff, 0xffff, 0xffff, 
0xffff, 0x193, 0xffff, 0xffff, 0xffff, 0x9b5, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xf7a, 0xffff, 0xffff, 0x1018, 0x120c, 
0x1233, 0x9c1, 0x7d, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xe99, 0xffff, 0x151, 0xc94, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0x75c, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0x4df, 0xffff, 0x165, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0x89, 0xffff, 0x80, 0xffff, 0xbcd, 
0xe18, 0xffff, 0xffff, 0xffff, 0xa7a, 0xffff, 0x657, 0x8c6, 
0xb2f, 0xcae, 0x6cc, 0xffff, 0xffff, 0x2d7, 0xffff, 0xffff, 
0xffff, 0xbc6, 0xfc7, 0xffff, 0x734, 0xfd3, 0xffff, 0x8f6, 
0xffff, 0xffff, 0xfde, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0x83b, 0xffff, 0xffff, 0xffff, 0xa87, 0xffff, 
0xffff, 0xffff, 0x11d, 0x42b, 0x2c7, 0xffff, 0xffff, 0xffff, 
0xffff, 0x128b, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xc20, 0xee7, 0xffff, 0x7d5, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x110d, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0x10c6, 0xffff, 0x8a2, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0x10d7, 0xffff, 0xffff, 0x88c, 0xffff, 0xedb, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x10b2, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x7d0, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0x83d, 0xffff, 0x897, 0xffff, 0xffff, 0xe41, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0x8d6, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0x12b7, 0xffff, 0xb1, 0x36d, 0xffff, 0xffff, 
0xffff, 0xffff, 0x125c, 0xffff, 0xffff, 0xffff, 0x81f, 0xfd9, 
0xffff, 0xffff, 0xffff, 0xffff, 0x3f8, 0x168, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x74a, 0xffff, 
0x1039, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xbf2, 0xffff, 0xffff, 0xe5b, 0x2a4, 0x656, 0x10a7, 
0xffff, 0x1116, 0xffff, 0x9c5, 0xffff, 0x5b, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xb0d, 0xffff, 0xffff, 0xffff, 0xffff, 0x366, 0xcc4, 
0xffff, 0xffff, 0xffff, 0xffff, 0x91b, 0xffff, 0x11af, 0xffff, 
0xffff, 0xffff, 0xffff, 0x159, 0xffff, 0xffff, 0xffff, 0xffff, 
0xac8, 0xffff, 0xd3b, 0xffff, 0xfc9, 0xffff, 0x6b2, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0x1dc, 0xffff, 0xffff, 0xffff, 0xffff, 0x365, 
0x86a, 0x9c0, 0xe69, 0xffff, 0x10f7, 0xffff, 0xdd6, 0xffff, 
0x4ef, 0xffff, 0xffff, 0xffff, 0xccd, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xdba, 0xaab, 0x10f1, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xc35, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0x2bb, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x1117, 
0xffff, 0xffff, 0xffff, 0x874, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0x12a4, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xbb9, 0xffff, 0x1d2, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x3f3, 0x756, 0xffff, 
0xffff, 0xa4f, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x48e, 
0xffff, 0xea7, 0xffff, 0xffff, 0xffff, 0xffff, 0xbd9, 0xe01, 
0xffff, 0xffff, 0xffff, 0xa40, 0xffff, 0xffff, 0x203, 0x118e, 
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0xffff, 0xbdf, 0xffff, 0xffff, 0xffff, 
0x1e1, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0x119, 0xffff, 0xffff, 0x59f, 0xffff, 
0xffff, 0xfe4, 0x33e, 0x1020, 0x11dd, 0xffff, 0xb5c, 0xffff, 
0xeaf, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
0xffff, 0xffff, 0xffff, 0x468, 0xffff, 0xb60, 0xb05, 0xffff, 
0x1138, 0x12e5, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 
};

